/* register.c */

#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "loadimg.h"

#ifdef __MINGW32__
#include <Windows.h>
#endif

/* these functions are available to the calling program */
tryLoadProc tryLoad;
loadBandProc loadBand;
endLoadProc endLoad;

/* if numLoader is negative, returns number of loader functions, otherwise selects the */
/* one which is passed as parameter */
/* loader 0 is uses heuristics on filename suffix to use the most probable loader first. */
/* if no such loader can be found, loader returns the most probable loader for the given application */
/* and keep calling loaders 1...NUM_LOADERS */
/* if a valid heuristics can be used, loader 0 returns the number of loader which it is using */
int returnLoaderFunctions(int numLoader, char *filename)
{
 if ((numLoader>=0)&&(numLoader<NUM_LOADERS))
  switch (numLoader)
   {
    case 0:
     {
      unsigned int suffix=0;
      int chosen;
      int fnameLen = strlen(filename);
      
      if (fnameLen<5)
       return(-1);
      suffix =  tolowerAZ(filename[fnameLen-1]);
      suffix += (tolowerAZ(filename[fnameLen-2])<<8);
      suffix += (tolowerAZ(filename[fnameLen-3])<<16);
      if ((filename[fnameLen-5]=='.') || (tolowerAZ(filename[fnameLen-5])=='d')) /* allow .xxxx or .drape suffix */
       suffix += (tolowerAZ(filename[fnameLen-4])<<24);
      switch (suffix)
       {
        case ('b'<<16)+('m'<<8)+'p': chosen = 8; break;
        case ('g'<<16)+('i'<<8)+'f': chosen = 1; break;
        case ('j'<<24)+('p'<<16)+('e'<<8)+'g':
        case ('j'<<16)+('p'<<8)+'g': chosen = 4; break;
        case ('p'<<16)+('c'<<8)+'x': chosen = 3; break;
        case ('p'<<16)+('n'<<8)+'g': chosen = 2; break;
        case ('t'<<24)+('i'<<16)+('f'<<8)+'f':
        case ('t'<<16)+('i'<<8)+'f': chosen = 5; break;
        case ('w'<<16)+('i'<<8)+'f': chosen = 7; break;
        default: chosen = 4; /* then it is most likely JPEG */
       }
      if (chosen>0)
       returnLoaderFunctions(chosen, filename);
      return(chosen);
     }
     break;
    case 1: registerLoadGIF(); break;
    case 2: registerLoadPNG(); break;   
    case 3: registerLoadPCX(); break;
    case 4: registerLoadJPEG(); break;
    case 5: registerLoadTIFF(); break;   
    case 6: registerLoadPBM(); break;
    case 7: registerLoadWIF(); break;   
    case 8: registerLoadBMP(); break;
   /* dont't forget to fix returnLoaderString() below, if you are adding new formats! */
  }
//printf("file %s using loader %d\n", filename, numLoader);
 return(NUM_LOADERS);
}

/* returns the string that should explain the loader's file type */
char *returnLoaderString(int numLoader)
{
 /* nota bene! the order of these strings must match the above case sentence! */
 static char *loaderID[] = {"", "GIF", "PNG", "PCX", "JPEG", "TIFF", "PBM", "WIF", "BMP" };

 if ((numLoader<NUM_LOADERS) && (numLoader))
  return(loaderID[numLoader]);
 else return("");
}

char *returnLoaderStringNoLogo(int numLoader)
{
 /* nota bene! the order of these strings must match the above case sentence! */
 static char *loaderID[] = {"", " GIF", " PNG", " PCX", " JPEG", " TIFF", " PBM", " WIF", " BMP" };

 if ((numLoader<NUM_LOADERS) && (numLoader))
  return(loaderID[numLoader]);
 else return("");
}

char *returnLoaderStringJustLogo(int numLoader)
{
 /* nota bene! the order of these strings must match the above case sentence! */
 static char *loaderID[] = {"", "emblem-photos", "emblem-photos", "emblem-photos", "emblem-photos", "emblem-photos", "", "", "emblem-photos"};

 if ((numLoader<NUM_LOADERS) && (numLoader))
  return(loaderID[numLoader]);
 else return("");
}

/* internal initialisation of the image loading library */
extern void initImageLoadLibrary()
{
#ifdef __MINGW32__
#ifdef DEBUG
 /* so we can use printf in windows DLL */
 AllocConsole();
 freopen("CONIN$", "r", stdin);
 freopen("CONOUT$", "w", stdout);
 freopen("CONOUT$", "w", stderr);
#endif
#endif
}

/* registers a loader with system */
void registerLoader(tryLoadProc tLoad, loadBandProc lBand, endLoadProc eLoad)
{ tryLoad = tLoad; loadBand = lBand; endLoad = eLoad; }
