/***************************************************************************
                          apaintBrushList.h  -  description
                             -------------------
    begin                : Thu Jul 15 2004
    copyright            : (C) 2004-2011 Arahne, d.o.o.
    email                : arahne@arahne.si
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef BRUSH_LIST_H
#define BRUSH_LIST_H

#include <qptrlist.h>
#include "apaintimage.h"
#include <qimage.h>
#include <qpointarray.h>

class IconList;
class Layer;

class ApaintBrushList:public QObject {
   Q_OBJECT
public:

  ~ApaintBrushList();
  static ApaintBrushList *getObject();
  static void deleteObject();

  void addBrush(Layer *selLayer);
  void loadBrush(int numCol, QRgb *ctab);
  void removeBrush();
  ApaintImage getBrush();
  QString getBrushName();
  QImage getBrushMask();
  bool isValid();
  void mergeColorTables(Layer *selLayer);
  bool isColorTabDifferent() {return diffColorTab;}
  IconList *getIconList() {return iList;}
  void updateBrushList(int fhei, bool showDirs);
  bool useBrush() {return isUsingBrush;}
  void enableBrush(bool enable) { isUsingBrush = enable;}
  int *getTransTable() {return transTable;}
  void show();
  void hide() {showBrush = false;}
  bool doShowBrush() {return showBrush;}
  void setPos(QRect newPos) {brushPos = newPos;}  
  QRect getPos() {return brushPos;}
  int posx() {return brushPos.x();}
  int posy() {return brushPos.y();}
  int posWid() {return curBrush.width();}
  int posHei() {return curBrush.height();}
  void mirrorHorizontal();
  void mirrorVertical();
  void mirrorFlip();
  void mirrorExplode();
  void setOffsetX(int x) {offsetX = x;}
  void setOffsetY(int y) {offsetY = y;}
  int getOffsetX() {return offsetX;}
  int getOffsetY() {return offsetY;}
  void setAbsOffset(bool enable) { absOffset = enable;}
  bool getAbsOffset() { return absOffset;}
  
signals:
  void listIsChanged();
  void hasBrushes(bool);
    
private:
   ApaintBrushList();

   QString getMaskName(QString imageName);
   
  ApaintImage curBrush;
  QImage curMask;
  QString curMaskPath;
  QRect brushPos;
  bool diffColorTab, isUsingBrush;
  IconList *iList;
  int transTable[256];
  QString brushDir;
  bool brushDirOK;
  bool showBrush;
  int offsetX, offsetY;
  bool absOffset;
};

#endif
