/***************************************************************************
                          apaintUtil.h  -  description
                             -------------------
    begin                : Thu Apr 15 2004
    copyright            : (C) 2004-2011 Arahne, d.o.o.
    email                : arahne@arahne.si
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef APAINTUTIL_H
#define APAINTUTIL_H

#include <qvalidator.h>
#include <qcheckbox.h>
#include "dialog_ui_enterText.h"

#define  DIR_SEPARATOR  "/"

class QVBoxLayout;

class CommonProp: public QWidget {
  Q_OBJECT
public:
  ~CommonProp();
   static CommonProp *getObject();
   static void createObject(QWidget *parent, const char *name);
   void setThickLinesProp(bool show);
   void setAspectDrawProp(bool show);
   bool doAspectDraw() {return cbDrawInAspect->isChecked();}
   void enableAspectDrawProp(bool enable);

signals:
  void setThickLinesSig(bool enable);
  void setAspectDrawSig(bool enable);
  void aspectDrawChangedSig();
  
private:
  CommonProp(QWidget *parent, const char *name);
  QCheckBox *cbThickLines, *cbDrawInAspect;
  QVBoxLayout *vbox;
  
};
/***********************************************************************************/

class ApaintUtil {
public:
 static QString getBaseName(QString &allName);
 static QString getDirName(QString &allName);
};

class  ApaintIntValidator : public QValidator
{
public:
    ApaintIntValidator( int bottom, int top, QObject * parent, const char *name);
    ~ApaintIntValidator();

    QValidator::State validate( QString &, int & ) const;


private:
    ApaintIntValidator( const ApaintIntValidator & );
    ApaintIntValidator& operator=( const ApaintIntValidator & );

    int b, t;
};

/***********************************************************************************/

class ApaintDoubleValidator : public QValidator
{

public:
    ApaintDoubleValidator( double bottom, double top, int decimals, QObject * parent, const char *name);
    ~ApaintDoubleValidator();

    QValidator::State validate( QString &, int & ) const;

private:
    ApaintDoubleValidator( const ApaintDoubleValidator & );
    ApaintDoubleValidator& operator=( const ApaintDoubleValidator & );

    double b, t;
    int d;
};

/***********************************************************************************/

class EnterTextDialog : public EnterTextDialogUI
{
  Q_OBJECT
private:
  EnterTextDialog(QWidget *parent, const char* name, bool modal);
public:
  ~EnterTextDialog(){};
  
  static bool  getText(QWidget *parent, QString title, QString label, QString descr, QString &out);
};

/***********************************************************************************/

class QWMatrix;

class MatrixUtil {
public:
 static void translatemx(QWMatrix *m, double x, double y);
 static void scalemx(QWMatrix *m, double x, double y);
 static void rotatemx(QWMatrix *m, double a, double xyRatio);
 static void shearmx(QWMatrix *m, double sx, double sy);  
};

/***********************************************************************************/

class QVBoxLayout;

class VerticalyAlignedWidget: public QWidget {
public:
    VerticalyAlignedWidget(QWidget *parent, const char *name);
    ~VerticalyAlignedWidget();

    void addWidget(QWidget *w);
private:
  QVBoxLayout *vbox;    
};

/***********************************************************************************/

class ApaintDoc;

class ApaintDocUtil {
public:
  static void destroyCurLayer(ApaintDoc *doc);
  static void moveCurLayer(ApaintDoc *doc, int dx, int dy, int border);
};
/***********************************************************************************/

/* map return codes */
#define PMAP_BAD       -1     /* mapping doesn't exist (bad correspondence) */
#define PMAP_AFFINE     0     /* mapping is affine */
#define PMAP_PROJECTIVE 1     /* mapping is projective */

class PerspectiveTransform {
public:  
   static double adjoint(double a[3][3], double b[3][3]);
   static int quad2rect(double u0,double v0,double u1,double v1, QPoint quad[4], double RQ[3][3] ,double QR[3][3]);
   static int square2quad(QPoint quad[4], double SQ[3][3]);
   static void transformPoint(double p[3], double a[3][3], double q[3]);
   static void transformPoint(int x, int y, int &tx, int &ty, double mx[3][3]);
};

#endif

