/***************************************************************************
                          apaintdoc.cpp  -  description
                             -------------------
    begin                : Mon Jan 27 11:15:43 CET 2003
    copyright            : (C) 2003-2011 Arahne, d.o.o.
    email                : arahne@arahne.si
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <qmessagebox.h>
#include <qfiledialog.h>
#include <qpainter.h>
#include <math.h>
#include <qbitmap.h>
#include <string.h>
#include <qclipboard.h>
#include "apaintdoc.h"
#include "toolTransform.h"
#include "apaintconfig.h"
#include "apaintimageUtil.h"
#include "apaintImageResize.h"
#include "apaintBrushList.h"
#include "apaint.h"

#define NUM_POINTS  8

extern "C"{
#include "../../iolib/acolor.h"   
#include "../../iolib/browseio.h"
}

#define rad_to_deg(angle) ((angle) * 360.0 / (2.0 * M_PI))
#define REF_SIZE   5


/************************LAYER DRAG***************************/
LayerDrag::LayerDrag():QMimeSource()
{
  useMask = false;
}

LayerDrag::~LayerDrag()
{

}

const char *LayerDrag::format(int i) const
{
  if (i==0)
   return (APAINT_MIME_SIG);
  else
   return 0;
}

bool LayerDrag::provides(const char *mimeType) const
{
 if (!strcmp(mimeType, APAINT_MIME_SIG))
  return true;
 else
  return false;
}

QByteArray LayerDrag::encodedData(const char *typ) const
{  
  if (!strcmp(typ, APAINT_MIME_SIG))
  {
    QByteArray d;
    QDataStream s(d, IO_ReadWrite);
    if (useMask)
     s << (Q_INT8)1;
    else
     s << (Q_INT8)0; 

    s << (ApaintImage &)img;          
    if (useMask)
      s << mask;
           
    return d;
  }
  else
    return QByteArray();
}

void LayerDrag::setAImage(ApaintImage *simg)
{
 img = simg->copy();
}

void  LayerDrag::setMask(QImage& m)
{ 
 mask = m.copy();
 useMask = true;
}

bool LayerDrag::usingMask()
{
  return useMask;
}

/************************LAYER DRAG***************************/

Layer::Layer(QPointArray &borderPts, LayerType ltyp, LayerGeometry lg, double ax, double ay)
{
 initLayer(ltyp, lg, ax, ay); 
 freehandBorder = borderPts.copy(); 
}

Layer::Layer(LayerType ltyp, LayerGeometry lg, double ax, double ay)
{
 initLayer(ltyp, lg, ax, ay);
}

Layer::Layer(int wid, int hei, int depth, uint fillCol, int numCol, LayerType ltyp, double ax, double ay)
{   
 initLayer(ltyp, RectangleG, ax, ay);
 imageBuf.newImage(wid, hei, depth);
 if (!imageBuf.isNull())
 {
   if (depth>8)
    imageBuf.fill(fillCol);
   else if (depth == 8)
   {
     imageBuf.setNumColors(numCol);
     imageBuf.setRandomColors(numCol);
     imageBuf.setColor(0, fillCol);
     imageBuf.fill(0);
   }
   setNewData(imageBuf, 0);
   setLayerName();
 }
}

Layer::Layer(const QString &fileName, LayerType ltyp, double ax, double ay)
{
 initLayer(ltyp, RectangleG, ax, ay); 
 if (!imageBuf.load(fileName))
   validDraw = false;
 else
  {
   setNewData(imageBuf, 0);
   setLayerName();
   /* here we saving the history file */
   if (!strstr(fileName.ascii(), "repeatDraw.png")) /* don't save history of reading this system configuration file */
    {
     char imageHistoryFile[500];

     strcpy(imageHistoryFile, getImageIconPath());
     strcat(imageHistoryFile, "/history");
     saveRecentFile((char *) fileName.local8Bit().data(), TRUE, imageHistoryFile, ApaintConfig::getNumRecentFiles());
    }
  }
}

Layer::Layer(ApaintImage &img, int x, int y, QImage *imask, LayerType ltyp, double ax, double ay)
{
 initLayer(ltyp, RectangleG, ax, ay); 
 setNewData(img, x, y, imask);
 setLayerName();
}

Layer::~Layer()
{
 tileList.clear();
 refArea.clear();
 if (srcZoomX) delete[] srcZoomX;
 if (srcZoomY) delete[] srcZoomY;
}

void Layer::initLayer(LayerType ltyp, LayerGeometry lg, double ax, double ay)
{ 
 resetColors();
 fgNum = 1;
 bgNum = 0;
 useMask=false;
 hidden = false;
 isResizeable = false;
 previewMode = false;
 setDrawDelimiterLines(false);
 vmode = NormalView;
 validDraw = true; 
 type = ltyp;
 geometry = lg;
 useTiles = false;
 tilex = tiley = 1;
 tileOffx = tileOffy = 0;
 tileList.append(new Tile(QRect(), QRect()));
 tileList.setAutoDelete(true);
 refArea.setAutoDelete(true);
 valid = false;
 pos = QRect(0, 0, 0, 0);
 ptsBoundary = 0;
 transPtsBoundary = 0;
 numBoundSeg = 0;
 aspectX = ax;
 aspectY = ay;
 srcZoomX = srcZoomY = 0;
 doShowBorder = true;
}

QRect Layer::drawAreaPos()
{
  if (isUsingTiles())
    return QRect((int)(pos.x()*aspectX), (int)(pos.y()*aspectY), pix1Bit.width()*tilex, pix1Bit.height()*tiley);

  return QRect((int)(pos.x()*aspectX), (int)(pos.y()*aspectY), pix1Bit.width(), pix1Bit.height());
}

void Layer::updateSize()
{
 if (isUsingTiles())
  pos = QRect(pos.x(), pos.y(), imageBuf.width()*tilex, imageBuf.height()*tiley);
 else
  pos = QRect(pos.x(), pos.y(), imageBuf.width(), imageBuf.height());
}


bool Layer::initPixmap(int wid, int hei)
{
  pix1Bit = QPixmap((int)(wid*aspectX), (int)(hei*aspectY), 1);
  if (pix1Bit.isNull())
   return(false);
  pix1Bit.fill(Qt::black);

  if (srcZoomX) delete[] srcZoomX;
  if (srcZoomY) delete[] srcZoomY;
  int swid = pix1Bit.width();
  int shei = pix1Bit.height();
  srcZoomX = new int[swid];
  srcZoomY = new int[shei];
  if (!srcZoomX || !srcZoomY)
  {
    qDebug("NULL pointer: ApaintDoc::setAspectRatio");
    return false;
  }
  for (int x=0; x<swid; x++)
    srcZoomX[x] = (int)(x/aspectX);
  for (int y=0; y<shei; y++)
    srcZoomY[y] = (int)(y/aspectY);
  
  return(true);
}

void Layer::setAspectRatio(double ax, double ay)
{
  aspectX = ax;
  aspectY = ay;  
  initPixmap(imageBuf.width(), imageBuf.height());
}

void Layer::setNewData(ApaintImage &img, QImage *imask)
{ 
  setNewData(img, pos.x(), pos.y(), imask); 
}

void Layer::setNewData(ApaintImage &img, int x, int y, QImage *imask)
{
  imageBuf = img;
  origBuf = imageBuf.copy();
  updateSize();
  newPos(x, y);
  if (isUsingTiles())
    createTiles(tilex, tiley, tileOffx, tileOffy);
  valid = dataLoaded =  initPixmap(imageBuf.width(), imageBuf.height());
  if (imask)
    setMask(*imask);
  else
    setMask();
}

/* loads data from the layer pos in main layer */
void Layer::loadDataFromLayer(Layer *from, QRect srcPos, bool fillBackGround)
{
 if (getType() == MainLayer) return;
 
 if (from->isUsingTiles())
  imageBuf = from->getAllTileData(srcPos, true);
 else
  imageBuf=from->imageBuf.copy(srcPos.x(), srcPos.y(), srcPos.width(), srcPos.height());

 if (fillBackGround)
  {
   int color = imageBuf.type() == LOAD_TRUE_COL ? apaint->getBgColor() : bgNum;

   int x = srcPos.x();
   int y = srcPos.y();
   int w = srcPos.width();
   int h = srcPos.height();
   int repx = 1;
   int repy = 1;

   if (from->isUsingTiles())
    {
     repx = from->tilex;
     repy = from->tiley;
    }

   for (int y1 = y; y1<0 && y1<y+h; y1++)
    for (int x1 = x; x1<x+w; x1++)
     imageBuf.setPixel(x1-x, y1-y, color);

   for (int y1 = QMAX(0,y); y1<y+h; y1++)
    for (int x1 = x; x1<0 && x1<x+w; x1++)
     imageBuf.setPixel(x1-x, y1-y, color);

   for (int y1 = QMAX(0,y); y1<y+h; y1++)
    for (int x1 = from->imageBuf.width()*repx; x1<x+w; x1++)
     imageBuf.setPixel(x1-x, y1-y, color);

   for (int y1 = from->imageBuf.height()*repy; y1<y+h; y1++)
    for (int x1 = x; x1<x+w; x1++)
     imageBuf.setPixel(x1-x, y1-y, color);
  }

 if (from->isUsingTiles())
  origBuf = imageBuf.copy();
 else
  origBuf = from->imageBuf.copy(srcPos.x(), srcPos.y(), srcPos.width(), srcPos.height());

 valid = dataLoaded = initPixmap(srcPos.width(), srcPos.height());
 pos = srcPos;
 updateSize();
 setMask();
 setLayerName();
}

QString Layer::getBaseName(QString name)
{
 QStringList sl = QStringList::split('/', name);
 if (!sl.count())
  return (ApaintConfig::getText(10));
 return (sl[sl.count()-1]);
}

void Layer::resetColors()
{
 for (int i=0; i<256; i++)
 {
   protectedColors[i] = false;
   transColors[i] = false;
 } 
}

void Layer::setupTileTable(int oldx, int oldy, int newx, int newy)
{
  int x, y;
  Tile *tile;
  /*first remove old ones*/
  if (oldy>newy)
  {
    for (y=newy; y<oldy; y++)
     for (x = 0; x<oldx; x++)
     tileList.removeLast();
    oldy = newy;
  }
  if (oldx>newx)
  {
    for (y=oldy-1; y>=0; y--)
     for (x=oldx-1; x>=newx; x--)
      tileList.remove(y*oldx+x);
  }
  if (newy>oldy)
  {
    for (y=oldy; y<newy; y++)
     for (x=0; x<oldx; x++)
     {
       tile = new Tile(QRect(), QRect());
       tileList.append(tile);
     }
  }
  if (newx>oldx)
  {
    for (y=0; y<newy; y++)
     for (x=oldx; x<newx; x++)
     {
      tile = new Tile(QRect(), QRect());
      tileList.insert(y*newx+x, tile);
     }
  }
}

void Layer::createTiles(int sizex, int sizey, int offx, int offy)
{
  int x, y, ox, oy;
  QRect posl, posr, layerSize;
  Tile *tile;
  int oldtx = tilex;
  int oldty = tiley;
  tilex = sizex;
  tiley = sizey;
  tileOffx = offx;
  tileOffy = offy;
  bool horOff = false;  
  
  setupTileTable(oldtx, oldty, tilex, tiley);
  if (tilex == 1 && tiley == 1)
  {
    enableTiles(false);
    return;
  }
  if (offx !=0 )
  {
   offy = 0;
   horOff = true;
  }
  else if (offy!=0)
  {
    offx = 0;
    horOff = false;
  }  
  
  layerSize = QRect(pos.x(), pos.y(), imageBuf.width()*tilex, imageBuf.height()*tiley);
  for (y = 0; y<tiley; y++)
  {
   for (x=0; x<tilex; x++)
   {
     ox = (offx*y)%imageBuf.width() + x*imageBuf.width();
     oy = (offy*x)%imageBuf.height() + y*imageBuf.height();

     if ((offx!=0) && (ox+imageBuf.width()>layerSize.width()))
     {
        posl = QRect(ox, oy, layerSize.width()-ox, imageBuf.height());
        posr = QRect(0, oy, ox+imageBuf.width()-layerSize.width(), imageBuf.height());
     }
     else if ((offy!=0) && (oy+imageBuf.height()>layerSize.height()))
     {
        posl = QRect(ox, oy, imageBuf.width(), layerSize.height()-oy);
        posr = QRect(ox, 0, imageBuf.width(), oy+imageBuf.height()-layerSize.height());
     }
     else
     {
        posl = QRect(ox, oy, imageBuf.width(), imageBuf.height());
        posr = QRect();
     }
     
     tile = tileList.at(y*tilex+x);
     tile->setMainPos(posl);
     tile->setSecondPos(posr);
     tile->setSecondHor(horOff);   
   }
  }
  enableTiles(true);
}

void Layer::enableTiles(bool enable)
{  
 if (tilex == 1 && tiley == 1)
   useTiles = false;
 else
   useTiles = enable;

  updateSize();
}

ApaintImage Layer::getTileData(Tile *tdata, QRect r, int sx, int sy, int imgx, int imgy,  bool doMirror)
{
  ApaintImage dImg;  
  int srcx, srcy;

  srcx = r.x()-sx+imgx;
  srcy = r.y()-sy+imgy;
  
  if (tdata->mirrorHor())
   srcx =(origBuf.width()-srcx)-r.width();
  if (tdata->mirrorVert())
   srcy =(origBuf.height()-srcy)-r.height();
  dImg = origBuf.copy(srcx, srcy, r.width(), r.height());
  if (doMirror)
   {
    if (tdata->mirrorHor()) dImg.mirrorHorizontal();
    if (tdata->mirrorVert()) dImg.mirrorVertical();
   }
  return dImg;
}

ApaintImage Layer::getAllTileData(QRect r,  bool doMirror)
{
  ApaintImage tmp;
  ApaintImage retImage;
  Tile *tdata;
  QRect recData, second;
  int sx, sy, imgx, imgy;

  retImage.newImage(r.width(), r.height(), imageBuf.depth());
  imageBuf.copyImageDataTo(&retImage);  
  for (int i=0; i<(int)tileList.count(); i++)
  {
   tdata = tileList.at(i);
   second = tdata->getPosSecond();
   if (tdata->intersectsMain(r))
   {
    recData = tdata->intersectMain(r);
    sx = tdata->getPosMain().x();
    sy = tdata->getPosMain().y();
    tmp = getTileData(tdata, recData, sx, sy, 0, 0, doMirror);
    bitBlt(&retImage, recData.x()-r.x(), recData.y()-r.y(), &tmp, 0, 0, recData.width(), recData.height(),
           true, false, 0, 0, 0);
   }
   if (second.isValid() && tdata->intersectsSecond(r))
   {     
    recData = tdata->intersectSecond(r);
    imgx = imgy = 0;
    if (tdata->getSecondHor())
      imgx = imageBuf.width() - second.width();
    else    
      imgy = imageBuf.height() - second.height();
    tmp = getTileData(tdata, recData, second.x(), second.y(), imgx, imgy, doMirror);
    bitBlt(&retImage, recData.x()-r.x(), recData.y()-r.y(), &tmp, 0, 0, recData.width(), recData.height(),
           true, false, 0, 0, 0);
   }
  }
  return retImage;
}

void Layer::getTileInfo(int &sizex, int &sizey, int &offx, int &offy)
{
  sizex = tilex;
  sizey = tiley;
  offx = tileOffx;
  offy = tileOffy;
}

Tile* Layer::pointInTile(int x, int y)
{
  Tile *tdata;
  
  if (isUsingTiles())
    for (int i=0; i<(int)tileList.count(); i++)
    {
     tdata = tileList.at(i);
     if (tdata->getPosMain().contains(x,y) || tdata->getPosSecond().contains(x,y))
      return tdata;
    }
  return 0;
}

int Layer::getTileNum(int x, int y)
{
  if (isUsingTiles() && layerPos().contains(x, y))
  {
   Tile *tdata;
   for (int i=0; i<(int)tileList.count(); i++)
   {
     tdata = tileList.at(i);
     if (tdata->getPosMain().contains(x,y) || tdata->getPosSecond().contains(x,y))
      return i;
   }
  }
  else if (layerPos().contains(x, y))
    return 0;
    
  return -1; 
}

QPtrList<QRect>& Layer::getLayerRefreshArea(QRect area)
{
  Tile *tile;
  QRect  mainr, extend, secr;
  int sx, sy;
  int imgx, imgy;

  refArea.clear();
  for (int i=0; i<(int)tileList.count(); i++)
  {
    tile = tileList.at(i);
    mainr = tile->getPosMain();
    extend = tile->getPosSecond();
    if (area.intersects(mainr))
    {
      secr = area.intersect(mainr); 
      sx = secr.x()-mainr.x();
      sy = secr.y()-mainr.y();
      if (tile->mirrorHor())
       sx = (imageBuf.width()-sx-secr.width());
      if (tile->mirrorVert())
       sy = (imageBuf.height()-sy-secr.height());
       
      refArea.append(new QRect(sx, sy,secr.width(), secr.height()));
    }
    if (extend.isValid() && area.intersects(extend))
    {
      secr = area.intersect(extend);
      imgx = imgy = 0;
      if (tile->getSecondHor())      
        imgx = imageBuf.width() - extend.width();
      else
        imgy = imageBuf.height() - extend.height();
      sx = secr.x() - extend.x() + imgx;
      sy = secr.y() - extend.y() + imgy;
      if (tile->mirrorHor())
       sx = (imageBuf.width()-sx-secr.width());
      if (tile->mirrorVert())
       sy = (imageBuf.height()-sy-secr.height());

      refArea.append(new QRect(sx, sy,secr.width(), secr.height()));
    }
  }
  return refArea;
}

void Layer::tilesToImage()
{
  if (getType() != MainLayer) return;
  if (isUsingTiles())
  {
    imageBuf = getAllTileData(pos, true);
    origBuf = imageBuf.copy();
    initPixmap(pos.width(), pos.height());
  }  
}

void Layer::copyRepeatOptions(Layer *newLayer)
{
  if (newLayer->getType() != MainLayer) return;
  if (isUsingTiles())
  {
    int sizex, sizey, offx, offy;
    Tile *srcData, *dstData;
    
    getTileInfo(sizex, sizey, offx, offy);
    newLayer->createTiles(sizex, sizey, offx, offy);
    newLayer->setDrawDelimiterLines(doDrawDelimiterLines());
    for (int i=0; i<(int)tileList.count(); i++)
    {
     srcData = tileList.at(i);
     dstData = newLayer->tileList.at(i);
     dstData->setMirrorHor(srcData->mirrorHor());
     dstData->setMirrorVert(srcData->mirrorVert());
    }    
  }
}

bool Layer::isValid()
{
  return valid;
}

QRect Layer::layerRefreshPos()
{
 return QRect(pos.x()-REF_SIZE, pos.y()-REF_SIZE, pos.width()+2*REF_SIZE, pos.height()+2*REF_SIZE);
}

void Layer::newPos(int x, int y)
{
 if (getType() == MainLayer) return; 
 int dx = x - pos.x();
 int dy = y - pos.y();
 move(-dx, -dy);
}

void Layer::move(int dx, int dy)
{
  if (geometry == Freehand)
   freehandBorder.translate(-dx, -dy);
  else if (geometry == FreehandSegment)
  {
    for (int i = 0; i<numBoundSeg;i++)
    {
      ptsBoundary[i].x1-=dx;
      ptsBoundary[i].y1-=dy;
      ptsBoundary[i].x2-=dx;
      ptsBoundary[i].y2-=dy;
    }
  }
  pos.moveBy(-dx, -dy);
  setResizeable(false);
}

void Layer::setMask(QImage &smask)
{
  if (getType() == MainLayer) return;
  mask = smask.copy();
  useMask=true;
}
#include <iostream>

using namespace std;

void Layer::setMask()
{
  QBitmap bmp = QBitmap(pos.width(), pos.height());
  QPainter p(&bmp);
  /*reset first*/
  p.setPen(QPen(Qt::color0, 1));
  p.setBrush(QBrush(Qt::color0, Qt::SolidPattern));
  p.drawRect(0, 0, pos.width(), pos.height());
  /**/
  p.setPen(QPen(Qt::color1, 1));
  p.setBrush(QBrush(Qt::color1, Qt::SolidPattern));
  if (geometry == RectangleG)
    p.drawRect(0, 0, pos.width(), pos.height());
  else if (geometry == CircleG)
   p.drawEllipse(0, 0, pos.width(), pos.height());
  else if (geometry == Freehand)
  {
    p.setPen(QPen(Qt::color1, 0));
    p.translate(-pos.x(), -pos.y());
    p.drawPolygon(freehandBorder);
  }
    
  QImage img = bmp.convertToImage();
  setMask(img);
}

void Layer::setGeometry( LayerGeometry g)
{
  geometry = g;
  if (geometry == FreehandSegment)
    ptsBoundary =  Boundary::generate_boundary(&mask, pos.x(), pos.y(), numBoundSeg);
}

void Layer::setValidDraw(bool isValid)
{
  validDraw = isValid;
}

void Layer::drawTileLine(QPainter *p, int x1, int y1, int x2, int y2, QRect &clip, bool horizontal)
{
  if (horizontal)
  {
    if (x1<clip.left()) x1 = clip.left();
    if (x2>clip.right()) x2 = clip.right();
    p->drawLine(x1, y1, x2, y2);
  }
  else
  {
    if (y1<clip.top()) y1 = clip.top();
    if (y2>clip.bottom()) y2 = clip.bottom();
    p->drawLine(x1, y1, x2, y2);
  }
}

void Layer::drawTileLines(QPainter *p, QRect clip, double zoomX, double zoomY)
{
  if (!isUsingTiles()) return;
  if (!doDrawDelimiterLines()) return;
  int lx=(int)(pos.x()*zoomX);
  int ty=(int)(pos.y()*zoomY);
  int rx=(int)((pos.x()+imageBuf.width())*zoomX);
  int by=(int)((pos.y()+imageBuf.height())*zoomY);
  int tileOffxZoom = (int) (tileOffx*zoomX);
  int tileOffyZoom = (int) (tileOffy*zoomY);
  
  p->setRasterOp(Qt::XorROP);
  p->setPen(QPen(QColor(128, 128,128),QMAX(2, (int)(zoomX/2)),Qt::DashLine));

  if (tileOffy>0)
  {
   for (int x=0; x<tilex; x++)
    for (int y=0; y<tiley; y++)
     drawTileLine(p, rx*x, y*by+(x*tileOffyZoom)%by, rx*(x+1), y*by+(x*tileOffyZoom)%by, clip, true);
  }
  else
  {
   for (int y=1; y<tiley; y++)
    drawTileLine(p, lx, y*by, rx*tilex, y*by, clip, true);
  }

  if (tileOffx>0)
   {
    for (int y=0; y<tiley; y++)
      for (int x=0; x<tilex; x++)
        drawTileLine(p, x*rx+(y*tileOffxZoom)%rx, by*y, x*rx+(y*tileOffxZoom)%rx, by*(y+1), clip, false);
   }
  else
  {
   for (int x=1; x<tilex; x++)
    drawTileLine(p, x*rx, ty, x*rx, by*tiley, clip, false);
  }
}

void Layer::drawBorderMain(QPainter *p, double zoomX, double zoomY, bool activeLayer)
{
 if (!validDraw) return;
 if (!doShowBorder) return;

 p->setRasterOp(Qt::XorROP);

 if (vmode == NormalView)
 {
   if (activeLayer)
     p->setPen(QPen(QColor(0, 128,0),QMAX(2, (int)(zoomX/2)),Qt::DashLine));
   else
     p->setPen(QPen(QColor(128, 128,0),QMAX(2, (int)(zoomX/2)),Qt::DashLine));   

   drawBorder(p, zoomX, zoomY);
 }
 else
  drawBorder(p, pts, zoomX, zoomY);
}

void Layer::drawBorder(QPainter *p, double zoomX, double zoomY)
{ 
  int x1, y1, x2, y2, i, count;
  
  int lx=(int)(pos.x()*zoomX);
  int ty=(int)(pos.y()*zoomY);
  int rx=(int)((pos.x()+pos.width())*zoomX);
  int by=(int)((pos.y()+pos.height())*zoomY); 
  
  if (geometry == RectangleG)
    p->drawRect(lx, ty, rx-lx, by-ty);
  else if (geometry == CircleG)
   p->drawEllipse(lx, ty, rx-lx, by-ty);
  else if (geometry == Freehand)
  {   
    QPoint *pt;
    QPoint tmp;

    tmp = freehandBorder.point(0);
    pt = &tmp;
    count = freehandBorder.count();
    if (!count) return;

    x1 = (int)(pt->x()*zoomX)+(int)(zoomX/2);
    y1 = (int)(pt->y()*zoomY)+(int)(zoomY/2);
    for (i =1; i<count; i++)
    {

      tmp = freehandBorder.point(i);
      pt = &tmp;

      x2 = (int)(pt->x()*zoomX)+(int)(zoomX/2);
      y2 = (int)(pt->y()*zoomY)+(int)(zoomY/2);
      p->drawLine(x1, y1, x2, y2);
      x1 = x2;
      y1 = y2;
    }      
  }
  else if (geometry == FreehandSegment)
  {
    for (i=0; i<numBoundSeg; i++)
    {
      x1 = (int)(ptsBoundary[i].x1*zoomX);//+(int)(zoomX/2);
      y1 = (int)(ptsBoundary[i].y1*zoomY);//+(int)(zoomY/2);
      x2 = (int)(ptsBoundary[i].x2*zoomX);//+(int)(zoomX/2);
      y2 = (int)(ptsBoundary[i].y2*zoomY);//+(int)(zoomY/2);
      p->drawLine(x1, y1, x2, y2);
    }
  }
}

void Layer::drawBorder(QPainter *p, QPointArray dpt, double zoomX, double zoomY)
{
 int j;
 QPoint pt[8];

 if (!dpt.count()) return;
 
 pt[0] = QPoint((int)((dpt.point(0).x())*zoomX), (int )((dpt.point(0).y())*zoomY));
 pt[1] = QPoint((int)((dpt.point(1).x()+1)*zoomX), (int )((dpt.point(1).y())*zoomY));
 pt[2] = QPoint((int)((dpt.point(2).x()+1)*zoomX), (int )((dpt.point(2).y()+1)*zoomY));
 pt[3] = QPoint((int)((dpt.point(3).x())*zoomX), (int )((dpt.point(3).y()+1)*zoomY));
 pt[4] = QPoint((pt[0].x()+pt[1].x())/2, (pt[0].y()+pt[1].y())/2);
 pt[5] = QPoint((pt[1].x()+pt[2].x())/2, (pt[1].y()+pt[2].y())/2);
 pt[6] = QPoint((pt[2].x()+pt[3].x())/2, (pt[2].y()+pt[3].y())/2);
 pt[7] = QPoint((pt[3].x()+pt[0].x())/2, (pt[3].y()+pt[0].y())/2);
 
 p->setRasterOp(Qt::NotROP);
 for (int i=0; i<NUM_POINTS; i++)
 {
   p->drawRect(pt[i].x()-RECT_SIZE, pt[i].y()-RECT_SIZE, 2*RECT_SIZE, 2*RECT_SIZE);

   if (i>=4)
	   continue;

   if (i==3)
    j = 0;
   else
    j = i+1;
  
   p->drawLine(pt[i].x(), pt[i].y(), pt[j].x(), pt[j].y());
 }  
}

void Layer::transformBoundary(QWMatrix &mx)
{
  if (geometry == FreehandSegment && transPtsBoundary)
  {
    for (int i = 0; i<numBoundSeg;i++)
    {
      mx.map( transPtsBoundary[i].x1, transPtsBoundary[i].y1, &ptsBoundary[i].x1, &ptsBoundary[i].y1);
      mx.map( transPtsBoundary[i].x2, transPtsBoundary[i].y2, &ptsBoundary[i].x2, &ptsBoundary[i].y2);
    }
  }
}

void Layer::beginTransform(QPointArray dpts)
{
  transPos=pos;
  saveMask = mask.copy();
  vmode = TransformView;
  tpts = dpts.copy();
  pts = dpts.copy();
  if (geometry == Freehand)
   xBorder = freehandBorder.copy();
  else if (geometry == FreehandSegment)
  {
     transPtsBoundary = new BoundSeg[numBoundSeg];
     if (transPtsBoundary)
       memcpy(transPtsBoundary, ptsBoundary, numBoundSeg*sizeof(BoundSeg));
  }

}
    
void Layer::xForm(QWMatrix &mx, int dstWidIn, int dstHeiIn)
{
  int dstWid,dstHei;

  dstWid =  dstWidIn;
  dstHei =  dstHeiIn;
  
  pts = mx.map( tpts );
  pos = pts.boundingRect().normalize();

  if (usingMask())
    mask = saveMask.copy();
  imageBuf = origBuf.xForm(mx, dstWid, dstHei, usingMask(), mask);
  
  /*main layer must always start at 0,0 and have image dimensions!*/
  if (type == MainLayer)
  {      
    pos = QRect(0, 0, imageBuf.width(), imageBuf.height());
    if (isUsingTiles())
      createTiles(tilex, tiley, tileOffx, tileOffy);
  }
  else
    pos = QRect(pos.x(), pos.y(),imageBuf.width(), imageBuf.height());
    
  freehandBorder = mx*xBorder;
  transformBoundary(mx);
}

void Layer::scaleRotate(QWMatrix &mx, double angle, int dstWidIn, int dstHeiIn,
                        bool resizeContents, uint col, bool rotateContour, bool doAverage, ResizeFilterType ftyp)
{
  int dstWid,dstHei;

  dstWid =  dstWidIn;
  dstHei =  dstHeiIn;
  
  if (!dstWid || !dstHei)
    return;
  
  pts = mx.map( tpts );
  pos = pts.boundingRect().normalize();

  if (!resizeContents)
  {
    imageBuf = origBuf.resize(dstWid, dstHei, col);
    pos = QRect(pos.x(), pos.y(),imageBuf.width(), imageBuf.height());
    if (usingMask())
      setMask(); 
  }
  else
  {
    if (usingMask())
    {
      if (dstWid != origBuf.width() || dstHei != origBuf.height())
        mask = saveMask.scale(dstWid, dstHei);
      else
        mask = saveMask.copy();
    }

    imageBuf = CResizeEngine::resize(origBuf, dstWid, dstHei, ftyp);
    if (angle != 0)
      imageBuf = imageBuf.rotate(angle, ftyp, usingMask(), mask);

    //change image back to cmap if we did filtering
    if (ftyp != FILTER_NONE && origBuf.type() == LOAD_CMAP)
      imageBuf = ApaintImageUtil::changeToColorTabImg(origBuf, imageBuf, angle, dstWid, dstHei, 
                                                      rotateContour, usingMask(), mask);

    if(doAverage)
        ApaintImageUtil::averageContour(imageBuf, 0, 0, imageBuf.width(), imageBuf.height(),
                                        protectedColors, isUsingTiles(), 0, 0, 0);
  }

  /*main layer must always start at 0,0 and have image dimensions!*/
  if (type == MainLayer)
  {
    pos = QRect(0, 0, imageBuf.width(), imageBuf.height());
    if (isUsingTiles())
      createTiles(tilex, tiley, tileOffx, tileOffy);
  }
  else
   pos = QRect(pos.x(), pos.y(),imageBuf.width(), imageBuf.height());  

  freehandBorder = mx*xBorder;
  transformBoundary(mx);
}

void Layer::endTransform()
{
  origBuf = imageBuf.copy();
  /*it has new size*/
  initPixmap(imageBuf.width(), imageBuf.height());
  vmode = NormalView;
  xBorder = QPointArray();
  if (geometry == FreehandSegment && transPtsBoundary)
    delete[] transPtsBoundary;
  
}

void Layer::horShift(int x)
{
  imageBuf.horShift(x);
  origBuf = imageBuf.copy();
  shiftx = x;
}

void Layer::vertShift(int y)
{
  imageBuf.vertShift(y);
  origBuf = imageBuf.copy();
  shifty = y;
}

bool Layer::isDataLoaded()
{
  return dataLoaded;
}

void Layer::setNumColors(int num, bool preview)
{
  if (!isValid()) return;
  int oldNum = origBuf.numColors();
   
  imageBuf.setNumColors(num);
  if (!preview)
  {
    if (num<oldNum)
    {     
      for (int i = num; i<oldNum; i++)
      {
        transColors[i] = false;
        protectedColors[i] = false;
      }
     }
     origBuf = imageBuf.copy();
   }
  previewMode = preview;
}  

void Layer::setType(int imgType, int numColors, int noise, bool dither, bool preview)
{
  if (!isValid()) return;

  imageBuf.setImageType(imgType, 0, numColors, noise, dither);
  if (!preview)
   origBuf = imageBuf.copy();

  previewMode = preview;
}

void Layer::removeUnusedColors()
{
  if (!isValid()) return;

   imageBuf.removeUnusedColors();
   origBuf.removeUnusedColors();
}

void Layer::mergeColors()
{
  if (!isValid()) return;

   imageBuf.mergeColors();
   origBuf.mergeColors();
}

void Layer::setExactColorTable(QRgb *ctab, int num)
{
  imageBuf.setExactColorTable(ctab, num);
  origBuf.setExactColorTable(ctab, num);
}

void Layer::setColorTable(QRgb *ctab, int num, bool dither,  bool preview)
{
  if (imageBuf.type() == LOAD_TRUE_COL)
    imageBuf.setImageType(LOAD_CMAP, ctab, num, 0, dither);
  else if (imageBuf.type() == LOAD_CMAP)      
    imageBuf.setColorTable(ctab, num);

  if (!preview)
   origBuf = imageBuf.copy();

  previewMode = preview;
}

void Layer::finishPreview()
{
  origBuf = imageBuf.copy();
  previewMode = false;
}

void Layer::undoPreview()
{
 imageBuf = origBuf.copy();
 previewMode = false;
}

void Layer::undoDataPreview()
{
  origBuf.copyImageDataTo(&imageBuf);
}

LayerDrag *Layer::getClipboardData()
{
  LayerDrag *ld = new LayerDrag();
  if (!ld)
  {
    qDebug("LayerDrag: NULL pointer!");
    return 0;
  }
  ld->setAImage(&imageBuf);
  if (useMask)
   ld->setMask(mask);   
  return ld;
}

void Layer::setColor(int n, QRgb col)
{
  imageBuf.setColor(n, col);
  origBuf.setColor(n, col);
}

void Layer::getUsedColors(bool *uc, int &num)
{
  imageBuf.getUsedColors(uc, num, false, 0);
}

void Layer::remapData(RemapArray &rmap, bool preview)
{
  imageBuf.remapData(rmap);
  if (!preview)
    origBuf.remapData(rmap);

  previewMode = preview;
}

void Layer::setBorderPoints(QPointArray &pts)
{
  if (geometry != Freehand) return;
  
  freehandBorder = pts.copy();
}

bool Layer::pointVisible(QImage &mask, int x, int y)
{
  QRect pos = QRect(0, 0, mask.width(), mask.height());
  if (!pos.contains(x,y)) return false;

  if (mask.bitOrder() == QImage::LittleEndian)
    return *(mask.scanLine(y)+ (x >> 3)) & 1 << (x&7);
  else
    return *(mask.scanLine(y)+ (x >> 3)) & 1 << (7 - (x&7));
}

bool Layer::pointVisible(int xin, int yin)
{
  if (!pos.contains(xin,yin)) return false;
  
  int x = xin-pos.x();
  int y = yin-pos.y();
  
  if (mask.bitOrder() == QImage::LittleEndian)
    return *(mask.scanLine(y)+ (x >> 3)) & 1 << (x&7);
  else
    return *(mask.scanLine(y)+ (x >> 3)) & 1 << (7 - (x&7));
}

void Layer::applyFilter(int what, int fgCol, int bgCol, int penX, int penY, int value, bool doVarColors, bool doVarSize, bool doGradient)
{
 applyFilter(0, 0, imageBuf.width(), imageBuf.height(), what, fgCol, bgCol, penX, penY, value, doVarColors, doVarSize, doGradient);
}

void Layer::applyFilter(int x, int y, int w, int h, int what, int fgCol, int bgCol, int penX, int penY, int value, bool doVarColors, bool doVarSize, bool doGradient)
{
 ApaintImageUtil::filterApply(imageBuf, x, y, w, h, protectedColors, isUsingTiles(), what, fgCol, bgCol, penX, penY, value, doVarColors, doVarSize, doGradient);
 origBuf = imageBuf.copy();
}

void Layer::contourAverage(int thickX, int thickY)
{
  contourAverage(0, 0, imageBuf.width(), imageBuf.height(), 0, 0, 0, thickX, thickY);
}

void Layer::contourDespecle(int what, int neighbours)
{
  contourDespecle(0, 0, imageBuf.width(), imageBuf.height(), what, neighbours);
}

void Layer::contourAverage(int x, int y, int w, int h, int rad, double aspectX, double aspectY, int thickX, int thickY)
{
 if (imageBuf.type() == LOAD_TRUE_COL)
  ApaintImageUtil::averageContour24thickness(imageBuf, x, y, w, h, isUsingTiles(), thickX, thickY);
 else
  {
   if ((thickX>1) || (thickY>1))
    ApaintImageUtil::averageContourThickness(imageBuf, x, y, w, h, protectedColors, isUsingTiles(), thickX, thickY);
   else
    ApaintImageUtil::averageContour(imageBuf, x, y, w, h, protectedColors, isUsingTiles(), rad, aspectX, aspectY);
  }
 origBuf = imageBuf.copy();
}

void Layer::contourDespecle(int x, int y, int w, int h, int what, int neighbours)
{
 if (imageBuf.type() == LOAD_CMAP)
  {
   ApaintImageUtil::despecleContour(imageBuf, x, y, w, h, protectedColors, isUsingTiles(), what, neighbours);
   origBuf = imageBuf.copy();
  }
}

void Layer::contourThick(uint col, int direction, int timesX, int timesY, int angle, bool doShade, bool doLines,
 bool doGradient, bool doBidirectional, bool repeatOperation)
{
 if ((!timesX) && (!timesY) )
  return;
 for (int i=0; i< (repeatOperation ? max(timesX, timesY) : 1); i++)
  ApaintImageUtil::contourThick(imageBuf, col, direction, protectedColors, timesX, timesY, angle, isUsingTiles(), doShade, doLines, doGradient, doBidirectional, repeatOperation);
 origBuf = imageBuf.copy();
}

void Layer::contourRemove(uint col)
{
  ApaintImageUtil::contourRemove(imageBuf, col, protectedColors, isUsingTiles());
  origBuf = imageBuf.copy();
}

void Layer::contourPrune(uint col)
{
  ApaintImageUtil::contourPrun(imageBuf, col, protectedColors, isUsingTiles());
  origBuf = imageBuf.copy();
}

void Layer::contourThin(uint col)
{
  ApaintImageUtil::contourThin(imageBuf, col, isUsingTiles());
  origBuf = imageBuf.copy();
}

void Layer::contourAdd(uint fgCol, uint bgCol, uint withCol, bool useWithCol, int direction, bool into, bool doShade, int numX, int numY)
{
 if ((!numX) && (!numY))
  return;
 ApaintImageUtil::contourAdd(imageBuf, fgCol, bgCol, withCol, useWithCol, direction, into, protectedColors, isUsingTiles(), doShade, numX, numY);  
 origBuf = imageBuf.copy();
}

void Layer::changeToBW(int treshHold, bool preview)
{
 origBuf.copyImageDataTo(&imageBuf);
 imageBuf.changeColtabToBW(treshHold);
 if (!preview)   
  {
   imageBuf.copyImageDataTo(&origBuf);
   imageBuf.mergeColors();
   origBuf.mergeColors(); 
  }
 previewMode = preview;
}

void Layer::updateLayerData()
{
  if (isUsingTiles())
   createTiles(tilex, tiley, tileOffx, tileOffy);
  else
   {
    pos.setWidth(imageBuf.width());
    pos.setHeight(imageBuf.height());
   }
  if (usingMask())
   setMask();
  initPixmap(imageBuf.width(), imageBuf.height());
}

void Layer::insertRemovePixelsHor(int x, int num, bool doInsert, bool useBgCol, bool useDataBuf, uchar *dataBuf)
{
  if (useBgCol)
    ApaintImageUtil::insertRemovePixelsHor(imageBuf, num, x, bgNum, bgCol, doInsert, useDataBuf, dataBuf);
  else
    ApaintImageUtil::insertRemovePixelsHor(imageBuf, num, x, fgNum, fgCol, doInsert, useDataBuf, dataBuf);

  origBuf = imageBuf.copy();
  updateLayerData();  
}

void Layer::insertRemovePixelsVert(int y, int num, bool doInsert, bool useBgCol, bool useDataBuf, uchar *dataBuf)
{
  if (useBgCol)
    ApaintImageUtil::insertRemovePixelsVert(imageBuf, num, y, bgNum, bgCol, doInsert, useDataBuf, dataBuf);
  else
    ApaintImageUtil::insertRemovePixelsVert(imageBuf, num, y, fgNum, fgCol, doInsert, useDataBuf, dataBuf);
    
  origBuf = imageBuf.copy();
  updateLayerData();
}
  
void Layer::borderMirror(bool vert)
{
  int x, y;

  if (geometry == Freehand)
  {
    for (uint i=0; i<freehandBorder.count(); i++)
    {
     x = freehandBorder.point(i).x();
     if (!vert)
      x = pos.left()+pos.right()-x;
     y = freehandBorder.point(i).y();
     if (vert)
      y = pos.top()+pos.bottom()-y;
     freehandBorder.setPoint(i, x, y);
    }
  }
  else if (geometry == FreehandSegment)
  {
    for (int i=0; i<numBoundSeg; i++)
    {      
       if (!vert)
       {
         ptsBoundary[i].x1 = pos.left()+pos.right()-ptsBoundary[i].x1+1;
         ptsBoundary[i].x2 = pos.left()+pos.right()-ptsBoundary[i].x2+1;
       }

       if (vert)
       {
         ptsBoundary[i].y1 = pos.top()+pos.bottom()-ptsBoundary[i].y1+1;
         ptsBoundary[i].y2 = pos.top()+pos.bottom()-ptsBoundary[i].y2+1;
       }      
    } 
  }
}

void Layer::mirrorHorizontal()
{
  imageBuf.mirrorHorizontal();
  origBuf = imageBuf.copy();
  borderMirror(false);
  if (usingMask())
    mask = mask.mirror(true, false);
}

void Layer::mirrorVertical()
{
  imageBuf.mirrorVertical();
  origBuf = imageBuf.copy();
  borderMirror(true);
  if (usingMask())  
    mask = mask.mirror(false, true);
}

void Layer::mirrorExplode()
{
  imageBuf.mirrorExplode();
  origBuf = imageBuf.copy();
  if (usingMask())
   mask = explodeQImage(&mask);
}

void Layer::flipBorder()
{  
 int x, y;

 if (geometry == Freehand)
  for (uint i=0; i<freehandBorder.count(); i++)
   {
    x = freehandBorder.point(i).y();
    x+= pos.x()-pos.y();
    y = freehandBorder.point(i).x();
    y+=pos.y()-pos.x();
    freehandBorder.setPoint(i, x, y);
   }
 else
  if (geometry == FreehandSegment)
   for (int i=0; i<numBoundSeg; i++)
    {
     x = ptsBoundary[i].y1;
     x+= pos.x()-pos.y();
     y = ptsBoundary[i].x1;
     y+=pos.y()-pos.x();
     ptsBoundary[i].x1 = x;
     ptsBoundary[i].y1 = y;

     x = ptsBoundary[i].y2;
     x+= pos.x()-pos.y();
     y = ptsBoundary[i].x2;
     y+=pos.y()-pos.x();
     ptsBoundary[i].x2 = x;
     ptsBoundary[i].y2 = y;  
    }
}

void Layer::mirrorFlip()
{
 imageBuf.mirrorFlip(FALSE);
 origBuf = imageBuf.copy();
 /*change wid & hei in selection*/
 if (isUsingTiles())
  createTiles(tilex, tiley, tileOffx, tileOffy);
 else
  {
   pos.setWidth(imageBuf.width());
   pos.setHeight(imageBuf.height());
  }
 flipBorder();
 if (usingMask())
  {
   mask = flipQImage(&mask);
   borderMirror(false);
   mask = mask.mirror(true, false);
  }

 initPixmap(imageBuf.width(), imageBuf.height());
}

void Layer::duplicate(int duplicateType)
{
 imageBuf.duplicate(duplicateType);
 origBuf = imageBuf.copy();
 /* changes wid & hei in selection */
 if (isUsingTiles())
  createTiles(tilex, tiley, tileOffx, tileOffy);
 else
  {
   pos.setWidth(imageBuf.width());
   pos.setHeight(imageBuf.height());
  }
 if (usingMask())
  mask = duplicateQImage(duplicateType, &mask);
 initPixmap(imageBuf.width(), imageBuf.height());
}

void Layer::scale2X()
{
 imageBuf.scale2X();
 origBuf = imageBuf.copy();
 /* changes wid & hei in selection */
 if (isUsingTiles())
  createTiles(tilex, tiley, tileOffx, tileOffy);
 else
  {
   pos.setWidth(imageBuf.width());
   pos.setHeight(imageBuf.height());
  }
 if (usingMask())
  mask = scale2XQImage(&mask);
 initPixmap(imageBuf.width(), imageBuf.height());
}

void Layer::scale3X()
{
 imageBuf.scale3X();
 origBuf = imageBuf.copy();
 /* changes wid & hei in selection */
 if (isUsingTiles())
  createTiles(tilex, tiley, tileOffx, tileOffy);
 else
  {
   pos.setWidth(imageBuf.width());
   pos.setHeight(imageBuf.height());
  }
 if (usingMask())
  mask = scale3XQImage(&mask);
 initPixmap(imageBuf.width(), imageBuf.height());
}

void Layer::shrink2X()
{
 imageBuf.shrink2X();
 origBuf = imageBuf.copy();
 /* changes wid & hei in selection */
 if (isUsingTiles())
  createTiles(tilex, tiley, tileOffx, tileOffy);
 else
  {
   pos.setWidth(imageBuf.width());
   pos.setHeight(imageBuf.height());
  }
 if (usingMask())
  mask = shrink2XQImage(&mask);
 initPixmap(imageBuf.width(), imageBuf.height());
}

void Layer::shrink3X()
{
 imageBuf.shrink3X();
 origBuf = imageBuf.copy();
 /* changes wid & hei in selection */
 if (isUsingTiles())
  createTiles(tilex, tiley, tileOffx, tileOffy);
 else
  {
   pos.setWidth(imageBuf.width());
   pos.setHeight(imageBuf.height());
  }
 if (usingMask())
  mask = shrink3XQImage(&mask);
 initPixmap(imageBuf.width(), imageBuf.height());
}

void Layer::shrink4X()
{
 imageBuf.shrink4X();
 origBuf = imageBuf.copy();
 /* changes wid & hei in selection */
 if (isUsingTiles())
  createTiles(tilex, tiley, tileOffx, tileOffy);
 else
  {
   pos.setWidth(imageBuf.width());
   pos.setHeight(imageBuf.height());
  }
 if (usingMask())
  mask = shrink4XQImage(&mask);
 initPixmap(imageBuf.width(), imageBuf.height());
}

void Layer::halve(int halveType)
{
 imageBuf.halve(halveType);
 origBuf = imageBuf.copy();
 /* changes wid & hei in selection */
 if (isUsingTiles())
  createTiles(tilex, tiley, tileOffx, tileOffy);
 else
  {
   pos.setWidth(imageBuf.width());
   pos.setHeight(imageBuf.height());
  }
 if (usingMask())
  mask = halveQImage(halveType, &mask);
 initPixmap(imageBuf.width(), imageBuf.height());
}

void Layer::shrink(int shrinkType)
{
 imageBuf.shrink(shrinkType);
 origBuf = imageBuf.copy();
 /* changes wid & hei in selection */
 if (isUsingTiles())
  createTiles(tilex, tiley, tileOffx, tileOffy);
 else
  {
   pos.setWidth(imageBuf.width());
   pos.setHeight(imageBuf.height());
  }
 if (usingMask())
  mask = shrinkQImage(shrinkType, &mask);
 initPixmap(imageBuf.width(), imageBuf.height());
}

void Layer::multiplyX(int times)
{
 imageBuf.multiplyX(times);
 origBuf = imageBuf.copy();
 /* changes wid & hei in selection */
 if (isUsingTiles())
  createTiles(tilex, tiley, tileOffx, tileOffy);
 else
  {
   pos.setWidth(imageBuf.width());
   pos.setHeight(imageBuf.height());
  }
 if (usingMask())
  mask = multiplyXQImage(times, &mask);
 initPixmap(imageBuf.width(), imageBuf.height());
}

void Layer::multiplyY(int times)
{
 imageBuf.multiplyY(times);
 origBuf = imageBuf.copy();
 /* changes wid & hei in selection */
 if (isUsingTiles())
  createTiles(tilex, tiley, tileOffx, tileOffy);
 else
  {
   pos.setWidth(imageBuf.width());
   pos.setHeight(imageBuf.height());
  }
 if (usingMask())
  mask = multiplyYQImage(times, &mask);
 initPixmap(imageBuf.width(), imageBuf.height());
}

void Layer::sortColors(int sortOrder)
{
 static int sortUp = TRUE;
 static int lastSortOrder = -1;

 if (!isValid()) return;

 if (lastSortOrder!=sortOrder)
  sortUp=TRUE;
 lastSortOrder = sortOrder;

 imageBuf.sortColors(sortOrder, getFgIndx(), sortUp);
 origBuf.sortColors(sortOrder, getFgIndx(), sortUp);
 sortUp = !sortUp;
}

void Layer::invertColors(int what, bool *protColors)
{
 if (!isValid()) return;
 imageBuf.invertColors(what, protColors);
 origBuf.invertColors(what, protColors);
}

void Layer::swapColors(int first, int second)
{
 if (!isValid()) return;

 imageBuf.swapColors(first, second);
 origBuf.swapColors(first, second);
}

void Layer::contrastStretch()
{
 if (!isValid()) return;
 imageBuf.contrastStretchAP();
 origBuf.contrastStretchAP();
}

void Layer::makeColorShades(int fgIndex, int bgIndex)
{
 if (!isValid()) return;
 imageBuf.makeColorShades(fgIndex, bgIndex);
 origBuf.makeColorShades(fgIndex, bgIndex);
}

void Layer::perspectiveXForm(Layer *layer, int dstx, int dsty, int dstw, int dsth, double mx[3][3])
{
 loadDataFromLayer(layer, QRect(dstx, dsty, dstw, dsth));
 imageBuf = imageBuf.perspectiveXForm(dstw, dsth, mx, usingMask(), mask);
 origBuf = imageBuf.copy();
 /*change wid & hei in selection*/
 if (isUsingTiles())
  createTiles(tilex, tiley, tileOffx, tileOffy);
 else
  {
   pos.setWidth(imageBuf.width());
   pos.setHeight(imageBuf.height());
  }
 initPixmap(imageBuf.width(), imageBuf.height());
 if (geometry == Freehand)  
  freehandBorder = QPointArray();
 else  if (geometry == FreehandSegment && transPtsBoundary)
   delete[] transPtsBoundary;

 geometry = RectangleG;
}


void Layer::setMaskWithColor(uint col)
{
  int x, y;
  uint d;  
  uchar **jtab;
  QRgb *ctab=0;
  
  mask = QImage(imageBuf.width(), imageBuf.height(), 1, 2, QImage::LittleEndian);
  mask.fill(1);
  jtab = imageBuf.jumpTable();
  if (imageBuf.type() == LOAD_CMAP)
   ctab = imageBuf.colorTable();
  for (y=0; y<imageBuf.height(); y++)
    if (imageBuf.type() == LOAD_CMAP)
    {
      for (x=0; x<imageBuf.width(); x++)
      {
        d= *(jtab[y]+x);
        if (ctab[d] == col)
          mask.setPixel(x, y, 0);
      }
    }
    else if (imageBuf.type() == LOAD_TRUE_COL)
    {
      for (x=0; x<imageBuf.width(); x++)
      {
        d=qRgb(*(jtab[y]+x*3), *(jtab[y]+x*3+1), *(jtab[y]+x*3+2));
        if (d == col)
          mask.setPixel(x, y, 0);
      }
    }
   useMask = true;
}

bool Layer::isBorderColor(uint col)
{
  int x, y;
  int  numPix, allPix;
  uint d;
  double ratio;
  uchar **jtab;
  QRgb *ctab=0;

  jtab = imageBuf.jumpTable();
  if (imageBuf.type() == LOAD_CMAP)
   ctab = imageBuf.colorTable();
  numPix = allPix = 0;
  
  for (y=0; y<imageBuf.height(); y++)
    if (imageBuf.type() == LOAD_CMAP)
    {
      if (y == 0 || y == imageBuf.height()-1)
      {
        for (x=0; x<imageBuf.width(); x++)
        {
          d= *(jtab[y]+x);
          if (ctab[d] == col)
            numPix++;
          allPix++;
        }
      }
      else
      {
          d= *(jtab[y]);
          if (ctab[d] == col) numPix++;
          d= *(jtab[y]+imageBuf.width()-1);
          if (ctab[d] == col) numPix++;
          allPix+=2;
      }
    }
    else if (imageBuf.type() == LOAD_TRUE_COL)
    {
      if (y == 0 || y == imageBuf.height()-1)
      {
        for (x=0; x<imageBuf.width(); x++)
        {
          d=qRgb(*(jtab[y]+x*3), *(jtab[y]+x*3+1), *(jtab[y]+x*3+2));
          if (d == col)
            numPix++;
          allPix++;
        }
      }
      else
      {
        d=qRgb(*(jtab[y]), *(jtab[y]+1), *(jtab[y]+2));
        if (d == col) numPix++;
        d=qRgb(*(jtab[y]+(imageBuf.width()-1)*3),
               *(jtab[y]+(imageBuf.width()-1)*3+1),
               *(jtab[y]+(imageBuf.width()-1)*3+2));
        if (d == col) numPix++;
        allPix+=2;
      }
    }
   ratio = ((double)numPix / (double) allPix)*100;
   return ((int)ratio>=50);
}

void Layer::setTransperantBg()
{
   uint borderCol;
  /*get background color*/
   borderCol = imageBuf.pixel(0, 0);

  /*if we have background color */
  if (isBorderColor(borderCol))
  {
    /*make it transparent...*/
    if (imageBuf.type() == LOAD_CMAP)
    {
       setTransColor(imageBuf.pixelIndex(0, 0), true);
       setMask();
    }
   /*mask the background*/
/* do not mask background on true color, since we do not have any way of disabling it afterwards 
    else if (imageBuf.type() == LOAD_TRUE_COL)
      setMaskWithColor(borderCol);
*/
  }
  else
    setMask();
}

QString Layer::getLayerName()
{
  return name;
}

void Layer::setLayerName()
{
 if (type == ImageLayer)
  name = getBaseName(imageBuf.imageName());
 else if (type == MainLayer)
  name = ApaintConfig::getText(648);
 else if (type == SelectionLayer)
  name = ApaintConfig::getText(646); 
}

void Layer::setLayerType(LayerType t)
{
  type = t;
  setLayerName();
}

void Layer::setLayerName(QString lname)
{
  name = lname;
}

void Layer::fill(int xr, int yr, int w, int h, uint col, bool useMask, QImage *mask)
{
 imageBuf.fill(xr, yr, w, h, col, useMask, mask);
 origBuf.fill(xr, yr, w, h, col, useMask, mask);
}

void Layer::setTransColors(bool *transTab)
{
 if (!transTab) return;
 if (type != MainLayer)
   memcpy(getTransColors(), transTab, 256*sizeof(bool));
}

void Layer::setProtectedColors(bool *protectedTab)
{
 if (!protectedTab) return;
   memcpy(getProtectColors(), protectedTab, 256*sizeof(bool));
}

void Layer::setExactNumColors(int num)
{
  imageBuf.setExactNumColors(num);
  origBuf.setExactNumColors(num);
}

ApaintDoc::ApaintDoc(QWidget *pparent)
{
  modified = false;
  aspectX = aspectY = 1;
  protectMode = false;
  mainLayer = 0;
  selectedLayer=0;
  nLayers=0;
  parent=pparent;
  useClip=false;
  layerIsTransforming = false;
  for (int i=0; i<MAX_LAYERS; i++)
   layerList[i] = 0;
  fileDlg = 0;
  connect(this, SIGNAL(newDocSig()), this, SLOT(redrawLayers()));
  connect(this, SIGNAL(newLayerSig(Layer*)), apaint, SLOT(layerOp()));
  connect(this, SIGNAL(removeLayer(int)), apaint, SLOT(layerOp()));
  connect(this, SIGNAL(layerSelected(int)), apaint, SLOT(layerOp()));
}

ApaintDoc::~ApaintDoc()
{
}

void ApaintDoc::setAspectRatio(double ax, double ay)
{
 aspectX = ax;
 aspectY = ay;
 for (int i=0; i<numLayers(); i++)
  layerList[i]->setAspectRatio(ax, ay);
}

void ApaintDoc::createNewLayer(int x, int y, int wid, int hei, LayerType ltyp, LayerGeometry lg, bool fillBackground)
{
 if (nLayers<MAX_LAYERS)
  {
   selectedLayer=nLayers;
   layerList[selectedLayer]= new Layer(ltyp, lg, aspectX, aspectY);
   if (!layerList[selectedLayer])
    {
     qDebug("ApaintDoc: NULL pointer!") ;
     return;
    }
   nLayers++;
   layerList[selectedLayer]->loadDataFromLayer(mainLayer, QRect(x, y, wid, hei), fillBackground);
   layerList[selectedLayer]->setResizeable(true);
   emit newLayerSig(layerList[selectedLayer]);
   emit layerChanged(layerList[selectedLayer]); 
  }
}

void ApaintDoc::createNewLayer(QPointArray &borderPts, LayerType ltyp, LayerGeometry lg)
{
  if (nLayers<MAX_LAYERS)
  {
   selectedLayer=nLayers;
   layerList[selectedLayer]= new Layer(borderPts, ltyp, lg, aspectX, aspectY);
   if (!layerList[selectedLayer])
   {
    qDebug("ApaintDoc: NULL pointer!") ;
    return;
   }
   nLayers++;
   layerList[selectedLayer]->loadDataFromLayer(mainLayer, (borderPts.boundingRect()).normalize());
   layerList[selectedLayer]->setResizeable(true);
   emit newLayerSig(layerList[selectedLayer]);
   emit layerChanged(layerList[selectedLayer]);  
  }
}

void ApaintDoc::createNewLayer(ApaintImage &img, QImage *imask, int x, int y)
{
 if (nLayers<MAX_LAYERS)
  {
   selectedLayer=nLayers;
   layerList[selectedLayer]= new Layer(img, x, y, imask, SelectionLayer, aspectX, aspectY);   
   if (!layerList[selectedLayer])
   {
    qDebug("ApaintDoc: NULL pointer!") ;
    return;
   }
   nLayers++;
   emit newLayerSig(layerList[selectedLayer]);
   emit layerChanged(layerList[selectedLayer]);    
  }  
}

void ApaintDoc::createNewLayer(QImage *imask, QRect area)
{
  createNewLayer(mainLayer, imask, area, false);
}

void ApaintDoc::createNewLayer(Layer *srcLayer, QImage *imask, QRect area, bool useSrcPos)
{
  int px, py;
  if (!srcLayer) return;
  if (nLayers<MAX_LAYERS && nLayers>0)
  {
   selectedLayer=nLayers;
   ApaintImage img = srcLayer->imageBuf.copy(area.x(), area.y(), area.width(), area.height());
   srcLayer->imageBuf.copyImageDataTo(&img);
   if (useSrcPos)
   {
     px = srcLayer->layerPos().x();
     py = srcLayer->layerPos().y();
   }
   else
   {
     px = area.x();
     py = area.y();     
   }
   layerList[selectedLayer]= new Layer(img, px, py, imask, SelectionLayer, aspectX, aspectY);
   if (!layerList[selectedLayer])
   {
    qDebug("ApaintDoc: NULL pointer!") ;
    return;
   }
   layerList[selectedLayer]->setGeometry(FreehandSegment);
   nLayers++;
   layerList[selectedLayer]->setResizeable(true);
   emit newLayerSig(layerList[selectedLayer]);
   emit layerChanged(layerList[selectedLayer]);
  }
}


void ApaintDoc::newDoc(int wid, int hei, int depth, uint fillCol, int numCol)
{
  QApplication::setOverrideCursor( waitCursor );
  freeLayers();  
  mainLayer = new Layer(wid, hei, depth, fillCol, numCol, MainLayer, aspectX, aspectY);
  if (!mainLayer)
  {
    qDebug("ApaintDoc: NULL pointer!");
    return;
  }
  layerList[0]=mainLayer;
  selectedLayer=0;
  nLayers++;
  layerList[selectedLayer]->setFilename("");
  emit newLayerSig(layerList[selectedLayer]);
  emit layerChanged(layerList[selectedLayer]);
  emit newDocSig();
  QApplication::restoreOverrideCursor();
}

bool ApaintDoc::loadLayer(const QString &filename)
{
  int prevLayer;
 
  if (nLayers>=MAX_LAYERS || !nLayers) return false;
  QApplication::setOverrideCursor( waitCursor );

  prevLayer = selectedLayer;
  selectedLayer = nLayers;
  layerList[selectedLayer] = new Layer(filename, ImageLayer, aspectX, aspectY);
  if (!layerList[selectedLayer])
  {
    qDebug("ApaintDoc: NULL pointer!");
    selectedLayer = prevLayer;
    return false;
  }
  if (!layerList[selectedLayer]->isValid())
  {
    QApplication::restoreOverrideCursor();
    QMessageBox::warning(0,ApaintConfig::getText(9), QString("%1 : %2").arg(ApaintConfig::getText(512)).arg(filename), ApaintConfig::getText(3));
    delete layerList[selectedLayer];
    selectedLayer = prevLayer;
    return false;
  }

  layerList[selectedLayer]->setTransperantBg();

  layerList[selectedLayer]->setFilename(filename);
  QApplication::restoreOverrideCursor();  
  nLayers++;
  emit newLayerSig(layerList[selectedLayer]);
  emit layerChanged(layerList[selectedLayer]);
  return true;
}

void ApaintDoc::setLayer(int pos, Layer *layer)
{
  if (pos<0 || pos >= numLayers())
   return;
  if (layerList[pos] && layerList[pos]->isUsingTiles() && layer->getType() == MainLayer)
  {
    //set repeat opt.
    layerList[pos]->copyRepeatOptions(layer);
    //set fg, bg colors
    layer->setBgCol(layerList[pos]->getBgCol());
    layer->setFgCol(layerList[pos]->getFgCol());
    layer->setBgIndx(layerList[pos]->getBgIndx());
    layer->setFgIndx(layerList[pos]->getFgIndx());
  }
  delete layerList[pos];
  layerList[pos] = layer;
  if (pos == 0)
   mainLayer = layerList[pos];
}

bool ApaintDoc::load(const QString &filename)
{
  Layer *mlayer;

  QApplication::setOverrideCursor( waitCursor );
  
  mlayer = new Layer(filename, MainLayer, aspectX, aspectY);
  if (!mlayer)
  {
    qDebug("ApaintDoc: NULL pointer!");
    return false;
  }
  if (!mlayer->isValid())
  {
    delete mlayer;
    QApplication::restoreOverrideCursor();
    QMessageBox::warning(0,ApaintConfig::getText(9), QString("%1 : %2").arg(ApaintConfig::getText(512)).arg(filename), ApaintConfig::getText(3));
    return false;
  }
  freeLayers();
  mainLayer = mlayer;
  layerList[0]=mainLayer;
  selectedLayer=0;
  layerList[selectedLayer]->setFilename(filename);
  nLayers++;
  emit newLayerSig(layerList[selectedLayer]);
  emit layerChanged(layerList[selectedLayer]);
  emit newDocSig();
  QApplication::restoreOverrideCursor();
  return true;
}

bool ApaintDoc::fixFilename(QString &fname, int imgType)
{
  bool saveAs;
   
  saveAs = fixPfx(fname);
  QString pfx = getPfx((char*)fname.ascii());
  if ((pfx == "jpg" || pfx =="jpeg") && (imgType == LOAD_CMAP))
   {
    chgPfx(fname, "png");
    saveAs = true;
   }
  return saveAs;
}

bool ApaintDoc::save(bool saveAs, QString &saveFileName)
{
  QString tmp, fileName = saveFileName;
  QDir dir;

  if (!mainLayer) return false;
  
  saveFileName = "";
  fileName = mainLayer->getFilename();
  if (fileName.isEmpty())
   {
    tmp = QFileDialog::getSaveFileName(dirName, 0, parent);
    if (tmp.isEmpty())
     return false;
    else
     fileName = tmp;
    saveAs = true;
   }
  else if (saveAs)
   {
    tmp = QFileDialog::getSaveFileName(fileName, 0, parent);
    if (tmp.isEmpty())
     return false;
    else
     fileName = tmp;
   }
  // fix filename
  if (fixFilename(fileName, mainLayer->getImageType()))
   saveAs = true;

  if (dir.exists(fileName) && saveAs)
   {
    int res = QMessageBox::warning(0,ApaintConfig::getText(9), ApaintConfig::getText(505), ApaintConfig::getText(3), ApaintConfig::getText(5));
    if (res == 1)
      return false;
   }
  collapsLayers();
  if (!mainLayer->imageBuf.save(fileName))
   {
    QMessageBox::warning(0,ApaintConfig::getText(9), ApaintConfig::getText(509), ApaintConfig::getText(3));
    return false;
   }
  if (saveAs)
   {
    mainLayer->origBuf = mainLayer->imageBuf.copy();
    emit layerChanged(layerList[selectedLayer]);
   }
  mainLayer->setFilename(fileName);
  saveFileName = fileName;

  /* here we saving the history file */
  {
   char imageHistoryFile[500];

   strcpy(imageHistoryFile, getImageIconPath());
   strcat(imageHistoryFile, "/history");
   saveRecentFile((char *) fileName.local8Bit().data(), FALSE, imageHistoryFile, ApaintConfig::getNumRecentFiles());
  }
 return true;
}

void ApaintDoc::saveLayer()
{
  QString saveFile;
  ApaintImage saveImage;
  QDir dir;
  
  if (!layerList[selectedLayer])
  {
   QMessageBox::warning(0,ApaintConfig::getText(9), ApaintConfig::getText(509), ApaintConfig::getText(3));  
   return;
  }

  saveFile = QFileDialog::getSaveFileName(dirName, 0, parent);
  if (saveFile.isEmpty()) return ;
  // fix filename
  fixFilename(saveFile, layerList[selectedLayer]->getImageType());
  
  if (dir.exists(saveFile))
   {
    int res = QMessageBox::warning(0,ApaintConfig::getText(9), ApaintConfig::getText(505), ApaintConfig::getText(3), ApaintConfig::getText(5));
    if (res == 1)
      return ;
   }
  saveImage = layerList[selectedLayer]->origBuf.copy();
  if (saveImage.type() == LOAD_CMAP)
  {
   int ncol = saveImage.numColors();
   
   if (ncol<255)
    {
     saveImage.setNumColors(ncol+1);
     saveImage.fill(ncol+1);     
    }
  }
  else  
   saveImage.fill(layerList[selectedLayer]->getBgCol());
  bitBlt(&saveImage, 0, 0, &layerList[selectedLayer]->origBuf, 0, 0, layerList[selectedLayer]->origBuf.width(),
         layerList[selectedLayer]->origBuf.height(), false ,
         layerList[selectedLayer]->usingMask(), &layerList[selectedLayer]->getMask(),
         layerList[selectedLayer]->getProtectColors(), layerList[selectedLayer]->getTransColors());
   
  if (!saveImage.save(saveFile))
    QMessageBox::warning(0,ApaintConfig::getText(9), ApaintConfig::getText(509), ApaintConfig::getText(3));
  else
    layerList[selectedLayer]->setFilename(saveFile);
}

bool ApaintDoc::newLayer(int wid, int hei, int depth, uint fillCol, int numCol)
{
 int prevLayer;

  if (nLayers>=MAX_LAYERS || !nLayers) return false;
  QApplication::setOverrideCursor( waitCursor );

  prevLayer = selectedLayer;
  selectedLayer = nLayers;  
  layerList[selectedLayer] = new Layer(wid, hei, depth, fillCol, numCol, ImageLayer, aspectX, aspectY);
  if (!layerList[selectedLayer])
   {
    qDebug("ApaintDoc: NULL pointer!");
    selectedLayer = prevLayer;
    return false;
   }
  if (!layerList[selectedLayer]->isValid())
   {
    QApplication::restoreOverrideCursor();    
    delete layerList[selectedLayer];
    selectedLayer = prevLayer;
    return false;
   }
  layerList[selectedLayer]->setFilename("");
  layerList[selectedLayer]->setMask();
  QApplication::restoreOverrideCursor();
  nLayers++;
  emit newLayerSig(layerList[selectedLayer]);
  emit layerChanged(layerList[selectedLayer]);
  refreshLayer();
  return true; 
}


void ApaintDoc::resizeLayer(int x, int y, int wid, int hei)
{
 if (!layerList[selectedLayer]) return;
 if (layerList[selectedLayer]->getType() == MainLayer) return;
 if (!layerList[selectedLayer]->getIsResizeable()) return;

 layerList[selectedLayer]->loadDataFromLayer(mainLayer, QRect(x, y, wid, hei));
 emit layerChanged(layerList[selectedLayer]);
}

bool ApaintDoc::isLayerValid(Layer *layer)
{
  if (!layer) return false;

  for (int i=0; i<numLayers(); i++)
   if (layerList[i] == layer)
    return true;
   
  return false;
}

void ApaintDoc::deleteLayer(int layer)
{
 if (layer==0) return; /*dont delete mainLayer*/
 if (!layerList[layer]) return;

 delete layerList[layer];
 layerList[layer] = 0;
 for (int i=layer; i<numLayers()-1; i++)
  layerList[i] = layerList[i+1];
 nLayers--;
 emit removeLayer(layer);
 if (selectedLayer == layer)
 {
  selectedLayer--;
  emit layerChanged(layerList[selectedLayer]);
  redrawLayers();
 }
}

void ApaintDoc::copyImageToLayer(Layer *layer, QRect area, ApaintImage& img, ApaintImage& origImg,
                                  bool sameImage, bool useMask, QImage *mask, bool *transCol, bool cmapAdd)
{
  QRect secr;
  int sx, sy, imgx, imgy, j;
  Tile *tile;
  int i, layerColors=1, groundColors=1, doCmapAdd;

  if ((layer->imageBuf.type()==LOAD_TRUE_COL) && (img.type()==LOAD_CMAP))
   layer->setType(LOAD_CMAP, 256, 0, false, false);

  if (cmapAdd && (layer->imageBuf.type()==LOAD_CMAP))
   layer->imageBuf.removeUnusedColors();
  if (cmapAdd && (img.type()==LOAD_CMAP))
   img.removeUnusedColors();

  doCmapAdd = (cmapAdd && (layer->imageBuf.type()==LOAD_CMAP) && (img.type()==LOAD_CMAP) && ((layer->imageBuf.numColors()+1)*img.numColors()<=255));
  if (doCmapAdd)
   {       
    layerColors = img.numColors();
    groundColors = layer->imageBuf.numColors();
    /* enlarge the colormap with all color combinations of ground and layer colors, and make the superimpose color values */
    layer->imageBuf.setNumColors(groundColors + groundColors*layerColors);
    layer->origBuf.setNumColors(groundColors + groundColors*layerColors);
    for (i=0; i<groundColors*layerColors; i++)
     {
      QRgb col1, col2;

      col1 = layer->imageBuf.color(i%groundColors);
      col2 = img.color(i/groundColors);
      layer->imageBuf.setColor(groundColors+i, qRgb(((int)qRed(col1)+qRed(col2))/2, ((int)qGreen(col1)+qGreen(col2))/2, ((int)qBlue(col1)+qBlue(col2))/2));
      layer->origBuf.setColor(groundColors+i, qRgb(((int)qRed(col1)+qRed(col2))/2, ((int)qGreen(col1)+qGreen(col2))/2, ((int)qBlue(col1)+qBlue(col2))/2));
     }
   }

  if (layer->isUsingTiles())
    {
      QPtrList<Tile>& tList = layer->getTileList();
      for (j=0; j<(int)tList.count(); j++)
       {
         tile = tList.at(j);
         QRect mainr = tile->getPosMain();
         QRect extend = tile->getPosSecond();
         if (area.intersects(mainr))
         {
           secr = area.intersect(mainr);
           sx = secr.x()-mainr.x();
           sy = secr.y()-mainr.y();
           if (tile->mirrorHor())
            sx = (layer->imageBuf.width()-sx-secr.width());
           if (tile->mirrorVert())
            sy = (layer->imageBuf.height()-sy-secr.height());

	   if (cmapAdd)
	    {
             if (!img.isNull())
               bitBltAddCmap(&layer->imageBuf, sx, sy, &img,
                      secr.x()-area.x(), secr.y()-area.y(), secr.width(), secr.height(), sameImage,
                      useMask, mask, layer->getProtectColors(), transCol, groundColors, layerColors);
             if (!origImg.isNull())
               bitBltAddCmap(&layer->origBuf, sx, sy, &origImg,
                      secr.x()-area.x(), secr.y()-area.y(), secr.width(), secr.height(),
                      sameImage, useMask, mask, layer->getProtectColors(), transCol, groundColors, layerColors);
            }
	   else
	    {
             if (!img.isNull())
               bitBlt(&layer->imageBuf, sx, sy, &img,
                      secr.x()-area.x(), secr.y()-area.y(), secr.width(), secr.height(), sameImage,
                      useMask, mask, layer->getProtectColors(), transCol);
             if (!origImg.isNull())
               bitBlt(&layer->origBuf, sx, sy, &origImg,
                      secr.x()-area.x(), secr.y()-area.y(), secr.width(), secr.height(),
                      sameImage, useMask, mask, layer->getProtectColors(), transCol);
            }
         }
         if (extend.isValid() && area.intersects(extend))
         {
           secr = area.intersect(extend);
           imgx = imgy = 0;
           if (tile->getSecondHor())
             imgx = layer->imageBuf.width() - extend.width();
           else
             imgy = layer->imageBuf.height() - extend.height();
           sx = secr.x() - extend.x() + imgx;
           sy = secr.y() - extend.y() + imgy;
           if (tile->mirrorHor())
            sx = (layer->imageBuf.width()-sx-secr.width());
           if (tile->mirrorVert())
            sy = (layer->imageBuf.height()-sy-secr.height());
	   if (cmapAdd)
	    {       
             if (!img.isNull())
               bitBltAddCmap(&layer->imageBuf, sx, sy, &img,
                      secr.x()-area.x(), secr.y()-area.y(), secr.width(), secr.height(),
                      sameImage, useMask, mask, layer->getProtectColors(), transCol, groundColors, layerColors);
             if (!origImg.isNull())
               bitBltAddCmap(&layer->origBuf, sx, sy, &origImg,
                      secr.x()-area.x(), secr.y()-area.y(), secr.width(), secr.height(),
                      sameImage, useMask, mask, layer->getProtectColors(), transCol, groundColors, layerColors);
            }
	   else
	    {       
             if (!img.isNull())
               bitBlt(&layer->imageBuf, sx, sy, &img,
                      secr.x()-area.x(), secr.y()-area.y(), secr.width(), secr.height(),
                      sameImage, useMask, mask, layer->getProtectColors(), transCol);
             if (!origImg.isNull())
               bitBlt(&layer->origBuf, sx, sy, &origImg,
                      secr.x()-area.x(), secr.y()-area.y(), secr.width(), secr.height(),
                      sameImage, useMask, mask, layer->getProtectColors(), transCol);
            }
         }
       }
      if (doCmapAdd)
       {
        layer->imageBuf.removeUnusedColors();
        layer->imageBuf.mergeColors();
       }
    }
   else
    {
     if (cmapAdd)
      {       
       if (!img.isNull())
         bitBltAddCmap(&layer->imageBuf, area.x(),area.y(),
          &img, 0,0, area.width(),area.height(), sameImage, useMask, mask, layer->getProtectColors(), transCol, groundColors, layerColors);

       if (!origImg.isNull())
         bitBltAddCmap(&layer->origBuf, area.x(),area.y(),
          &origImg, 0,0, area.width(),area.height(), sameImage, useMask, mask, layer->getProtectColors(), transCol, groundColors, layerColors);
       layer->imageBuf.removeUnusedColors();
       layer->imageBuf.mergeColors();
      }
     else
      {
       if (!img.isNull())
         bitBlt(&layer->imageBuf, area.x(),area.y(),
        	&img, 0,0, area.width(),area.height(), sameImage, useMask, mask, layer->getProtectColors(), transCol);
       if (!origImg.isNull())
         bitBlt(&layer->origBuf, area.x(),area.y(),
        	&origImg, 0,0, area.width(),area.height(), sameImage, useMask, mask, layer->getProtectColors(), transCol);
      }
    }
}

void ApaintDoc::freeLayers()
{
 int i;

 for (i=0; i<numLayers(); i++)
 {
  delete layerList[i];
  layerList[i]=0;
 }
 nLayers=0;
 selectedLayer=0;
}


void ApaintDoc::collapsLayer(int i, bool cmapAdd)
{
 bool sameImage = false;
 QRect area = layerList[i]->layerPos();

 if (layerInBox(i, area))
   copyImageToLayer(mainLayer, area, layerList[i]->imageBuf, layerList[i]->origBuf,
                    sameImage, layerList[i]->usingMask(), &layerList[i]->getMask(), layerList[i]->getTransColors(), cmapAdd);

 deleteLayer(i);
 Layer *layer = layerList[selectedLayer];
 if (layer)
  layer->origBuf = layer->imageBuf.copy();
}

void ApaintDoc::collapsSelectedLayer(bool cmapAdd)
{
 if (selectedLayer == 0)  return;
 if (!layerList[selectedLayer]) return;
  collapsLayer(selectedLayer, cmapAdd);

}

void ApaintDoc::collapsLayers()
{
 int i;
 bool sameImage = false;
 for (i=1; i<numLayers(); i++)
 {
  QRect area = layerList[i]->layerPos();

  copyImageToLayer(mainLayer, area, layerList[i]->imageBuf, layerList[i]->origBuf,
                  sameImage, layerList[i]->usingMask(), &layerList[i]->getMask(), layerList[i]->getTransColors(), FALSE);
  delete layerList[i];
  emit removeLayer(i);
  layerList[i] = 0;
 }
 nLayers=1;
 selectedLayer=0;
 emit layerChanged(layerList[selectedLayer]);
}

Layer *ApaintDoc::getLayer(int num)
{
 if (num>=0 && num<numLayers())
  return(layerList[num]);
 return 0;
}

void ApaintDoc::redrawLayers()
{
 for (int i=1; i<numLayers(); i++)
  refreshLayer(getLayerPos(i));
}

void ApaintDoc::sendSignals(bool layerChg, bool mainLayerChg)
{
 if (!layerList[selectedLayer]) return ;
 if (layerChg)
  emit layerChanged(layerList[selectedLayer]);
 if (mainLayerChg)
  emit mainLayerChanged();
}

bool ApaintDoc::layerInBox()
{
 return (layerInBox(selectedLayer, mainLayer->layerPos()));
}

bool ApaintDoc::layerInBox(int i, QRect box)
{
 if (!layerList[i])
  return false;
 QRect pos = layerList[i]->layerPos();
 if (box.intersects(pos))
   return true;
 else
   return false;
}


bool ApaintDoc::removeNullLayer()
{
 if (!layerList[selectedLayer])
  return false;
 QRect r = layerList[selectedLayer]->layerPos();
 if (!r.width() || !r.height() || !r.isValid())
  {
   deleteLayer(selectedLayer);
   return true;
  }
 return false;
}


void ApaintDoc::moveLayer(int dx, int dy)
{
 if (!layerList[selectedLayer])
  return;
 layerList[selectedLayer]->move(dx, dy);
 emit layerMovedSig(layerList[selectedLayer]->layerPos());
}

bool ApaintDoc::isModified() const
{
 return modified;
}


void ApaintDoc::refreshLayer()
{
 if (!layerList[selectedLayer])
  return;
 QRect pos = layerList[selectedLayer]->layerPos();
 emit refreshSig(pos.x()-2, pos.y()-2, pos.width()+4, pos.height()+4);  
}

void ApaintDoc::refreshLayer(QRect pos)
{
 if (!layerList[selectedLayer])
  return;
 if (!layerList[selectedLayer]->isUsingTiles())
  emit refreshSig(pos.x()-2, pos.y()-2, pos.width()+4, pos.height()+4);
}

void ApaintDoc::refreshLayer(QRect oldPos, QRect newPos, int border)
{
 if (!layerList[selectedLayer])
  return;
 emit refreshLayerSig(oldPos, newPos, border) ;
}

/*
void ApaintDoc::setLayerMask()
{
 if (!layerList[selectedLayer])
  return;
 layerList[selectedLayer]->setMask();
}

void ApaintDoc::setLayerMask(QImage mask)
{
 if (!layerList[selectedLayer])
  return;
 layerList[selectedLayer]->setMask(mask);
}
*/

void ApaintDoc::beginLayerTransform(QPointArray pt)
{
 if (!layerList[selectedLayer])
  return;
 layerIsTransforming = true;
 layerList[selectedLayer]->beginTransform(pt);
}

void ApaintDoc::transformLayer(QWMatrix &mx, int dstWid, int dstHei)
{
 if (!layerList[selectedLayer])
  return;
 if (!layerIsTransforming)
  return;
   
 layerList[selectedLayer]->xForm(mx, dstWid, dstHei);
 if (selectedLayer == 0)
  emit mainLayerChanged();
}

void ApaintDoc::transformLayer(QWMatrix &mx, double angle, int dstWid, int dstHei, bool resizeContents,
                               bool rotateContour, bool doAverage, ResizeFilterType ftyp)
{
 int mulXY=1;

 if (!layerList[selectedLayer]) return;
 if (!layerIsTransforming) return;

if (resizeContents)
 {
  if (ftyp==FILTER_SMOOTH2)
   {
    layerList[selectedLayer]->scale2X();
    mulXY=2;
   }
  if (ftyp==FILTER_SMOOTH3)
   {
    layerList[selectedLayer]->scale3X();
    mulXY=3;
   }
  if (ftyp==FILTER_SMOOTH4)
   {
    layerList[selectedLayer]->scale2X();
    layerList[selectedLayer]->scale2X();
    mulXY=4;
   }
 }

 if (layerList[selectedLayer]->getImageType() == LOAD_TRUE_COL)
  layerList[selectedLayer]->scaleRotate(mx, angle, dstWid*mulXY, dstHei*mulXY, resizeContents,
                                        layerList[selectedLayer]->getBgCol(), rotateContour, doAverage, ftyp);
 else
   layerList[selectedLayer]->scaleRotate(mx, angle, dstWid*mulXY, dstHei*mulXY, resizeContents,
                                         layerList[selectedLayer]->getBgIndx(), rotateContour, doAverage, ftyp);
if (resizeContents)
 {
  if (ftyp==FILTER_SMOOTH2)
   layerList[selectedLayer]->shrink2X();
  if (ftyp==FILTER_SMOOTH3)
   layerList[selectedLayer]->shrink3X();
  if (ftyp==FILTER_SMOOTH4)
   {
    layerList[selectedLayer]->shrink2X();
    layerList[selectedLayer]->shrink2X();
   }
 }

 if (selectedLayer == 0)
  emit mainLayerChanged(); 
}

void ApaintDoc::endLayerTransform()
{
  if (!layerList[selectedLayer]) return;
  if (!layerIsTransforming) return;
  
  layerList[selectedLayer]->endTransform();
  emit layerChanged(layerList[selectedLayer]);
  layerIsTransforming = false;
}

void ApaintDoc::putUndoStack(LayerChangeType ctyp)
{
 putUndoStack(layerList[selectedLayer], ctyp);
}

void ApaintDoc::putUndoStack(Layer *sel, LayerChangeType ctyp)
{
 if (!sel) return;
 emit putUndoStackSig(sel, ctyp);
}

void  ApaintDoc::setLayerPos(int x, int y)
{
  if (!layerList[selectedLayer]) return;
  
  layerList[selectedLayer]->newPos(x, y);
  emit layerPosChanged();
}

void ApaintDoc::setLayerType(LayerType typ)
{
  if (!layerList[selectedLayer]) return;
  layerList[selectedLayer]->setLayerType(typ);
}

void ApaintDoc::setLayerNumColors(int num, int noise, bool dither, bool preview)
{
 if (!layerList[selectedLayer]) return;

 if (layerList[selectedLayer]->getImageType() == LOAD_TRUE_COL)
  layerList[selectedLayer]->setType(LOAD_CMAP, num, noise, dither, preview);
 else if (layerList[selectedLayer]->getImageType() == LOAD_CMAP)
  layerList[selectedLayer]->setNumColors(num, preview);

 emit layerChanged(layerList[selectedLayer]);
 if (layerList[selectedLayer]->getType()!=MainLayer)
   refreshLayer();
}

void ApaintDoc::setLayerExactNumColors(int num)
{
 if (!layerList[selectedLayer]) return;

 layerList[selectedLayer]->setExactNumColors(num);

 emit layerChanged(layerList[selectedLayer]);
 if (layerList[selectedLayer]->getType()!=MainLayer)
  refreshLayer(); 
}

void ApaintDoc::setLayerImageType(int imgType, bool dither)
{
  if (!layerList[selectedLayer]) return;
  
  layerList[selectedLayer]->setType(imgType, 256, 0, dither, false);
  emit layerChanged(layerList[selectedLayer]);
  if (layerList[selectedLayer]->getType()!=MainLayer)
   refreshLayer();
}

/*
void ApaintDoc::setLayerExactColorTable(QRgb *ctab, int num)
{
  if (!layerList[selectedLayer]) return;

  layerList[selectedLayer]->setExactColorTable(ctab, num);
  emit layerChanged(layerList[selectedLayer]);
  if (layerList[selectedLayer]->getType()!=MainLayer)
   refreshLayer();
}
*/

void ApaintDoc::setLayerColorTable(QRgb *ctab, int num, bool dither, bool preview)
{
  if (!layerList[selectedLayer]) return;
  if (!ctab || !num) return;
  
  layerList[selectedLayer]->setColorTable(ctab, num, dither, preview);
  emit layerChanged(layerList[selectedLayer]);
  if (layerList[selectedLayer]->getType()!=MainLayer)
   refreshLayer();
}

void ApaintDoc::finishLayerPreview()
{
  if (!layerList[selectedLayer]) return;
  layerList[selectedLayer]->finishPreview();
  emit layerChanged(layerList[selectedLayer]);
  if (layerList[selectedLayer]->getType()!=MainLayer)
     refreshLayer();

}

void ApaintDoc::undoLayerPreview(bool dataPreview, bool emitRefresh)
{
  if (!layerList[selectedLayer]) return;
  
  if (dataPreview)
    layerList[selectedLayer]->undoDataPreview();
  else
    layerList[selectedLayer]->undoPreview();

  if (emitRefresh)
  {
    emit layerChanged(layerList[selectedLayer]);
    if (layerList[selectedLayer]->getType()!=MainLayer)
     refreshLayer();
  }  
}


void ApaintDoc::cutSelectedLayer()
{
  if (!layerList[selectedLayer]) return;
  if (layerList[selectedLayer]->getType() == MainLayer) return;

  LayerDrag *ld = layerList[selectedLayer]->getClipboardData();
  QClipboard *cb = QApplication::clipboard();
  cb->setData(ld);
  putUndoStack(DestroyLayer);
  deleteLayer();
  refreshLayer();
}
  
void ApaintDoc::copySelectedLayer()
{
  if (!layerList[selectedLayer]) return;
   if (layerList[selectedLayer]->getType() == MainLayer) return;
    
  LayerDrag *ld = layerList[selectedLayer]->getClipboardData();
  QClipboard *cb = QApplication::clipboard();
  
  cb->setData(ld);
}

void ApaintDoc::pasteSelectedLayer(int posx, int posy)
{  
  int px, py;
  
  if (!layerList[selectedLayer]) return;
  px = posx;
  py = posy;
  if (nLayers>1)
  {
    px = layerList[selectedLayer]->layerPos().x();
    py = layerList[selectedLayer]->layerPos().y();
    putUndoStack(DestroyLayer);
    collapsSelectedLayer(false);
  }
  if (nLayers<MAX_LAYERS)
  {
   QClipboard *cb = QApplication::clipboard();
   QMimeSource *ms = cb->data();
   if (!ms->provides(APAINT_MIME_SIG)) return;

   QByteArray rd = ms->encodedData(APAINT_MIME_SIG);
   QDataStream s(rd, IO_ReadWrite);
   Q_INT8 um;
   ApaintImage aimg;   
   QImage img;
   
   s >> um;
   s >> aimg;
   if (um)
    s >> img;  

   img = img.convertBitOrder(QImage::LittleEndian);
   aimg.recalcJumpTable();   
   selectedLayer = nLayers;
   layerList[selectedLayer] = new Layer(aimg, px, py,  &img, SelectionLayer, aspectX, aspectY);
   if (!layerList[selectedLayer])
   {
     qDebug("ApaintDoc: NULL pointer!");
     return;
   }
   putUndoStack(CreateLayer);
   nLayers++;
   emit newLayerSig(layerList[selectedLayer]);
   emit layerChanged(layerList[selectedLayer]);
   refreshLayer();  
  }
}

LayerDrag *ApaintDoc::getClipboardData()
{
  QClipboard *cb = QApplication::clipboard();
  QMimeSource *ms = cb->data();
  if (!ms->provides(APAINT_MIME_SIG)) return 0;
  return((LayerDrag *)ms);
}

void ApaintDoc::mainLayerToSelection(Layer *newMainLayer)
{
  if (!newMainLayer) return;
  if (newMainLayer->getType() != MainLayer) return;
  if (nLayers>=MAX_LAYERS) return;

  /*shift all layers*/
  for (int i=0; i<nLayers; i++)
   layerList[i+1]  = layerList[i];
  nLayers++;
  newMainLayer->enableMask(false);
  layerList[0] = newMainLayer;  
  mainLayer = layerList[0];
  selectedLayer = 1;
  layerList[selectedLayer]->setLayerType(SelectionLayer);
  emit mainLayerChanged();
  emit layerChanged(layerList[selectedLayer]);
  emit newLayerSig(layerList[selectedLayer]);
}

void ApaintDoc::selectionToMainLayer()
{
  if (!layerList[selectedLayer]) return;
  if (layerList[selectedLayer]->getType()!= SelectionLayer) return;

  QString name = mainLayer->getFilename();

  delete layerList[0];/* del main layer*/
  layerList[0] = layerList[selectedLayer];
  mainLayer = layerList[selectedLayer];
  layerList[0]->newPos(0, 0);  
  layerList[0]->setLayerType(MainLayer);
  layerList[0]->enableMask(false);
  layerList[0]->resetColors();
  layerList[0]->enableMask(false);
  for (int i=selectedLayer; i<numLayers()-1; i++)
   layerList[i] = layerList[i+1];
  nLayers--;
  selectedLayer=0;  
  /* restore the previous filename */
  mainLayer->setFilename(name);
  emit layerChanged(layerList[selectedLayer]);
  emit mainLayerChanged();
  emit newLayerSig(layerList[selectedLayer]);
}

void ApaintDoc::setFgIndx(int indx)
{
  if (!layerList[selectedLayer]) return;
  layerList[selectedLayer]->setFgIndx(indx);
}

void ApaintDoc::setFgCol(QRgb rgb)
{
  if (!layerList[selectedLayer]) return;
  layerList[selectedLayer]->setFgCol(rgb);
}

void ApaintDoc::setBgIndx(int indx)
{
  if (!layerList[selectedLayer]) return;
  layerList[selectedLayer]->setBgIndx(indx); 
}

void ApaintDoc::setBgCol(QRgb rgb)
{
  if (!layerList[selectedLayer]) return;
  layerList[selectedLayer]->setBgCol(rgb); 
}

QRgb *ApaintDoc::getLayerColorTable()
{
 if (!layerList[selectedLayer]) return 0;
 
 return layerList[selectedLayer]->getColorTable();
}

int ApaintDoc::layerNumColors()
{
 if (!layerList[selectedLayer]) return 0;

 return layerList[selectedLayer]->numColors(); 
}

/*
void ApaintDoc::getLayerUsedColors(bool *uc, int &num)
{
 if (!layerList[selectedLayer] || !uc) return;

 layerList[selectedLayer]->getUsedColors(uc, num);
}
*/

void ApaintDoc::setLayerBorderPoints(QPointArray &pts)
{
  if (!layerList[selectedLayer]) return ;
  
  layerList[selectedLayer]->setBorderPoints(pts);  
}

bool ApaintDoc::isLayerPointVisible(int x, int y)
{
  if (!layerList[selectedLayer]) return false;
  if (layerList[selectedLayer]->getType() == MainLayer) return false;

  return layerList[selectedLayer]->pointVisible(x, y);  
}

void ApaintDoc::layerApplyFilter(int what, int fgCol, int bgCol, int penX, int penY, int value, bool doVarColors, bool doVarSize, bool doGradient)
{
 if (!layerList[selectedLayer]) return ;

 layerList[selectedLayer]->applyFilter(what, fgCol, bgCol, penX, penY, value, doVarColors, doVarSize, doGradient);
 refreshLayer();
}

void ApaintDoc::layerContourAverage(int thickX, int thickY)
{
  if (!layerList[selectedLayer]) return ;

  layerList[selectedLayer]->contourAverage(thickX, thickY);
  refreshLayer();
}

void ApaintDoc::layerContourDespecle(int what, int neighbours)
{
  if (!layerList[selectedLayer]) return ;

  layerList[selectedLayer]->contourDespecle(what, neighbours);
  refreshLayer();
}

void ApaintDoc::layerContourThick(uint col, int direction, int timesX, int timesY, int angle, bool doShade, bool doLines,
 bool doGradient, bool doBidirectional, bool repeatOperation)
{
  if (!layerList[selectedLayer]) return ;

  layerList[selectedLayer]->contourThick(col, direction, timesX, timesY, angle, doShade, doLines, doGradient, doBidirectional, repeatOperation);
  refreshLayer();
}

void ApaintDoc::layerContourRemove(uint col)
{
  if (!layerList[selectedLayer]) return ;

  layerList[selectedLayer]->contourRemove(col);
  refreshLayer();

}

void ApaintDoc::layerContourPrune(uint col)
{
  if (!layerList[selectedLayer]) return ;

  layerList[selectedLayer]->contourPrune(col);
  refreshLayer();

}

void ApaintDoc::layerContourThin(uint col)
{
  if (!layerList[selectedLayer]) return ;

  layerList[selectedLayer]->contourThin(col);
   refreshLayer();
}

void ApaintDoc::layerContourAdd(uint fgCol, uint bgCol, uint withCol, bool useWithCol, int direction, bool into, bool doShade, int numX, int numY)
{
  if (!layerList[selectedLayer]) return ;

  layerList[selectedLayer]->contourAdd(fgCol, bgCol, withCol, useWithCol, direction, into, doShade, numX, numY);
  refreshLayer();
}

void ApaintDoc::changeLayerToBW(int treshHold, bool preview)
{
  if (!layerList[selectedLayer]) return ;

  layerList[selectedLayer]->changeToBW(treshHold, preview);
  emit layerChanged(layerList[selectedLayer]);
  if (layerList[selectedLayer]->getType() !=MainLayer)
   refreshLayer();
}

void ApaintDoc::layerMirrorHorizontal()
{
  if (!layerList[selectedLayer]) return ;

  layerList[selectedLayer]->mirrorHorizontal();
  refreshLayer();
}

void ApaintDoc::layerMirrorVertical()
{
 if (!layerList[selectedLayer]) return ;

 layerList[selectedLayer]->mirrorVertical();
 refreshLayer();
}

void ApaintDoc::layerMirrorExplode()
{
 if (!layerList[selectedLayer]) return ;

 layerList[selectedLayer]->mirrorExplode();
 refreshLayer();
}

/*
void ApaintDoc::layerMirrorFlip()
{
 if (!layerList[selectedLayer]) return;
 QRect oldPos = layerList[selectedLayer]->layerPos();
 layerList[selectedLayer]->mirrorFlip();
 QRect newPos = layerList[selectedLayer]->layerPos();
 if (layerList[selectedLayer]->getType() == MainLayer)
  emit mainLayerChanged();
 else
  {
   refreshLayer(oldPos);
   refreshLayer(newPos);
  }
 emit layerChanged(layerList[selectedLayer]);
}
*/
void ApaintDoc::layerMirrorFlip()
{
 if (!layerList[selectedLayer]) return;
 QRect oldPos = layerList[selectedLayer]->layerPos();
 layerList[selectedLayer]->mirrorFlip();
 QRect newPos = layerList[selectedLayer]->layerPos();
 if (layerList[selectedLayer]->getType() == MainLayer)
  {
   int xdpi, ydpi;

   xdpi = layerList[selectedLayer]->imageBuf.xdpi();
   ydpi = layerList[selectedLayer]->imageBuf.ydpi();
   layerList[selectedLayer]->imageBuf.setXdpi(ydpi);
   layerList[selectedLayer]->imageBuf.setYdpi(xdpi);
   emit mainLayerChanged();
  }
 else
  {
   refreshLayer(oldPos);
   refreshLayer(newPos);
  }
 emit layerChanged(layerList[selectedLayer]);
}

void ApaintDoc::layerDuplicate(int duplicateType)
{
 if (!layerList[selectedLayer]) return;

 putUndoStack(ChangeLayerForm);

 QRect oldPos = layerList[selectedLayer]->layerPos();
 layerList[selectedLayer]->duplicate(duplicateType);
 QRect newPos = layerList[selectedLayer]->layerPos();
 if (layerList[selectedLayer]->getType() == MainLayer)
  emit mainLayerChanged();
 else
  {
   refreshLayer(oldPos);
   refreshLayer(newPos);
  }
 emit layerChanged(layerList[selectedLayer]);
}

void ApaintDoc::layerScale2X()
{
 if (!layerList[selectedLayer]) return;

 putUndoStack(ChangeLayerForm);

 QRect oldPos = layerList[selectedLayer]->layerPos();
 layerList[selectedLayer]->scale2X();
 QRect newPos = layerList[selectedLayer]->layerPos();
 if (layerList[selectedLayer]->getType() == MainLayer)
  emit mainLayerChanged();
 else
  {
   refreshLayer(oldPos);
   refreshLayer(newPos);
  }
 emit layerChanged(layerList[selectedLayer]);
}

void ApaintDoc::layerScale3X()
{
 if (!layerList[selectedLayer]) return;

 putUndoStack(ChangeLayerForm);

 QRect oldPos = layerList[selectedLayer]->layerPos();
 layerList[selectedLayer]->scale3X();
 QRect newPos = layerList[selectedLayer]->layerPos();
 if (layerList[selectedLayer]->getType() == MainLayer)
  emit mainLayerChanged();
 else
  {
   refreshLayer(oldPos);
   refreshLayer(newPos);
  }
 emit layerChanged(layerList[selectedLayer]);
}

void ApaintDoc::layerScale4X()
{
 if (!layerList[selectedLayer]) return;

 putUndoStack(ChangeLayerForm);

 QRect oldPos = layerList[selectedLayer]->layerPos();
 layerList[selectedLayer]->scale2X();
 layerList[selectedLayer]->scale2X();
 QRect newPos = layerList[selectedLayer]->layerPos();
 if (layerList[selectedLayer]->getType() == MainLayer)
  emit mainLayerChanged();
 else
  {
   refreshLayer(oldPos);
   refreshLayer(newPos);
  }
 emit layerChanged(layerList[selectedLayer]);
}

void ApaintDoc::layerShrink2X()
{
 if (!layerList[selectedLayer]) return;

 putUndoStack(ChangeLayerForm);

 QRect oldPos = layerList[selectedLayer]->layerPos();
 layerList[selectedLayer]->shrink2X();
 QRect newPos = layerList[selectedLayer]->layerPos();
 if (layerList[selectedLayer]->getType() == MainLayer)
  emit mainLayerChanged();
 else
  {
   refreshLayer(oldPos);
   refreshLayer(newPos);
  }
 emit layerChanged(layerList[selectedLayer]);
}

void ApaintDoc::layerShrink3X()
{
 if (!layerList[selectedLayer]) return;

 putUndoStack(ChangeLayerForm);

 QRect oldPos = layerList[selectedLayer]->layerPos();
 layerList[selectedLayer]->shrink3X();
 QRect newPos = layerList[selectedLayer]->layerPos();
 if (layerList[selectedLayer]->getType() == MainLayer)
  emit mainLayerChanged();
 else
  {
   refreshLayer(oldPos);
   refreshLayer(newPos);
  }
 emit layerChanged(layerList[selectedLayer]);
}

void ApaintDoc::layerShrink4X()
{
 if (!layerList[selectedLayer]) return;

 putUndoStack(ChangeLayerForm);

 QRect oldPos = layerList[selectedLayer]->layerPos();
 layerList[selectedLayer]->shrink4X();
 QRect newPos = layerList[selectedLayer]->layerPos();
 if (layerList[selectedLayer]->getType() == MainLayer)
  emit mainLayerChanged();
 else
  {
   refreshLayer(oldPos);
   refreshLayer(newPos);
  }
 emit layerChanged(layerList[selectedLayer]);
}

void ApaintDoc::layerHalve(int halveType)
{
 if (!layerList[selectedLayer]) return;

 putUndoStack(ChangeLayerForm);

 QRect oldPos = layerList[selectedLayer]->layerPos();
 layerList[selectedLayer]->halve(halveType);
 QRect newPos = layerList[selectedLayer]->layerPos();
 if (layerList[selectedLayer]->getType() == MainLayer)
  emit mainLayerChanged();
 else
  {
   refreshLayer(oldPos);
   refreshLayer(newPos);
  }
 emit layerChanged(layerList[selectedLayer]);
}

void ApaintDoc::layerShrink(int shrinkType)
{
 if (!layerList[selectedLayer]) return;

 putUndoStack(ChangeLayerForm);

 QRect oldPos = layerList[selectedLayer]->layerPos();
 layerList[selectedLayer]->shrink(shrinkType);
 QRect newPos = layerList[selectedLayer]->layerPos();
 if (layerList[selectedLayer]->getType() == MainLayer)
  emit mainLayerChanged();
 else
  {
   refreshLayer(oldPos);
   refreshLayer(newPos);
  }
 emit layerChanged(layerList[selectedLayer]);
}

void ApaintDoc::layerMultiplyX(int times)
{
 if (!layerList[selectedLayer]) return;

 putUndoStack(ChangeLayerForm);

 QRect oldPos = layerList[selectedLayer]->layerPos();
 layerList[selectedLayer]->multiplyX(times);
 QRect newPos = layerList[selectedLayer]->layerPos();
 if (layerList[selectedLayer]->getType() == MainLayer)
  emit mainLayerChanged();
 else
  {
   refreshLayer(oldPos);
   refreshLayer(newPos);
  }
 emit layerChanged(layerList[selectedLayer]);
}

void ApaintDoc::layerMultiplyY(int times)
{
 if (!layerList[selectedLayer]) return;

 putUndoStack(ChangeLayerForm);

 QRect oldPos = layerList[selectedLayer]->layerPos();
 layerList[selectedLayer]->multiplyY(times);
 QRect newPos = layerList[selectedLayer]->layerPos();
 if (layerList[selectedLayer]->getType() == MainLayer)
  emit mainLayerChanged();
 else
  {
   refreshLayer(oldPos);
   refreshLayer(newPos);
  }
 emit layerChanged(layerList[selectedLayer]);
}

void ApaintDoc::layerSortColors(int sortOrder)
{
 if (!layerList[selectedLayer]) return;

 if (layerList[selectedLayer]->getImageType()!=LOAD_CMAP)
  return;

/* currently we save the whole image for undo, but for colormap images we should only save the palette */
/* should work like this:
 if (layerList[selectedLayer]->getImageType() == LOAD_TRUE_COL)
  putUndoStack(ChangeLayerForm);
 else
  putUndoStack(ChangeColorTable);
 but then undo/redo does not work correctly */
 putUndoStack(ChangeLayerForm);
 layerList[selectedLayer]->sortColors(sortOrder);
 if (sortOrder==4) /* SORT_SIMILARITY */
  setFgIndx(0);
 emit layerChanged(layerList[selectedLayer]);
 if (layerList[selectedLayer]->getType()!=MainLayer)
  refreshLayer();
}

void ApaintDoc::layerInvertColors(int what)
{
 bool *protCol;
 if (!layerList[selectedLayer]) return;

 protCol = layerList[selectedLayer]->getProtectColors();

 if ((what==SWAP_COLORS_PALETTE) && (layerList[selectedLayer]->getImageType() == LOAD_TRUE_COL))
  { QApplication::beep(); return; } 

/* currently we save the whole image for undo, but for colormap images we should only save the palette */
/* should work like this:
 if (layerList[selectedLayer]->getImageType() == LOAD_TRUE_COL)
  putUndoStack(ChangeLayerForm);
 else
  putUndoStack(ChangeColorTable);
 but then undo/redo does not work correctly */

 switch (what)
  {
   case INVERT_COLORS_ALL: putUndoStack(InvertAllColors); break;
   case INVERT_COLORS_RED: putUndoStack(InvertRed); break;
   case INVERT_COLORS_GREEN: putUndoStack(InvertGreen); break;
   case INVERT_COLORS_BLUE: putUndoStack(InvertBlue); break;
   case INVERT_COLORS_CYAN: putUndoStack(InvertCyan); break;
   case INVERT_COLORS_YELLOW: putUndoStack(InvertYellow); break;
   case INVERT_COLORS_MAGENTA: putUndoStack(InvertMagenta); break;
   case SWAP_COLORS_RGB_GBR: putUndoStack(SwapRGB_GBR); break;
   case SWAP_COLORS_RGB_BRG: putUndoStack(SwapRGB_BRG); break;
   case SWAP_COLORS_RG: putUndoStack(SwapRG); break;
   case SWAP_COLORS_RB: putUndoStack(SwapRB); break;
   case SWAP_COLORS_GB: putUndoStack(SwapGB); break;
   default: putUndoStack(ChangeLayerForm); break;
  }

 layerList[selectedLayer]->invertColors(what, protCol);
 emit layerChanged(layerList[selectedLayer]);
 if (layerList[selectedLayer]->getType()!=MainLayer)
  refreshLayer();
}

void ApaintDoc::layerSwapColors(int first, int second)
{
 if (!layerList[selectedLayer]) return;

 if (layerList[selectedLayer]->getImageType() == LOAD_TRUE_COL)
  { QApplication::beep(); return; } 


// putUndoStack(ChangeLayerForm);
 layerList[selectedLayer]->setSwapColor1(first);
 layerList[selectedLayer]->setSwapColor2(second);
 putUndoStack(SwapColors);

 layerList[selectedLayer]->swapColors(first, second);
 emit layerChanged(layerList[selectedLayer]);
 if (layerList[selectedLayer]->getType()!=MainLayer)
  refreshLayer();
}

void ApaintDoc::layerContrastStretch()
{
 if (!layerList[selectedLayer]) return;

/* currently we save the whole image for undo, but for colormap images we should only save the palette */
/* should work like this:
 if (layerList[selectedLayer]->getImageType() == LOAD_TRUE_COL)
  putUndoStack(ChangeLayerForm);
 else
  putUndoStack(ChangeColorTable);
 but then undo/redo does not work correctly */
 putUndoStack(ChangeLayerForm);
 layerList[selectedLayer]->contrastStretch();
 emit layerChanged(layerList[selectedLayer]);
 if (layerList[selectedLayer]->getType()!=MainLayer)
  refreshLayer();
}

void ApaintDoc::layerMakeColorShades(int fgIndex, int bgIndex)
{
 if (!layerList[selectedLayer]) return;

 putUndoStack(ChangeLayerForm);
 layerList[selectedLayer]->makeColorShades(fgIndex, bgIndex);
 emit layerChanged(layerList[selectedLayer]);
 if (layerList[selectedLayer]->getType()!=MainLayer)
  refreshLayer();
}

void ApaintDoc::layerInvertProtectedColors()
{
 bool *protCol;
 int i, numColors;

 if (!layerList[selectedLayer])
  return;
 if (layerList[selectedLayer]->getImageType()!=LOAD_CMAP)
  return;

 numColors = layerList[selectedLayer]->numColors();
 protCol = layerList[selectedLayer]->getProtectColors();

 for (i=0; i<numColors; i++)
  protCol[i]=!protCol[i];
 layerList[selectedLayer]->setProtectedColors(protCol);

 emit layerChanged(layerList[selectedLayer]);
 if (layerList[selectedLayer]->getType()!=MainLayer)
  refreshLayer();
}

void ApaintDoc::layerClearProtectedColors()
{
 bool *protCol;
 int i, numColors;

 if (!layerList[selectedLayer])
  return;
 if (layerList[selectedLayer]->getImageType()!=LOAD_CMAP)
  return;

 numColors = layerList[selectedLayer]->numColors();
 protCol = layerList[selectedLayer]->getProtectColors();

 for (i=0; i<numColors; i++)
  protCol[i]=FALSE;
 layerList[selectedLayer]->setProtectedColors(protCol);

 emit layerChanged(layerList[selectedLayer]);
 if (layerList[selectedLayer]->getType()!=MainLayer)
  refreshLayer();
}

void ApaintDoc::layerInvertTransparentColors()
{
 bool *transCol;
 int i, numColors;

 if (!layerList[selectedLayer])
  return;
 if (layerList[selectedLayer]->getImageType()!=LOAD_CMAP)
  return;
 if (layerList[selectedLayer]->getType()==MainLayer)
  return;

 numColors = layerList[selectedLayer]->numColors();
 transCol = layerList[selectedLayer]->getTransColors();

 for (i=0; i<numColors; i++)
  transCol[i]=!transCol[i];
 layerList[selectedLayer]->setTransColors(transCol);

 emit layerChanged(layerList[selectedLayer]);
 refreshLayer();
}

void ApaintDoc::layerClearTransparentColors()
{
 bool *transCol;
 int i, numColors;

 if (!layerList[selectedLayer])
  return;
 if (layerList[selectedLayer]->getImageType()!=LOAD_CMAP)
  return;
 if (layerList[selectedLayer]->getType()==MainLayer)
  return;

 numColors = layerList[selectedLayer]->numColors();
 transCol = layerList[selectedLayer]->getTransColors();

 for (i=0; i<numColors; i++)
  transCol[i]=FALSE;
 layerList[selectedLayer]->setTransColors(transCol);

 emit layerChanged(layerList[selectedLayer]);
 refreshLayer();
}

void ApaintDoc::layerPerspectiveXForm(int dstx, int dsty, int dstw, int dsth, double mx[3][3])
{
  if (!layerList[selectedLayer]) return;
  layerList[selectedLayer]->perspectiveXForm(mainLayer, dstx, dsty, dstw, dsth, mx);
  QRect pos = layerList[selectedLayer]->layerPos();
  if (layerList[selectedLayer]->getType() == MainLayer)
   emit mainLayerChanged();
  else
   refreshLayer(pos);

  emit layerChanged(layerList[selectedLayer]);
}

void ApaintDoc::setLayerTransColor(int i, bool enable)
{
 if (!layerList[selectedLayer]) return;
 if (layerList[selectedLayer]->getType() == MainLayer) return;
 layerList[selectedLayer]->setTransColor(i, enable);
 emit transProtColorsChanged();
 refreshLayer();
}

void ApaintDoc::setLayerTransColors(bool *transTab)
{
 if (!layerList[selectedLayer]) return;
 if (layerList[selectedLayer]->getType() == MainLayer) return;
 layerList[selectedLayer]->setTransColors(transTab);
 refreshLayer();
}

void ApaintDoc::setLayerProtectColor(int i, bool enable)
{
 if (!layerList[selectedLayer]) return;
 layerList[selectedLayer]->setProtectedColor(i, enable);
 emit transProtColorsChanged();
 refreshLayer();  
}

void ApaintDoc::setLayerProtectColors(bool *protectTab)
{
 if (!layerList[selectedLayer]) return;
 layerList[selectedLayer]->setProtectedColors(protectTab);
 refreshLayer(); 
}

void ApaintDoc::enableMainLayerTiles(bool enable)
{
  if( !mainLayer) return ;
  if (layerList[selectedLayer] != mainLayer)
   setActiveLayer(0);

  mainLayer->enableTiles(enable);
  emit layerChanged(mainLayer);
  emit mainLayerChanged();
}

void ApaintDoc::createMainLayerTiles(int sizex, int sizey, int offx, int offy, bool justRefresh)
{
  if (!mainLayer) return;
  if (layerList[selectedLayer] != mainLayer) 
   setActiveLayer(0);

  if (!justRefresh)
    mainLayer->createTiles(sizex, sizey, offx, offy);
  emit layerChanged(mainLayer);
  emit mainLayerChanged();
}


void ApaintDoc::setActiveLayer(int num)
{
 if (num>=0 && num<numLayers())
 {
  if (layerList[selectedLayer]!=0)
   if (layerList[selectedLayer]->isPreviewMode())
     layerList[selectedLayer]->undoPreview();
  setLayerNum(num);
  emit layerChanged(layerList[selectedLayer]);
  redrawLayers();
 }
}

void ApaintDoc::newImageFromTiles()
{
  if (!mainLayer) return;
  if (!mainLayer->isUsingTiles()) return;

  mainLayer->tilesToImage();
  mainLayer->createTiles(1, 1, 0, 0);
  emit layerChanged(mainLayer);
  emit mainLayerChanged();
}

bool ApaintDoc::layerPosUp()
{
 Layer *lp;
 
 if ((selectedLayer>0) && (selectedLayer < numLayers()-1))
 {
   lp = layerList[selectedLayer];
   layerList[selectedLayer] = layerList[selectedLayer+1];
   layerList[selectedLayer+1] = lp;
   selectedLayer++;
   refreshLayer();
   return true;
 }
 return false;
}

bool ApaintDoc::layerPosDown()
{
 Layer *lp;

 if ((selectedLayer>1) && (selectedLayer < numLayers()))
 {
   lp = layerList[selectedLayer];
   layerList[selectedLayer] = layerList[selectedLayer-1];
   layerList[selectedLayer-1] = lp;
   selectedLayer--;
   refreshLayer();
   return true;
 }
 return false;
}

void ApaintDoc::removeLayerUnusedColors()
{
  if (!layerList[selectedLayer]) return;

  layerList[selectedLayer]->removeUnusedColors();
  emit layerChanged(layerList[selectedLayer]);
}

void ApaintDoc::mergeLayerColors()
{
  if (!layerList[selectedLayer]) return;

  layerList[selectedLayer]->mergeColors();
  emit layerChanged(layerList[selectedLayer]);
}

void ApaintDoc::layerSetNewData(ApaintImage &img, int x, int y, QImage *imask)
{
 if (!layerList[selectedLayer]) return;

 layerList[selectedLayer]->setNewData(img, x, y, imask);
 emit layerChanged(layerList[selectedLayer]);
 if (layerList[selectedLayer]->getType() == MainLayer)
  emit mainLayerChanged();
}


void ApaintDoc::layerSetNewData(ApaintImage &img, QImage *imask)
{
  if (!layerList[selectedLayer]) return;
  
  layerList[selectedLayer]->setNewData(img, imask);
  emit layerChanged(layerList[selectedLayer]);
}

void ApaintDoc::setLayerResizeable(bool enable)
{
 if (!layerList[selectedLayer]) return;
 layerList[selectedLayer]->setResizeable(enable); 
}

bool ApaintDoc::getLayerResizeable()
{
  if (!layerList[selectedLayer]) return false;
  return (layerList[selectedLayer]->getIsResizeable());

}

void ApaintDoc::protectLayer(bool enable)
{
  protectMode = enable;
  emit protectLayerSig(enable);
}

LayerType ApaintDoc::getLayerType()
{
  if (!layerList[selectedLayer]) return MainLayer;
  
  return layerList[selectedLayer]->getType();
}

void ApaintDoc::setLayerNum(int selLayer)
{
	selectedLayer = selLayer;
	emit layerSelected(selectedLayer);
}

void ApaintDoc::setLayerValidDraw(bool valid)
{
  if (!layerList[selectedLayer]) return;
  
  layerList[selectedLayer]->setValidDraw(valid);
}

void ApaintDoc::layerRemapData(RemapArray &rmap, bool preview)
{
  if (!layerList[selectedLayer]) return;
  layerList[selectedLayer]->remapData(rmap, preview);
  refreshLayer();
}

void ApaintDoc::reduceSelectionBorder()
{
 int numColors=0;
 bool used[256];

 if (!layerList[selectedLayer]) return;
 if (layerList[selectedLayer]->getType() == MainLayer) return;
 if (!layerList[selectedLayer]->usingMask()) return;
 if (layerList[selectedLayer]->getImageType() == LOAD_TRUE_COL) return;

 /* don't calculate border reduction if we only have one color */
 layerList[selectedLayer]->getUsedColors(used, numColors);
 if (numColors<=1)
  return;

 QRect cpyRec;
 uint borderCol = layerList[selectedLayer]->imageBuf.pixel(0, 0);
 /*if we have background color */
 if (layerList[selectedLayer]->isBorderColor(borderCol))
 {
   QRect oldPos = layerList[selectedLayer]->layerPos();
   QImage newMask = setMaskWithColor(layerList[selectedLayer]->imageBuf, borderCol);
   newMask = cutCorners(newMask, cpyRec);
   if (newMask.width() == layerList[selectedLayer]->imageBuf.width() && newMask.height() == layerList[selectedLayer]->imageBuf.height())
    return; // nothing to do!!
   if ((newMask.width()<1) || (newMask.height()<1))
    return; // nothing to do!!
   ApaintImage newImg = layerList[selectedLayer]->imageBuf.copy(cpyRec.x(), cpyRec.y(), cpyRec.width(), cpyRec.height());
   layerList[selectedLayer]->setNewData(newImg, &newMask);
   emit layerChanged(layerList[selectedLayer]);
   refreshLayer(oldPos, layerList[selectedLayer]->layerPos(), 3);
 }
}

ApaintBrushList *ApaintDoc::getBrushList()
{
 ApaintBrushList *blist = ApaintBrushList::getObject();

 return blist;
}

double ApaintDoc::getXYratio()
{
 if (layerList[selectedLayer]->imageBuf.xdpi()==0)
  return(1.0);
 if (layerList[selectedLayer]->imageBuf.ydpi()==0)
  return(1.0);
 return (double)layerList[selectedLayer]->imageBuf.xdpi()/(double)layerList[selectedLayer]->imageBuf.ydpi();
}

