/***************************************************************************
                          apaintdraw.h  -  description
                             -------------------
    begin                : Mon Mar 10 2003
    copyright            : (C) 2003-2011 Arahne, d.o.o.
    email                : arahne@arahne.si
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef APAINTDRAW_H
#define APAINTDRAW_H

#include <qobject.h>
#include <qpixmap.h>
#include <qimage.h>
#include <string.h>
#include <qpainter.h>
#include "apaintimage.h"
#include "apaintdoc.h"

class ApaintView;
class ApaintBrushList;

class ApaintDraw : public QObject  {
   Q_OBJECT
public: 
	ApaintDraw(ApaintDoc *docp, ApaintView *viewp);
	~ApaintDraw();

   QRect getDrawArea();
   
signals:
  void unRedoData(QRect &srcRec, QRect &dstRec);
  void colorTableChanged(Layer *selLayer);
  
public slots:

  void layerChangedSlot(Layer *sel);
  void refreshRBand(QRect ref, QRect drawArea, bool useFg, bool finalDraw);
  void refreshDirectDraw(QRect ref, bool useFg, bool doDraw, bool finalDraw);
  void refreshImageDraw(int x, int y, int wid, int hei, bool finalDraw);
  void zoomChangedSlot(double zx, double zy);
  void screenRefresh(int x, int y, int w, int h, int border);
  void screenRefresh(int x, int y, int w, int h);
  void screenRefreshRepeat(int x, int y, int w, int h);
  void screenRefreshRepeat(Layer *layer, int x, int y, int w, int h);
  void screenRefresh(const QRect &oldRec, const QRect &newRec, int border);
  void clearRectSlot(QRect imgArea, QRect drawArea);
  void setFgColorIndx(int indx);
  void setBgColorIndx(int indx);
  void setFgColor(QRgb rgb);
  void setBgColor(QRgb rgb);
  void setFatLines(bool enable) ;


private:

  void refreshOldData(int x1, int y1, int x2, int y2, Layer *layer);
  void refreshDrawImage(QRect &ref);
  void finalizeDraw(int xs, int ys, int wids, int heis);
  void setMaxRect(int rnum, QRect r);
  void setGlobalMaxRect(QRect &refRect, QRect &drawRect);
  void resetCtrlData();
  QRect getScaleRect(QRect norm);
  QRect getScaleRect(int x, int y, int wid, int hei);
  void refreshImg(int x, int y, int wid, int hei);
  void drawToImage(QRect srcRec, QRect &dstRec, bool useFg, bool final, bool useRelativ);
  void drawToImageWithBrush(QRect srcRec, QRect dstRec, bool useRelativ, ApaintBrushList *bList, QImage &img1Bit, Layer *selLayer);
  void refreshIntersectedArea(const QRect &oldRec, const QRect &newRec, int border);
  void resetAndUndo(QRect &srcRec, QRect &dstRec);
  void doFatLines(QImage &img, QPainter *p, int xs, int ys, bool final);

private:

  QImage  *bImg;
  uchar colorIndexFg, colorIndexBg;
  bool docIsLoaded;
  int redfg, greenfg, bluefg;
  int redbg, greenbg, bluebg; 
  int mx1, my1, mx2, my2;
  double zoomX, zoomY;
  bool usingZoom;
  ApaintDoc *doc;
  ApaintView *view;
  QRect oldRefresh, globalMaxRef, globalMaxDraw;
  QPtrList<QRect> maxRectList;
  bool drawFatLines;
};

#endif
