/***************************************************************************
                          dialogcontour.cpp  -  description
                             -------------------
    begin                : Tue Feb 17 2004
    copyright            : (C) 2003-2011 Arahne, d.o.o.
    email                : arahne@arahne.si
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "dialogContour.h"
#include "apaintconfig.h"
#include <qlabel.h>
#include <qpushbutton.h>
#include <qgroupbox.h>
#include <qcheckbox.h>
#include <qtoolbutton.h>
#include <qcombobox.h>
#include <qbuttongroup.h>
#include <qpainter.h>
#include <qdrawutil.h>
#include <qspinbox.h>
#include <qcursor.h>
#include "apaintdoc.h"
#include "dialogColorPalette.h"

extern char *getCommonPNG(const char *pixName);

#define COLOR_MED           381

ColorButton::ColorButton(QWidget *parent, const char *name):QWidget(parent, name)
{
 num = -1;
 col = QColor(red);
 setMinimumSize(sizeHint());
}

QSize ColorButton::sizeHint() const
{
 return QSize(30, 30);
}

void ColorButton::mousePressEvent(QMouseEvent *e)
{
 if (e->button() == RightButton)
  emit rightBtnClicked();
}

void ColorButton::drawNumber(QPainter *p,  int x, int y, int w, int h, int n)
{
  int light = col.red()+col.green()+col.blue();
  if (light<COLOR_MED)
   p->setPen(QPen(white));
  else
   p->setPen(QPen(black));

  p->drawText(x, y, w, h, AlignCenter, QString("%1").arg(n + ApaintConfig::getFirstPaletteColor()));
}

void ColorButton::paintEvent(QPaintEvent *e)
{
  QPainter p(this);
  QBrush b = QBrush(col);
  
  qDrawShadeRect(&p, 0, 0, sizeHint().width(), sizeHint().height(), colorGroup(), false, 2, 0, &b);
  drawNumber(&p, 0, 0, sizeHint().width(), sizeHint().height(), num);
}

void ColorButton::setColor(QRgb c, int n)
{
  col = QColor(c);
  num = n;
  repaint();
}

DialogContour::DialogContour(QWidget *parent, const char *name, ApaintDoc *docp, ColPaletteDlg *colpp):ContourDialogUI(parent, name)
{
  setCaption(ApaintConfig::getText(36));
  btnBack = btnRight->paletteBackgroundColor();
  opType = AddC;
  Q_CHECK_PTR(docp);
  Q_CHECK_PTR(colpp);
  doc = docp;
  colp = colpp;
  cbType->setFocusPolicy(NoFocus);
  lbType->setText(ApaintConfig::getText(626)+":");
  lbCol1->setText(ApaintConfig::getText(627)+":");
  lbCol2->setText(ApaintConfig::getText(666)+":");
  lbSizeX->setText(ApaintConfig::getText(167)+" X");
  lbSizeY->setText(ApaintConfig::getText(167)+" Y");
  btnOK->setText(ApaintConfig::getText(OK_MSG));
  btnClose->setText(ApaintConfig::getText(86));
  gbColors->setTitle(ApaintConfig::getText(COLORS_MSG));
  gbThick->setTitle(ApaintConfig::getText(640));
  bgDirection->setTitle(ApaintConfig::getText(643));
  cbInto->setText(ApaintConfig::getText(670));
  cbAll->setText(ApaintConfig::getText(669));
  cbAll->setChecked(true);
  cbRepeat->setText(ApaintConfig::getText(789));
  cbRepeat->setChecked(true);
  cbShade->setText(ApaintConfig::getText(810));
  cbShade->setChecked(false);
  cbLines->setText(ApaintConfig::getText(688));
  cbLines->setChecked(false);

  cbGradient->setText(ApaintConfig::getText(793));
  cbGradient->setChecked(false);

  cbBidirectional->setText(ApaintConfig::getText(811));
  cbBidirectional->setChecked(false);

  lbAngle->setText(ApaintConfig::getText(50));

  btnUp->setPixmap(QPixmap(getCommonPNG("up")));
  btnDown->setPixmap(QPixmap(getCommonPNG("down")));
  btnLeft->setPixmap(QPixmap(getCommonPNG("left")));
  btnRight->setPixmap(QPixmap(getCommonPNG("right")));
  btnAll->setPixmap(QPixmap(getCommonPNG("arrow-all-8")));
  btnCross->setPixmap(QPixmap(getCommonPNG("arrow-cross")));
  btnDiag->setPixmap(QPixmap(getCommonPNG("arrow-diag")));
  btnLeftUp->setPixmap(QPixmap(getCommonPNG("leftUp")));
  btnLeftDown->setPixmap(QPixmap(getCommonPNG("leftDown")));
  btnRightUp->setPixmap(QPixmap(getCommonPNG("rightUp")));
  btnRightDown->setPixmap(QPixmap(getCommonPNG("rightDown")));  
  btnClear->setPixmap(QPixmap(getCommonPNG("edit-clear")));

  cbType->insertItem(ApaintConfig::getText(ADD_CONT_MSG), AddC);
  cbType->insertItem(ApaintConfig::getText(THICKEN_CONT_MSG), ThickenC);
  connect(cbType, SIGNAL(activated(int)), this, SLOT(typeSelect(int)));  
  connect(btnClose, SIGNAL(clicked()), this, SLOT(closeDlg()));
  connect(btnOK, SIGNAL(clicked()), this, SLOT(doContourOp()));
  connect(bgDirection, SIGNAL(clicked(int)), this, SLOT(directionSelect(int)));
  connect(colAround, SIGNAL(rightBtnClicked()), this, SLOT(aroundBtnClicked()));
  connect(colWith, SIGNAL(rightBtnClicked()), this, SLOT(withBtnClicked()));
  connect(cbAll, SIGNAL(toggled(bool)), this, SLOT(setWithColor(bool)));  
  connect(docp, SIGNAL(newDocSig()), this, SLOT(newDocSlot()));

  connect(cbGradient, SIGNAL(toggled(bool)), this, SLOT(setThickOptions(bool)));  
  connect(cbRepeat, SIGNAL(toggled(bool)), this, SLOT(setThickOptions(bool)));  

  connect(sbThickX, SIGNAL(valueChanged(int)), this, SLOT(showThickOptions()));
  connect(sbThickY, SIGNAL(valueChanged(int)), this, SLOT(showThickOptions()));

  directionSelect(bgDirection->id(btnAll));
  setupDlg();
  showThickOptions();

  // init color buttons
}

DialogContour::~DialogContour()
{
}

void DialogContour::newDocSlot()
{
  aroundBtnClicked();
  withBtnClicked();  
}

void DialogContour::bgFgChangedSlot(int fgn, QRgb fgc, int bgn, QRgb bgc)
{
 if (bgc != colAround->getColor())
  aroundBtnClicked();

 if (bgc != colWith->getColor())
  withBtnClicked(); 
}

void DialogContour::aroundBtnClicked()
{
 int n;

 QRgb c = colp->getSelColor(n);
 colAround->setColor(c, n);
}

void DialogContour::withBtnClicked()
{
 int n;

 QRgb c = colp->getSelColor(n);
 colWith->setColor(c, n);
}

void DialogContour::enableAroundColor(bool enable)
{
 if (enable)
  {
   lbCol1->show();
   colAround->show();
   cbInto->show();
  }
 else
  {
   lbCol1->hide();
   colAround->hide();
   cbInto->hide(); 
  }
}

void DialogContour::setWithColor(bool disable)
{
 if (disable)
  {
   lbCol2->hide();
   colWith->hide();
  }
 else
  {
   lbCol2->show();
   colWith->show();
  }
}

void DialogContour::enableWithColor(bool enable)
{
 if (enable)
  {
   cbAll->show();
   setWithColor(cbAll->isChecked());
  }
 else
  {
   cbAll->hide();
   setWithColor(true);
  }
}

void DialogContour::setupColors()
{
 if (opType == AddC)
  {
   enableAroundColor(true);
   enableWithColor(true);
  }
 else
  {
   enableAroundColor(false);
   enableWithColor(false);
  }
}

void DialogContour::setThickOptions(bool enable)
{
 showThickOptions();
}

void DialogContour::showThickOptions()
{
 if ((sbThickX->value()>1)||(sbThickX->value()>1))
  {
   cbRepeat->show();
   cbShade->show();
   if (cbRepeat->isChecked())
    cbLines->hide();
   else
    cbLines->show();
   cbGradient->show();
   if (cbGradient->isChecked())
    {
     sbAngle->show();
     lbAngle->show();
     cbBidirectional->show();
    }
   else
    {
     sbAngle->hide();
     lbAngle->hide();
     cbBidirectional->hide();
    }
  }
 else
  {
   cbRepeat->hide();
   cbShade->hide();
   cbLines->hide();
   cbGradient->hide();
   sbAngle->hide();
   lbAngle->hide();
   cbBidirectional->hide();
  }
}

void DialogContour::setupDlg()
{
 setupColors();

 if (opType == AddC || opType == ThickenC)
  {
   bgDirection->show();
   gbThick->show();
  }
 else
  {
   bgDirection->hide();
   gbThick->hide();
  }

 if (opType == AddC)
  gbColors->show();
 else
  gbColors->hide();

 if (opType == ThickenC)
  cbRepeat->show();
 else
  cbRepeat->hide();
}

void DialogContour::typeSelect(int type)
{
  opType = (ContourOp) type;
  setupDlg();
}

void DialogContour::showDlg()
{
 int n;
 QRgb c;

 c = colp->getSelColor(n);
 colAround->setColor(c, n);
 c = colp->getSelColor(n);
 colWith->setColor(c, n);
 show();
}

void DialogContour::closeDlg()
{
  close();
}

void DialogContour::directionSelect(int btn)
{
 QPushButton *which =(QPushButton *)bgDirection->find(btn);
 int bit=CDIR_NONE;
 QColor btnFront = QColor(210, 200, 10);

 if (which == btnAll)
  {
   direction = CDIR_ALL;
   btnLeft->setPaletteBackgroundColor(btnFront);
   btnRight->setPaletteBackgroundColor(btnFront);
   btnUp->setPaletteBackgroundColor(btnFront);
   btnDown->setPaletteBackgroundColor(btnFront);
   btnLeftUp->setPaletteBackgroundColor(btnFront);
   btnLeftDown->setPaletteBackgroundColor(btnFront);
   btnRightUp->setPaletteBackgroundColor(btnFront);
   btnRightDown->setPaletteBackgroundColor(btnFront);
  }
 else if (which == btnClear)
  {
   direction = CDIR_NONE;
   btnLeft->setPaletteBackgroundColor(btnBack);
   btnRight->setPaletteBackgroundColor(btnBack);
   btnUp->setPaletteBackgroundColor(btnBack);
   btnDown->setPaletteBackgroundColor(btnBack);
   btnLeftUp->setPaletteBackgroundColor(btnBack);
   btnLeftDown->setPaletteBackgroundColor(btnBack);
   btnRightUp->setPaletteBackgroundColor(btnBack);
   btnRightDown->setPaletteBackgroundColor(btnBack);
  }
 else if (which == btnCross)
  {
   direction = CDIR_LEFT | CDIR_RIGHT | CDIR_UP | CDIR_DOWN;
   btnLeft->setPaletteBackgroundColor(btnFront);
   btnRight->setPaletteBackgroundColor(btnFront);
   btnUp->setPaletteBackgroundColor(btnFront);
   btnDown->setPaletteBackgroundColor(btnFront);
   btnLeftUp->setPaletteBackgroundColor(btnBack);
   btnLeftDown->setPaletteBackgroundColor(btnBack);
   btnRightUp->setPaletteBackgroundColor(btnBack);
   btnRightDown->setPaletteBackgroundColor(btnBack);
  }
 else if (which == btnDiag)
  {
   direction =  CDIR_LEFT_UP | CDIR_LEFT_DOWN | CDIR_RIGHT_UP | CDIR_RIGHT_DOWN;
   btnLeft->setPaletteBackgroundColor(btnBack);
   btnRight->setPaletteBackgroundColor(btnBack);
   btnUp->setPaletteBackgroundColor(btnBack);
   btnDown->setPaletteBackgroundColor(btnBack);
   btnLeftUp->setPaletteBackgroundColor(btnFront);
   btnLeftDown->setPaletteBackgroundColor(btnFront);
   btnRightUp->setPaletteBackgroundColor(btnFront);
   btnRightDown->setPaletteBackgroundColor(btnFront);
  }
 else
  {
   if (which == btnLeft)
    bit = CDIR_LEFT;
   else if (which == btnRight)
    bit = CDIR_RIGHT;
   else if (which == btnUp)
    bit = CDIR_UP;
   else if (which == btnDown)
    bit = CDIR_DOWN;
   else if (which == btnLeftUp)
    bit = CDIR_LEFT_UP;
   else if (which == btnLeftDown)
    bit = CDIR_LEFT_DOWN;
   else if (which == btnRightUp)
    bit = CDIR_RIGHT_UP;
   else if (which == btnRightDown)
    bit = CDIR_RIGHT_DOWN;

   if (direction & bit)
    direction &= ~bit;
   else
    direction |= bit;
   which->setPaletteBackgroundColor((direction & bit) ? btnFront : btnBack);
  }
}

void DialogContour::doContourOp()
{
 int selCol;

 colp->getSelColor(selCol);
 QApplication::setOverrideCursor( QCursor(Qt::WaitCursor) );

 doc->putUndoStack(ChangeLayerForm);
 switch (opType)
  {
   case AddC:
    if (selCol == colAround->getIndex())
     doc->layerContourThick(selCol, direction, sbThickX->value(),  sbThickY->value(), sbAngle->value(), cbShade->isChecked(), cbLines->isChecked(), cbGradient->isChecked(), cbBidirectional->isChecked(), cbRepeat->isChecked());
    else
     doc->layerContourAdd(selCol, colAround->getIndex(), colWith->getIndex(), !cbAll->isChecked(), direction, cbInto->isChecked(), cbShade->isChecked(), sbThickX->value(), sbThickY->value());
   break;
   case ThickenC:
    doc->layerContourThick(selCol, direction, sbThickX->value(), sbThickY->value(), sbAngle->value(), cbShade->isChecked(), cbLines->isChecked(), cbGradient->isChecked(), cbBidirectional->isChecked(), cbRepeat->isChecked());
   break;  
  }
 QApplication::restoreOverrideCursor();
}


