/***************************************************************************
                          dialogGuessWeave.cpp  -  description
                             -------------------
    begin                : Tue Mar 23 2004
    copyright            : (C) 2003-2011 Arahne, d.o.o.
    email                : arahne@arahne.si
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "dialogGuessWeave.h"
#include <qpushbutton.h>
#include <qlabel.h>
#include <qspinbox.h>
#include <stdlib.h>
#include <math.h>
#include "apaintconfig.h"
#include "apaintdoc.h"

#define MIN_GRID  4

GuessWeaveDialog::GuessWeaveDialog(QWidget *parent, ApaintDoc *docp, const char *name):GuessWeaveDialogUI(parent, name)
{
  setCaption(ApaintConfig::getText(195));
  lbWidth->setText(ApaintConfig::getText(23)+":");
  lbHeight->setText(ApaintConfig::getText(22)+":");
  btnOK->setText(ApaintConfig::getText(OK_MSG));
  btnCancel->setText(ApaintConfig::getText(CANCEL_MSG));
  btnPreview->setText(ApaintConfig::getText(158));
  Q_CHECK_PTR(docp);
  doc = docp;
  minSquareX = minSquareY = MIN_GRID;
  connect(btnCancel, SIGNAL(clicked()), this, SLOT(closeDlg()));
  connect(btnPreview, SIGNAL(clicked()), this, SLOT(doPreview()));
  connect(btnOK, SIGNAL(clicked()), this, SLOT(guessWeaveOK()));
  connect(doc, SIGNAL(layerChanged(Layer*)), this, SLOT(layerChangedSlot(Layer*)));
  connect(sbWidth, SIGNAL(valueChanged(int)), this, SLOT(widHeiChanged(int)));
  connect(sbHeight, SIGNAL(valueChanged(int)), this, SLOT(widHeiChanged(int)));
  previewMode = false;
  dlgIsOn = false;
}

GuessWeaveDialog::~GuessWeaveDialog()
{

}

void GuessWeaveDialog::layerChangedSlot(Layer* layer)
{
  if (dlgIsOn)
   setupDlg();
}

void GuessWeaveDialog::showDlg()
{
 dlgIsOn = true;
 setupDlg(); 
 doc->protectLayer(true);
 show(); 
}

void GuessWeaveDialog::setupDlg()
{
  int guessX, guessY;

  guessX = guessY = 0;
  Layer *selLayer = doc->getLayer();
  if (!selLayer) return;

  if (selLayer->getImageType() == LOAD_TRUE_COL)
  {
    btnOK->setEnabled(false);
    btnPreview->setEnabled(false);
  }
  else if (selLayer->getImageType() == LOAD_CMAP)
  {
   guessWeave(selLayer->imageBuf, guessX, guessY);
   btnOK->setEnabled(true);
   btnPreview->setEnabled(true);
  }
  sbWidth->setValue(guessX);
  sbHeight->setValue(guessY);   
}

void GuessWeaveDialog::widHeiChanged(int val)
{
  if (previewMode)
    guessPreview();
}

void GuessWeaveDialog::closeDlg()
{
  close();
}

void GuessWeaveDialog::doOnClose()
{
  dlgIsOn = false;
  doc->protectLayer(false);
  if (previewMode)
   doUndo(); 
}

void GuessWeaveDialog::closeEvent(QCloseEvent *e)
{
  doOnClose();
  GuessWeaveDialogUI::closeEvent(e);
}

void GuessWeaveDialog::hideEvent(QHideEvent *e)
{
  doOnClose();
  GuessWeaveDialogUI::hideEvent(e);
}

void GuessWeaveDialog::doUndo()
{
  Layer *selLayer = doc->getLayer();
  if (!selLayer) return;
  
  if (previewMode)
  {
    previewMode = false;
    btnPreview->setText(ApaintConfig::getText(158));
    selLayer->undoPreview();
    doc->refreshLayer();
  }
}

void GuessWeaveDialog::doPreview()
{
  if (previewMode)
   doUndo();
  else
  {
    previewMode = true;
    btnPreview->setText(ApaintConfig::getText(26));
    guessPreview();
  }
}

void GuessWeaveDialog::guessWeave(ApaintImage &img, int &gx, int &gy)
{
 int swidth, sheight;
 int i,j,k, d1,d2,d3, foundX, foundY;
 int maxwid, maxhei, bpl;
 int maxSize = QMAX(img.width(), img.height());
 unsigned short diffTable[0xFFFF]; /* for pre-calculation of color differences of all color pairs */
 unsigned char *data;

 if (img.type() == LOAD_TRUE_COL)
  {
   gx = gy = 0;
   return;
  }

 unsigned short *midIndex = new ushort[maxSize];
 unsigned short *gridX = new ushort[maxSize / MIN_GRID];
 unsigned long *gridError = new unsigned long[maxSize / MIN_GRID];
 
 QRgb *colTab = img.colorTable();
 swidth = img.width();
 sheight = img.height();
 
 for (i=0;i<img.numColors();i++)
  for (j=0;j<img.numColors();j++)
   { /* calculate the difference based on allocCells, so we do not do it in the main loop! */
    d1=qRed(colTab[i]) - qRed(colTab[j]);
    d2=qGreen(colTab[i]) - qGreen(colTab[j]);
    d3=qBlue(colTab[i]) - qBlue(colTab[j]);
    diffTable[(i<<8)+j] = (unsigned short) sqrt(d1*d1+d2*d2+d3*d3);
   }
 maxwid = swidth/minSquareX;
 maxhei = sheight/minSquareY;
 memset(gridError, 0, sizeof(unsigned long)*maxwid);
 bpl = img.bpl();
 /* find the horizontal grid */
 /* testing hypothesis of grid width k */
 for (k=MIN_GRID; k<maxwid; k++)
  {
   for (i=0; i<swidth; i++)
    midIndex[i]= (unsigned short) (((i*k)/swidth)*((swidth*1.0)/k)+((swidth*1.0)/(k*2)));
   for (j=0; j<sheight; j++)
    {
     data = img.scanLine(j);
     for (i=swidth-1; i>=0; i--)
      gridError[k]+= diffTable[(((int)data[i])<<8)+data[midIndex[i]]];
    }
  }
  /* search for horizontal minimum */
  foundX=MIN_GRID;
  for (i=MIN_GRID; i<maxwid; i++)
   if (gridError[foundX]>gridError[i])
    foundX=i;
  /* now find vertical grid, but look only at points at center of horizontal grid */
  for (i=0; i<foundX; i++)
   gridX[i]= (unsigned short) ((swidth*i*1.0)/foundX + (swidth*1.0)/(foundX*2));
  memset(gridError, 0, sizeof(unsigned long)*maxhei);
  /* find the vertical grid */
  /* testing hypothesis of grid width k */
  for (k=MIN_GRID; k<maxhei; k++)
   {
    for (i=0; i<sheight; i++)
     midIndex[i]= (unsigned short) (((i*k)/sheight)*((sheight*1.0)/k)+((sheight*1.0)/(k*2)));
    for (j=0; j<foundX; j++)
     {
      data = img.scanLine(0)+gridX[j];
      for (i=sheight-1; i>=0; i--)
       gridError[k]+= diffTable[(((int)data[i*bpl])<<8)+data[midIndex[i]*bpl]];
     }
   }
  /* search for vertical minimum */
  foundY=MIN_GRID;
  for (i=MIN_GRID; i<maxhei; i++)
   if (gridError[foundY]>gridError[i])
    foundY=i;
  gx = foundX;
  gy = foundY;
  delete[] midIndex;
  delete[] gridX;
  delete[] gridError;
}


void GuessWeaveDialog::guessPreview()
{
 Layer *selLayer = doc->getLayer();

 if (!selLayer)
  return;
 if (selLayer->getImageType() == LOAD_TRUE_COL)
  return;
 /* do undo first */
 selLayer->undoPreview();
 selLayer->setPreviewMode(true);
 QRect br = selLayer->layerPos();
 
 int sx, sy, swidth, sheight, foundX, foundY;
 int i,j;
 int maxSize = QMAX(br.width(), br.height());
  int gridOffX, gridLenX, gridOffY, gridLenY;

 foundX = sbWidth->value();
 foundY = sbHeight->value();
 swidth = br.width();
 sheight = br.height();
 sx = 0;
 sy = 0;
 
 if ((foundX*MIN_GRID>swidth)||(foundY*MIN_GRID>sheight))
  return;

 ushort *gridX = new ushort[maxSize / MIN_GRID];
 ushort *gridY = new ushort[maxSize / MIN_GRID];

 /* it is useful to still see the grid in the preview, so only draw squares of odd size */
 gridLenX=swidth/foundX;
 if (!(gridLenX&1))
  gridLenX--;
 gridOffX=gridLenX/2;
 gridLenY=sheight/foundY;
 if (!(gridLenY&1))
  gridLenY--;
 gridOffY=gridLenY/2;

 for (i=0; i<foundX; i++)
  gridX[i]= (unsigned short) ((swidth*i*1.0)/foundX + (swidth*1.0)/(foundX*2));
 for (i=0; i<foundY; i++)
  gridY[i]= (unsigned short) ((sheight*i*1.0)/foundY + (sheight*1.0)/(foundY*2));
 for (i=0; i<foundX; i++)
  for (j=0; j<foundY; j++)
   selLayer->imageBuf.fill(sx+gridX[i]-gridOffX, sy+gridY[j]-gridOffY, gridLenX, gridLenY,
                           selLayer->imageBuf.pixelIndex(gridX[i], gridY[j]), false, 0);     
 delete[] gridX;
 delete[] gridY;
 doc->refreshLayer();
}

void GuessWeaveDialog::guessWeaveOK()
{
 Layer *selLayer = doc->getLayer();
 if (!selLayer) return;
 if (selLayer->getImageType() == LOAD_TRUE_COL) return;

 doUndo();
 
 QRect br = selLayer->layerPos();
 int swidth, sheight, foundX, foundY;
 int i,j, bpl, bpl2;
 int maxSize = QMAX(br.width(), br.height());
 unsigned char *data, *data2;
 ApaintImage img2;
 
 foundX = sbWidth->value();
 foundY = sbHeight->value();
 swidth = br.width();
 sheight = br.height();
 
 if ((foundX*MIN_GRID>swidth)||(foundY*MIN_GRID>sheight)) return;
 
 ushort *gridX = new ushort[maxSize / MIN_GRID];
 ushort *gridY = new ushort[maxSize / MIN_GRID];

 minSquareX = max(MIN_GRID, swidth/foundX-1); /* so it will try to find it in the same range on the next guess */
 minSquareY = max(MIN_GRID, sheight/foundY-1);

 /* the new destination selection image */
 img2.newImage(foundX, foundY, selLayer->imageBuf.depth());
 selLayer->imageBuf.copyImageDataTo(&img2);

 bpl = selLayer->imageBuf.bpl();
 data = selLayer->imageBuf.scanLine(0);
 bpl2 = img2.bpl();;
 data2 = img2.scanLine(0);

 for (i=0; i<foundX; i++)
  gridX[i]= (unsigned short) ((swidth*i*1.0)/foundX + (swidth*1.0)/(foundX*2));
 for (i=0; i<foundY; i++)
  gridY[i]= (unsigned short) ((sheight*i*1.0)/foundY + (sheight*1.0)/(foundY*2));
 for (i=0; i<foundX; i++)
  for (j=0; j<foundY; j++)
   data2[bpl2*j+i] = data[bpl*gridY[j]+gridX[i]];

 
 QImage mask = QImage(img2.width(), img2.height(), 1, 2, QImage::LittleEndian);
 mask.fill(1);
 doc->putUndoStack(CreateLayer);
 doc->createNewLayer(img2, &mask, br.x(), br.y());
 close();  
 delete[] gridX;
 delete[] gridY;
}


