/***************************************************************************
                          dialogPrint.cpp  -  description
                             -------------------
    begin                : Wed May 19 2004
    copyright            : (C) 2003-2011 Arahne, d.o.o.
    email                : arahne@arahne.si
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "dialogPrint.h"
#include "apaintdoc.h"
#include "apaintconfig.h"
#include <qlabel.h>
#include <qcheckbox.h>
#include <qpushbutton.h>
#include <qgroupbox.h>
#include <qspinbox.h>
#include <qlineedit.h>
#include <qcombobox.h>
#include <qmessagebox.h>
#include <qprogressbar.h>
#include "apaintUtil.h"
#include "apaintimageUtil.h"
#include "apaintconfig.h"
#include "dialogPreview.h"
#include "dialogRepeatDimen.h"

#define PRINT_CAP_PATH_1 "/etc/printcap"
#define PRINT_CAP_PATH_2 "/etc/printcap.cups"
#define PRINT_CAP_PATH_3 "/etc/cups/printcap"
#define PRINT_CAP_PATH_4 "/var/run/cups/printcap"
/*
#define LEFT_MARGIN_MM 3
#define TOP_MARGIN_MM    3
#define A4_WIDTH_MM      203
#define A4_HEIGHT_MM     279
*/
#define TITLE_HEI_MM     7

#ifdef __APPLE__ 
extern char *parseOneLineTagXML(char *line, char *tag);
#endif

ShowProgressDialog::ShowProgressDialog(QWidget *parent, const char *name, QPtrList<PrintDataProducer> &pdpList):
ShowProgressDialogUI(parent, name)
{
  PrintDataProducer *pdp;
  int allHei= 0;
  
  allSteps = 0;
  doCancel = false;
  setCaption(ApaintConfig::getText(130));
  btnCancel->setText(ApaintConfig::getText(5));
  for (int i=0; i<(int)pdpList.count(); i++)
  {
    pdp = pdpList.at(i);
    connect(pdp, SIGNAL(setProgress(PrintDataProducer *)), this, SLOT(updateProgress(PrintDataProducer *)));
    allHei+=pdp->pageHeight();
  }
  pbar->setPercentageVisible(true);
  pbar->setTotalSteps(allHei);
  pbar->reset();
  connect(btnCancel, SIGNAL(clicked()), this, SLOT(cancelClicked()));
}

ShowProgressDialog::~ShowProgressDialog()
{
  
}

void ShowProgressDialog::updateProgress(PrintDataProducer *pdp)
{
  if (doCancel)
  {
    pdp->setPrtStatus(PRT_USER_ABORT);
    return;
  }
  allSteps+=pdp->curLine();
  pbar->setProgress(allSteps);
  qApp->processEvents(); // check btn click
}

void ShowProgressDialog::cancelClicked()
{
  doCancel = true; 
}

PrinterSelectDialog::PrinterSelectDialog(QWidget *parent, const char *name, ApaintDoc *pdoc):PrinterSelectDialogUI(parent, name)
{
  setCaption(ApaintConfig::getText(79));
  lbPrinter->setText(ApaintConfig::getText(132));
  lbPrintMode->setText(ApaintConfig::getText(133));
  lbDensityX->setText(ApaintConfig::getText(134));
  lbDensityY->setText(ApaintConfig::getText(135));
  cbPrintToFile->setText(ApaintConfig::getText(137));
  btnClose->setText(ApaintConfig::getText(5));
  btnOK->setText(ApaintConfig::getText(3));
  cbDir->insertItem("/tmp", 0);
  Q_CHECK_PTR(pdoc);
  doc = pdoc;
  printerIndx = modeIndx = denXIndx = denYIndx =0;  
  connect(btnClose, SIGNAL(clicked()), this, SLOT(cancelBtnclicked()));
  connect(btnOK, SIGNAL(clicked()), this, SLOT(okBtnClicked()));
  connect(cbPrintToFile, SIGNAL(toggled(bool)), this, SLOT(setPrintToFile(bool)));
  connect(listPrinter, SIGNAL(highlighted(int)), this, SLOT(printerChanged(int)));
  connect(listPrintMode, SIGNAL(highlighted(int)), this, SLOT(modeChanged(int)));
  connect(listDensityX, SIGNAL(highlighted(int)), this, SLOT(denXChanged(int)));
  connect(listDensityY, SIGNAL(highlighted(int)), this, SLOT(denYChanged(int)));
  addPrinters();
  setPrinterName(ApaintConfig::getPrinterType());
  
}

PrinterSelectDialog::~PrinterSelectDialog()
{
  
}

void PrinterSelectDialog::setPrinterName(QString name)
{
  int indx = printerIndx;
  for (int i=0; i<(int)listPrinter->count(); i++)
   if (listPrinter->text(i) == name)
    indx = i;

  listPrinter->setCurrentItem(indx);
  updateData();
}

void PrinterSelectDialog::showDlg()
{  
  show();
}

void PrinterSelectDialog::cancelBtnclicked()
{
  close();
  listPrinter->setCurrentItem(printerIndx);
  listPrintMode->setCurrentItem(modeIndx);
  listDensityX->setCurrentItem(denXIndx);
  listDensityY->setCurrentItem(denYIndx);
}

void PrinterSelectDialog::updateData()
{
  printerIndx = listPrinter->currentItem();
  modeIndx = listPrintMode->currentItem();
  denXIndx = listDensityX->currentItem();
  denYIndx = listDensityY->currentItem();
  emit dataChanged(); 
}

void PrinterSelectDialog::okBtnClicked()
{
  if (saveToFile() && edFilename->text()=="")
  {
    QMessageBox::information(this, ApaintConfig::getText(132),ApaintConfig::getText(709) , ApaintConfig::getText(3));
    edFilename->setFocus();
    return;
  }
  updateData();
  close();
}

bool PrinterSelectDialog::saveToFile()
{
  return cbPrintToFile->isChecked();
}

void PrinterSelectDialog::setPrintToFile(bool isOn)
{
  cbDir->setEnabled(isOn);
  edFilename->setEnabled(isOn);
}

void PrinterSelectDialog::setFileName(QString& name)
{
 edFilename->setText(name);
}

bool PrinterSelectDialog::getFileName(QString& name)
{

  if (edFilename->text() == "")
   return false;
   
  switch(cbDir->currentItem())
  {
    case 0:
     name = QString("/tmp/%1").arg(edFilename->text());
    break;
  }
  return true;
}

void PrinterSelectDialog::printerChanged(int i)
{
  edPrinter->setText(listPrinter->currentText());
  addPrtModes(i);
  addDensityX(i);
  addDensityY(i);
  if (i<4)
   {
    cbPrintToFile->setChecked(true);
    cbPrintToFile->setEnabled(false);
   }
  else
   {
    cbPrintToFile->setChecked(false);
    cbPrintToFile->setEnabled(true);
   }
}

void PrinterSelectDialog::modeChanged(int i)
{
  edPrintMode->setText(listPrintMode->currentText()); 
}

void PrinterSelectDialog::denXChanged(int i)
{
  edDensityX->setText(listDensityX->currentText()); 
}

void PrinterSelectDialog::denYChanged(int i)
{
  edDensityY->setText(listDensityY->currentText());  
}


void PrinterSelectDialog::addPrinters()
{
  listPrinter->clear();
  ApaintImageIO *io = ApaintImageIO::getObject();
  for (int i=0; i<io->getNumPrinters(); i++)
    listPrinter->insertItem(QString("%1").arg(io->getPrinterName(i)));

}

void PrinterSelectDialog::addPrtModes(int i)
{
  int num;
  listPrintMode->clear();
  ApaintImageIO *io = ApaintImageIO::getObject();
  printerMode *prtMode = io->getPrinterModes(i, &num);
  for (int i=0; i<num; i++)
    switch (prtMode[i].printType)
    {
     case IMG_BW:
      listPrintMode->insertItem(ApaintConfig::getText(139), i);
     break;
     case IMG_CMAP:
      listPrintMode->insertItem(ApaintConfig::getText(152), i);
     break;
     case IMG_TRUE_COL:
      listPrintMode->insertItem(ApaintConfig::getText(141), i);
     break;
     case IMG_TRUE_COL_DITH:
      listPrintMode->insertItem(ApaintConfig::getText(140), i);
     break;
     
    }
    listPrintMode->setCurrentItem(0);
}

void PrinterSelectDialog::addDensityX(int i)
{
  int num;
  listDensityX->clear();
  ApaintImageIO *io = ApaintImageIO::getObject();
  printerDensity *prtDenX = io->getPrtDensityX(i, &num);
  for (int i=0; i<num; i++)
   listDensityX->insertItem(QString("%1").arg(prtDenX[i].dpi/1000), i);
  listDensityX->setCurrentItem(0);
}

void PrinterSelectDialog::addDensityY(int i)
{
  int num;
  listDensityY->clear();
  ApaintImageIO *io = ApaintImageIO::getObject();
  printerDensity *prtDenY = io->getPrtDensityY(i, &num);
  for (int i=0; i<num; i++)
   listDensityY->insertItem(QString("%1").arg(prtDenY[i].dpi/1000), i);
  listDensityY->setCurrentItem(0);
}

void PrinterSelectDialog::getPrinterLimitWidMM(int &left, int &wid)
{
 int num, pix, dpi, l;
 ApaintImageIO *io = ApaintImageIO::getObject();
 l = io->getNonPrintableLeft(printerIndx);
 printerDensity *prtDenX = io->getPrtDensityX(printerIndx, &num);
 pix = prtDenX[denXIndx].pixels;
 dpi = prtDenX[denXIndx].dpi/1000;
 wid = (int)(((double)pix / (double)dpi)*25.4);
 left = (int)(((double)l / (double)dpi)*25.4);

}

void PrinterSelectDialog::getPrinterLimitHeiMM(int &top, int &hei)
{
 int num, pix, dpi, t;

 ApaintImageIO *io = ApaintImageIO::getObject();
 t = io->getNonPrintableTop(printerIndx);
 printerDensity *prtDenY = io->getPrtDensityY(printerIndx, &num);
 pix = prtDenY[denYIndx].pixels;
 dpi = prtDenY[denYIndx].dpi/1000;
 hei = (int)(((double)pix / (double)dpi)*25.4);
 top = (int)(((double)t / (double)dpi)*25.4);
}


/******************PrintDialog***************************/

PrintDialog::PrintDialog(QWidget *parent, const char *name, ApaintDoc *pdoc):PrintDialogUI(parent, name)
{
  pageLeft = pageTop = pageWidth = pageHeight =0;
  numPrints = 0;
  printer = printMode = 0;
  densityX = densityY = 100;
  setCaption(ApaintConfig::getText(16));  
  lbPrinter0->setText(ApaintConfig::getText(132)+":");
  lbPrintMode0->setText(ApaintConfig::getText(133)+":");
  lbDensityX0->setText(ApaintConfig::getText(134)+":");
  lbDensityY0->setText(ApaintConfig::getText(135)+":");
  btnSetup->setText(ApaintConfig::getText(79)+"...");
  if (ApaintConfig::useMetricSystem())
  {
   lbLeftBorder->setText(ApaintConfig::getText(119)+"(mm):");
   lbUpBorder->setText(ApaintConfig::getText(120)+"(mm):");
   lbPrintWid->setText(ApaintConfig::getText(121)+"(mm):");
   lbPrintHei->setText(ApaintConfig::getText(122)+"(mm):");
  }
  else
  {
   lbLeftBorder->setText(ApaintConfig::getText(119)+"(inch):");
   lbUpBorder->setText(ApaintConfig::getText(120)+"(inch):");
   lbPrintWid->setText(ApaintConfig::getText(121)+"(inch):");
   lbPrintHei->setText(ApaintConfig::getText(122)+"(inch):");
  }  
  btnFullPage->setText(ApaintConfig::getText(200));
  lbZoom->setText(ApaintConfig::getText(96)+"(%):");
  cbFitToPage->setText(ApaintConfig::getText(76));
  cbOneRepeat->setText(ApaintConfig::getText(716));
  lbRepeatWid0->setText(ApaintConfig::getText(105)+":");
  lbRepeatHei0->setText(ApaintConfig::getText(106)+":");
  lbNumPrints->setText(ApaintConfig::getText(201)+":");
  cbTitle->setText(ApaintConfig::getText(715));
  cbLandscape->setText(ApaintConfig::getText(77));
  btnPreview->setText(ApaintConfig::getText(158)+"...");
  btnClose->setText(ApaintConfig::getText(5));
  btnOK->setText(ApaintConfig::getText(3));
  gbPrinterInfo->setTitle(ApaintConfig::getText(714));
  gbPrintSize->setTitle(ApaintConfig::getText(78));
  cbUsedCol->setText(ApaintConfig::getText(17));
  lbFilename0->setText(ApaintConfig::getText(137)+":");
  
  /* connections*/
  connect(edPageWid, SIGNAL(textChanged(const QString &)), this, SLOT(pwidChanged(const QString &)));
  connect(edPageHei, SIGNAL(textChanged(const QString &)), this, SLOT(pheiChanged(const QString &)));
  connect(sbZoom, SIGNAL(valueChanged(int)), this, SLOT(zoomChanged(int)));
  connect(sbNumPrints, SIGNAL(valueChanged(int)), this, SLOT(setNumPrints(int)));
  connect(btnSetup, SIGNAL(clicked()), this, SLOT(showSetup()));

  connect(btnOK, SIGNAL(clicked()), this, SLOT(doPrint()));
  connect(btnPreview, SIGNAL(clicked()), this, SLOT(doPreview()));
  connect(btnClose, SIGNAL(clicked()), this, SLOT(closeDlg()));
  connect(btnFullPage, SIGNAL(clicked()), this, SLOT(setFullPage()));
  connect(cbFitToPage, SIGNAL(toggled(bool)), this, SLOT(fitToPageClicked(bool)));
  connect(cbOneRepeat, SIGNAL(toggled(bool)), this, SLOT(oneRepeatClicked(bool)));
  connect(cbLandscape, SIGNAL(clicked()), this, SLOT(calcRepeatInfo()));
  /**/  
  Q_CHECK_PTR(pdoc);
  doc = pdoc;
  pselDlg = new PrinterSelectDialog(this, "Printer select dialog", doc);
  Q_CHECK_PTR(pselDlg);
  connect(pselDlg, SIGNAL(dataChanged()), this, SLOT(setPrinterData()));
  readPrintcapFile(pcapList);
  cbType->insertStringList(pcapList);
  setPrintCap(ApaintConfig::getPrintCap());
  previewDlg = new PreviewDialog(this, "Print preview");
  Q_CHECK_PTR(previewDlg);
  setPageLeftMM(ApaintConfig::getLeftBorderMM());
  setPageUpMM(ApaintConfig::getUpperBorderMM());
  setPageWidMM(ApaintConfig::getPrintWidthMM());
  setPageHeiMM(ApaintConfig::getPrintHeightMM());
}

PrintDialog::~PrintDialog()
{
  
}

//wid and hei are in mm
void PrintDialog::setPageSizeMinMax(int wid, int hei)
{
 
 if (ApaintConfig::useMetricSystem())
  {
    ApaintIntValidator *intValWid, *intValHei;

    intValWid = (ApaintIntValidator *) edPageLeft->validator();
    if (intValWid) delete intValWid;
    intValHei = (ApaintIntValidator *) edPageUp->validator();
    if (intValHei) delete intValHei;
    
    intValWid = new ApaintIntValidator(0, wid, 0, "int validator");
    intValHei = new ApaintIntValidator(0, hei, 0, "int validator");
    edPageLeft->setValidator(intValWid);
    if (edPageLeft->text().toInt()>wid)
      edPageLeft->setText(QString("%1").arg(wid));
    edPageUp->setValidator(intValHei);
    if (edPageUp->text().toInt()>hei)
      edPageUp->setText(QString("%1").arg(hei));
    edPageWid->setValidator(intValWid);
    if (edPageWid->text().toInt()>wid)
      edPageWid->setText(QString("%1").arg(wid));
    edPageHei->setValidator(intValHei);
    if (edPageHei->text().toInt()>hei)
      edPageHei->setText(QString("%1").arg(hei));
  }
  else
  {
    double dwid, dhei;
    dwid = (double)wid/24.5;
    dhei = (double)hei/24.5;
    
    ApaintDoubleValidator *doubleValWid, *doubleValHei;
    
    doubleValWid = (ApaintDoubleValidator *) edPageLeft->validator();
    if (doubleValWid) delete doubleValWid;
    doubleValHei = (ApaintDoubleValidator *) edPageUp->validator();
    if (doubleValHei) delete doubleValHei;
    
    doubleValWid = new ApaintDoubleValidator(0, dwid, 2, 0, "double validator");
    doubleValHei = new ApaintDoubleValidator(0, dhei, 2, 0, "double validator");
    edPageLeft->setValidator(doubleValWid);
    if (edPageLeft->text().toDouble()>dwid)
      edPageLeft->setText(QString("%1").arg(dwid));
    edPageUp->setValidator(doubleValHei);
    if (edPageUp->text().toDouble()>dhei)
      edPageUp->setText(QString("%1").arg(dhei));
    edPageWid->setValidator(doubleValWid);
    if (edPageWid->text().toDouble()>dwid)
      edPageWid->setText(QString("%1").arg(dwid));
    edPageHei->setValidator(doubleValHei);
    if (edPageHei->text().toDouble()>dhei)
      edPageHei->setText(QString("%1").arg(dhei));
  } 
}

void PrintDialog::setPrintCap(QString pcap)
{
  int indx = 0;
  for (int i=0; i < cbType->count(); i++)
  {
    if (cbType->text(i) == pcap)
    {
     indx = i;
     break;
    }
  }
  cbType->setCurrentItem(indx);
}

double PrintDialog::getPageLeftMM()
{
  if (edPageLeft->text() == "") return 0;
  double val = edPageLeft->text().toDouble();
  if (!ApaintConfig::useMetricSystem())
   val*=25.4;
  return val;
}

double PrintDialog::getPageUpMM()
{
  if (edPageUp->text() == "") return 0;
  double val = edPageUp->text().toDouble();
  if (!ApaintConfig::useMetricSystem())
   val*=25.4;
  return val;
}

double PrintDialog::getPageWidMM()
{
  if (edPageWid->text() == "") return 0;
  double val = edPageWid->text().toDouble();
  if (!ApaintConfig::useMetricSystem())
   val*=25.4;
  return val;  
}

double PrintDialog::getPageHeiMM()
{
  if (edPageHei->text() == "") return 0;
  double val = edPageHei->text().toDouble();
  if (!ApaintConfig::useMetricSystem())
   val*=25.4;
  return val;    
}

void PrintDialog::setPageLeftMM(double val)
{
  if (!ApaintConfig::useMetricSystem())
    val = (double)((int)((val/24.5)*1000))/1000.0;
    
  edPageLeft->setText(QString("%1").arg(val));
}

void PrintDialog::setPageUpMM(double val)
{
  if (!ApaintConfig::useMetricSystem())
    val = (double)((int)((val/24.5)*1000))/1000.0;
  
  edPageUp->setText(QString("%1").arg(val));
  
}

void PrintDialog::setPageWidMM(double val)
{
  if (!ApaintConfig::useMetricSystem())
    val = (double)((int)((val/24.5)*1000))/1000.0;
  
  edPageWid->setText(QString("%1").arg(val));  
}

void PrintDialog::setPageHeiMM(double val)
{
  if (!ApaintConfig::useMetricSystem())
    val = (double)((int)((val/24.5)*1000))/1000.0;
  
  edPageHei->setText(QString("%1").arg(val));    
}


void PrintDialog::setNumPrints(int val)
{
  numPrints = val;
}

void PrintDialog::showSetup()
{
  pselDlg->show();
}

void PrintDialog::setRepeatInfo(double repWid, double repHei)
{
  repWid = (double)((int)(repWid*1000))/1000;
  lbRepeatWid1->setText(QString("%1").arg(repWid));
  repHei = (double)((int)(repHei*1000))/1000;
  lbRepeatHei1->setText(QString("%1").arg(repHei));
}

void PrintDialog::calcRepeatInfo()
{
  int xdpi, ydpi, lb, tb, pwid, phei, imgWid, imgHei;
  ApaintImageIO *io = ApaintImageIO::getObject();
  Layer *layer = doc->getLayer();
  if (!layer || !io) return;
  
  xdpi = io->getDensityX(pselDlg->getPrinterIndx(), pselDlg->getDensityXIndx())/1000;
  ydpi = io->getDensityY(pselDlg->getPrinterIndx(), pselDlg->getDensityYIndx())/1000;
  calcPageSize((int)getPageLeftMM(), (int)getPageUpMM(), (int)getPageWidMM(), (int)getPageHeiMM(), xdpi, ydpi, lb, tb, pwid, phei);
  calcImageSize(&layer->imageBuf, xdpi, ydpi, pwid, phei, imgWid, imgHei);
  setRepeatInfo((double)pwid/(double)imgWid, (double)phei/(double)imgHei);
}

void PrintDialog::pwidChanged(const QString& val)
{
  calcRepeatInfo(); 
}

void PrintDialog::pheiChanged(const QString& val)
{
  calcRepeatInfo(); 
}

void PrintDialog::zoomChanged(int val)
{
  calcRepeatInfo();
}
 
void PrintDialog::hideZoom(bool enable)
{
  if (enable)
  {
    lbZoom->hide();
    sbZoom->hide();
  }
  else
  {
    lbZoom->show();
    sbZoom->show();    
  }
}

void PrintDialog::oneRepeatClicked(bool on)
{
  calcRepeatInfo();
}

void PrintDialog::fitToPageClicked(bool on)
{
  hideZoom(on);
  calcRepeatInfo();
}

void PrintDialog::showDlg()
{
  Layer *layer = doc->getLayer();
  if (!layer)
   return;
  QString name = ApaintUtil::getBaseName(layer->imageBuf.imageName());
  pselDlg->setFileName(name);
  setupDlg();
  show();
  adjustSize(); 
}

void PrintDialog::setupDlg()
{      
  Layer *layer = doc->getLayer();
  if (!layer)
   return;
   
  calcRepeatInfo();
  numPrints = 1;
  sbNumPrints->setValue(numPrints);
  cbTitle->setChecked(false);
  cbLandscape->setChecked(false);
  if (layer->imageBuf.type() == LOAD_CMAP)
    cbUsedCol->show();
  else
    cbUsedCol->hide();
  setPrinterData();  
}

void PrintDialog::setPrinterData()
{
  int left, top, wid, hei;
  QString name;
  lbPrinter1->setText( pselDlg->getPrinter());
  lbPrintMode1->setText(pselDlg->getPrintMode());
  lbDensityX1->setText(pselDlg->getDensityX());
  lbDensityY1->setText(pselDlg->getDensityY());
  if (pselDlg->saveToFile())
  {
   lbFilename1->show();
   lbFilename0->show();
   pselDlg->getFileName(name);
   lbFilename1->setText(name);
  }
  else
  {
    lbFilename1->hide();
    lbFilename0->hide();
  }
  pselDlg->getPrinterLimitWidMM(left, wid);
  pselDlg->getPrinterLimitHeiMM(top, hei);
  setPageSizeMinMax(wid, hei);  
}


void PrintDialog::setFullPage()
{
  int left, top, wid, hei;
  pselDlg->getPrinterLimitWidMM(left, wid);
  pselDlg->getPrinterLimitHeiMM(top, hei);
  
  setPageLeftMM(0);
  setPageUpMM(0);
  setPageWidMM(wid);
  setPageHeiMM(hei);
}

//pwid and phei are in mm                                                                                                                                                                        
void PrintDialog::calcPageSize(int pl, int pt, int pwid, int phei, int xdpi, int ydpi,
                               int &lb, int &tb, int &wid, int &hei)
{
   wid = (int)(((double)pwid/25.4)*xdpi);
   hei = (int)(((double)phei/25.4)*ydpi);
   lb = (int)(((double)pl/25.4)*xdpi);
   tb = (int)(((double)pt/25.4)*ydpi);
}

int PrintDialog::getPixelSize(int mm, int dpi)
{
  return (int)(((double)mm/25.4)*dpi);
}

void PrintDialog::calcImageSize(ApaintImage *img, int xdpi, int ydpi, int &pwid, int &phei, int &wid, int &hei)
{
  double inchWid, inchHei, scale, scalex, scaley;
  int imgDpiX, imgDpiY;

  scale = (double)sbZoom->value()/(double)100;
  imgDpiX = img->xdpi();
  if (!imgDpiX)
    imgDpiX = 100000;
  imgDpiY = img->ydpi();
  if (!imgDpiY)
    imgDpiY = 100000;

   
  inchWid = ((double)img->width()*scale) / (double)(imgDpiX/1000);
  inchHei = ((double)img->height()*scale) / (double)(imgDpiY/1000);
  
  if (cbLandscape->isChecked())
  {
    wid = (int)(inchHei*xdpi);
    hei = (int)(inchWid*ydpi);
  }
  else
  {
   wid = (int)(inchWid*xdpi);
   hei = (int)(inchHei*ydpi); 
  }

  if (cbFitToPage->isChecked())
  {
    scalex = (double)pwid / (double)wid;
    scaley = (double)phei / (double)hei;
    if (scalex<scaley)
    {
     wid = pwid;
     hei = (int)(hei*scalex);
     if (cbOneRepeat->isChecked() && phei>hei)
      phei = hei;
    }
    else
    {
     hei = phei;
     wid = (int)(wid*scaley);
     if (cbOneRepeat->isChecked() && pwid>wid)
      pwid = wid;                
    }
  }
  else if (cbOneRepeat->isChecked())
  {      
    if (pwid>wid)
      pwid = wid;
    if (phei>hei)
      phei = hei;
  }    
}

bool PrintDialog::getBands(int xdpi, int ydpi, int bandHei, int imgType, QPtrList<PrintDataProducer> &pdpList, int &lb, int &tb)
{
  int pwid, phei;
  int imgWid, imgHei, titleHei, colorsHei;
  bool showTitle = cbTitle->isChecked();
  bool showColors = false;
  int titleHeiMM, colorsHeiMM;
  int bestSize;
  ApaintImage titleImg, colorsImg;
  PrintDataProducerImage *pdp=NULL, *pdpTitle=NULL, *pdpColors=NULL;

  Layer *layer= doc->getMainLayer();
  if (!layer) return false;

  showColors = (layer->imageBuf.type() == LOAD_CMAP) && cbUsedCol->isChecked();
 
  if (layer->imageBuf.type() == LOAD_TRUE_COL && imgType != IMG_TRUE_COL)
  {
    QMessageBox::information(this, ApaintConfig::getText(132), ApaintConfig::getText(128) , ApaintConfig::getText(3));
    return false;
  }

  titleHeiMM =0;
  colorsHeiMM = 0;
  if (showTitle)
   titleHeiMM+= TITLE_HEI_MM;
  if (showColors)
   colorsHeiMM+= TITLE_HEI_MM;

  calcPageSize((int)getPageLeftMM(), (int)getPageUpMM(), (int)getPageWidMM(), (int)getPageHeiMM()-titleHeiMM-colorsHeiMM, xdpi, ydpi, lb, tb, pwid, phei);
  calcImageSize(&layer->imageBuf, xdpi, ydpi, pwid, phei, imgWid, imgHei);
  titleHei = getPixelSize(titleHeiMM, ydpi);
  colorsHei = getPixelSize(colorsHeiMM, ydpi);
  
  QFont font = QFont(ApaintConfig::getPrintFontName(), 10);  

  if (showColors)
  {
    bestSize = ApaintConfig::getBestPointSize(font, (int)(colorsHei*0.5));
    font.setPointSize(bestSize);
    colorsImg = ApaintImageUtil::getColorsImage(layer->imageBuf, font, pwid, colorsHei); 
    pdpColors = new PrintDataProducerImage(colorsImg,
                                          pwid, colorsHei, pwid, colorsHei, bandHei,
                                          (char)imgType, false);
    if (!pdpColors)
    {
      qDebug("PrintDialog::doPrint(): NULL pointer!");
      close();
      return false;
    }
  }
  
  if (showTitle)
  {    
    bool doAntialias = (xdpi<250 && ydpi<250);   
    if (doAntialias)
      bestSize = ApaintConfig::getBestPointSize(font, (int)(titleHei*2));
    else
      bestSize = ApaintConfig::getBestPointSize(font, (int)(titleHei*0.5));
    font.setPointSize(bestSize);

    QString titleTxt = QString("%1  %2x%3").arg(ApaintUtil::getBaseName(layer->imageBuf.imageName()))
                                .arg(layer->imageBuf.width())
                                .arg(layer->imageBuf.height());
    titleImg = ApaintImageUtil::getTextImage(pwid, titleHei, font, titleTxt, doAntialias);
    
    pdpTitle = new PrintDataProducerImage(titleImg,
                                          pwid, titleHei, pwid, titleHei, bandHei,
                                          (char)imgType, false);
    if (!pdpTitle)
    {
      qDebug("PrintDialog::doPrint(): NULL pointer!");
      close();
      return false;
    }

  }
  pdp = new PrintDataProducerImage(layer->imageBuf,
                                   pwid, phei, imgWid, imgHei, bandHei,
                                   (char)imgType, cbLandscape->isChecked());

  if (!pdp)
  {
    qDebug("PrintDialog::doPrint(): NULL pointer!");
    close();
    return false;
  }
  pdp->setSaveToFile(pselDlg->saveToFile());
  if (showTitle)
    pdpTitle->setSaveToFile(pselDlg->saveToFile());
  if (showColors)
    pdpColors->setSaveToFile(pselDlg->saveToFile());
  if (pselDlg->saveToFile())
  {
    QString name;
    if (!pselDlg->getFileName(name))
    {
      QMessageBox::information(this, ApaintConfig::getText(132),ApaintConfig::getText(709) , ApaintConfig::getText(3));
      delete pdp;
      return false;
    }
    pdp->setFilename(name);
    if (showTitle)      
      pdpTitle->setFilename(name);
    if (showColors)
      pdpColors->setFilename(name);
  }
  if (showTitle)
    pdpList.append(pdpTitle);  
  pdpList.append(pdp);
  if (showColors)
    pdpList.append(pdpColors);
  return true;  
}

void PrintDialog::doPreview()
{
  int xdpi, ydpi, i, allHeight, lb, tb;
  PrintDataProducerImage *pdp=NULL;
  
  // calc page, image size  
  xdpi = ApaintConfig::getScreenDpiX();
  ydpi = ApaintConfig::getScreenDpiY();
  QPtrList<PrintDataProducer> pdpList;
  pdpList.setAutoDelete(true);
  
  if (!getBands(xdpi, ydpi, 250, IMG_TRUE_COL, pdpList, lb, tb))
   return;
  allHeight = 0;
  for (i=0; i<(int)pdpList.count(); i++)
   {
    pdp = (PrintDataProducerImage *) pdpList.at(i); 
    allHeight+=pdp->pageHeight();
   }
  // COPY ALL DATA TO PREVIEW IMAGE
  ApaintImage previewImage;
  
  previewImage.newImage(pdp->pageWidth()+lb, allHeight+tb, pdp->depth());
  QRgb whiteCol = qRgb(255, 255, 255);
  previewImage.fill(0, 0, pdp->pageWidth()+lb, tb, whiteCol, false, 0);
  previewImage.fill(0, 0, lb, allHeight+tb, whiteCol, false, 0);
  allHeight = 0;
  lb*=3; //it is IMG_TRUE_COL
  for (i=0; i<(int)pdpList.count(); i++)
   {
    pdp = (PrintDataProducerImage *) pdpList.at(i); 
    while (!pdp->endOfData())
    {  
     pdp->nextBand();
     if (lb>0)
     {
       for (int y = 0; y<(pdp->lastLine()-pdp->curLine()); y++)
        memcpy(previewImage.scanLine(allHeight+pdp->curLine()+tb+y)+lb, pdp->curDataPtr()+y*pdp->bpl(), pdp->bpl());     
     }
     else
       memcpy(previewImage.scanLine(allHeight+pdp->curLine()+tb), pdp->curDataPtr(), pdp->bpl()*(pdp->lastLine()-pdp->curLine()));     
    }
    allHeight+=pdp->lastLine();
   }

  // REMOVE PREVIUOS DATA
  pdpList.clear();
    
  //SET PREVIEW IMAGE TO PREVIEW DIALOG AND SHOW IT
  previewDlg->setImage(previewImage);
  previewDlg->showDlg();  
}

void PrintDialog::doPrint()
{
  int xdpi, ydpi, lb, tb, imgType;

  ApaintImageIO *io = ApaintImageIO::getObject();  
  imgType = io->getPrinterMode(pselDlg->getPrinterIndx(), pselDlg->getPrintModeIndx());
  
  xdpi = io->getDensityX(pselDlg->getPrinterIndx(), pselDlg->getDensityXIndx())/1000;
  ydpi = io->getDensityY(pselDlg->getPrinterIndx(), pselDlg->getDensityYIndx())/1000;

  //because gimp-print doesn't support diff. xdpi and ydpi
  if (xdpi == 1440 && ydpi == 720)
   xdpi/=2;
  if (xdpi == 2880 && ydpi == 720)
   xdpi/=4;
  if (xdpi == 2880 && ydpi == 1440)
   xdpi/=2;
   
  QPtrList<PrintDataProducer> pdpList;
  pdpList.setAutoDelete(true);

  if (!getBands(xdpi, ydpi, 250, imgType, pdpList, lb, tb))
   return;
     
  io->setPrintCapType(cbType->currentText());
  ShowProgressDialog  *progressDlg = new ShowProgressDialog(this, "progress dlg", pdpList);
  progressDlg->show();
  bool res = io->printFile(pselDlg->getPrinterIndx(), pdpList, numPrints, xdpi*1000, ydpi*1000, lb, tb);
  progressDlg->close();
  if (res)
  {
    if (!pselDlg->saveToFile())
      QMessageBox::information(this, ApaintConfig::getText(132),ApaintConfig::getText(123) , ApaintConfig::getText(3));
  }
  else
    QMessageBox::information(this, ApaintConfig::getText(132),ApaintConfig::getText(125) , ApaintConfig::getText(3));

  pdpList.clear();
  delete progressDlg;
  close();
}

void PrintDialog::closeDlg()
{
  close();                                           
}

void PrintDialog::readPrintcapFile(QStringList& list)
{
 int  doublePointPos, orPos;
 list.clear();
#ifdef __APPLE__ 
#define MAX_LINE	5000
 FILE *fp;
 int lastLinePrinterName=FALSE;
 char *strVal;
 char printcapFile[200];
 char oneLine[MAX_LINE+1];

 strcpy(printcapFile, "/Library/Preferences/org.cups.printers.plist");
 fp = fopen(printcapFile, "r" );
 if (!fp)
  return;

 while (!feof(fp) && fgets(oneLine, MAX_LINE, fp))
  {
   if (strstr(oneLine, "<key>printer-name</key>"))
    lastLinePrinterName=TRUE;
   else
    {
     if (lastLinePrinterName)
      if (strVal = parseOneLineTagXML(oneLine, "string"))
       if (strlen(strVal))
        list.append(strVal);
     lastLinePrinterName=FALSE;
    }
  }
 fclose(fp);
#else
 QFile pcapFile(PRINT_CAP_PATH_1);

 if (!pcapFile.exists())
  {
   pcapFile.setName(PRINT_CAP_PATH_2);
   if (!pcapFile.exists())
    {
     pcapFile.setName(PRINT_CAP_PATH_3);
     if (!pcapFile.exists())
      {
       pcapFile.setName(PRINT_CAP_PATH_4);
       if (!pcapFile.exists())
        return;
      }
    }
  }
 if (pcapFile.open(IO_ReadOnly))
  {
   QTextStream stream(&pcapFile);
   QString line;
   while (!stream.eof())
    {
     line = stream.readLine();
     line.stripWhiteSpace();
     if (!line.startsWith("#"))
      {
       doublePointPos = line.find(":");
       orPos = line.find("|");
       if (doublePointPos != -1)
        {
         if (orPos != -1)
          line.truncate(orPos);
         else
          line.truncate(doublePointPos);
         list.append(line);
        }
      }
    }
   pcapFile.close();
 }
#endif
}



