/***************************************************************************
                          eliparctool.cpp  -  description
                             -------------------
    begin                : Thu Mar 27 2003
    copyright            : (C) 2003-2011 Arahne, d.o.o.
    email                : arahne@arahne.si
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "toolElipArc.h"
#include <math.h>
#include <stdlib.h>
#include <qpainter.h>
#include "apaintUtil.h"
#include "apaintdrawrepeat.h"

extern char *getCommonPNG(const char *pixName);

#define PIV2  102944    /* fixed point PI/2 */
#define TWOPI 411775    /* fixed point 2*PI */
#define HALF  32768     /* fixed point 1/2 */
#define C_SIZE 15
typedef long FIX;       /* 32-bit fixed point, 16-bit fraction */


ElipArcTool::ElipArcTool(DrawInRepeat *drawp, ToolType type, PropObject *propObj):ToolBase(type, propObj)
{
 int i;
 
 cur=QCursor(ArrowCursor);
 setToggleBtn(QPixmap(getCommonPNG("draw-curve")), tr("Ellipse Arc Tool"));
 pt=0;
 for(i=0;i<3;i++)
 {
   xr[i]=0;
   yr[i]=0;
 }
 pts=0;
 numPts=0;
 if (propObj)
  setUsepWidget(true);
 Q_CHECK_PTR(drawp);
 draw = drawp;  
}
ElipArcTool::~ElipArcTool()
{  
   delete pts;
}

void ElipArcTool::showCommonProp(bool enable)
{
  CommonProp *cProp = CommonProp::getObject();
  cProp->setAspectDrawProp(enable);
  cProp->setThickLinesProp(enable);
}

void ElipArcTool::drawCross(int x, int y, QPainter *p)
{
  int xi, yi;
  if (!p) return;
  xi=(int)(x*zoomX)+(int)(zoomX/2);
  yi=(int)(y*zoomY)+(int)(zoomY/2);
  
  p->setRasterOp(Qt::XorROP);
  p->setPen(QPen(QColor(128, 128,128),1));
  p->drawLine(xi-C_SIZE, yi, xi+C_SIZE, yi);
  p->drawLine(xi, yi-C_SIZE, xi, yi+C_SIZE);  
}

void ElipArcTool::mousePressed(QMouseEvent *e, QPainter *pc, QPainter *pi)
{
 int xi=(int)(e->x()/zoomX);
 int yi=(int)(e->y()/zoomY);

 if (pt==3 && drawStarted) return;
 btnPressed = e->button();
 if (btnPressed == MidButton) return;
 drawStarted=true;

 pi->setRasterOp(NotROP);
 pc->setRasterOp(NotROP);
 if (pt<2)
 {
  xr[pt] = xi; yr[pt] = yi;
  drawCross(xr[pt], yr[pt], pc);
  pt++;
 }else
 {
  drawCross(xr[0], yr[0], pc);
  drawCross(xr[1], yr[1], pc); 
  xr[pt] = xi; yr[pt] = yi;
  numPts=drawElArc(xr[0], yr[0], xr[1], yr[1], xr[2], yr[2], pts);
  draw->drawPolygon(pi, pts, true, (btnPressed == LeftButton), false);
  pt++;
 }
}

void ElipArcTool::mouseMoved(QMouseEvent *e, QPainter *pc, QPainter *pi)
{ 
 if (pt==3 && drawStarted)
 {
   pi->setRasterOp(NotROP);
   pc->setRasterOp(NotROP);
   draw->drawPolygon(pi, pts, false, false, false);
   xr[2]=(int)(e->pos().x()/zoomX);
   yr[2]=(int)(e->pos().y()/zoomY);      
   numPts=drawElArc(xr[0], yr[0], xr[1], yr[1], xr[2], yr[2], pts);
   draw->drawPolygon(pi, pts, true, (btnPressed == LeftButton), false);
 }
}

void ElipArcTool::mouseReleased(QMouseEvent *e, QPainter *pc, QPainter *pi)
{
  if (pt==3)
  {
    pt=0;
    draw->drawPolygon(pi, pts, true, (btnPressed == LeftButton), true);  
  }
  drawStarted=false;

}

int ElipArcTool::drawElArc(int Px, int Py, int Qx, int Qy, int Kx, int Ky, QPointArray &epts)
{
 long xP, yP, xQ, yQ, xK, yK, m, len;
 int xPQ, yPQ, n_pts, diff;
 unsigned int dist;

 /* popravi tocko K tako, da bo 2* dlje od daljice PQ tako da bo blizje vrhu elipsoide */
 xP = Px << 16;
 yP = Py << 16;
 xQ = Qx << 16;
 yQ = Qy << 16;
 xK = Kx;
 yK = Ky;
 xPQ = Px + (int) ((int) (Qx - Px) / 2);
 yPQ = Py + (int) ((int) (Qy - Py) / 2);

 xK += (int) (Kx - xPQ);
 yK += (int) (Ky - yPQ);

 diff = abs (Px - Kx);
 dist = diff * diff;
 diff = abs (Py - Ky);
 dist += diff * diff;

 diff = abs (Qx - Kx);
 dist += diff * diff;
 diff = abs (Qy - Ky);
 dist += diff * diff;

 n_pts = (int)(sqrt (dist) / 4);
 m = (int)(log10 (n_pts / (M_PI_2)) / log10 (2.0));
 if (m < 2)
   m = 2;
 else if (m > 12)
   m = 12; /* maximum */

 len = qtr_elips (xP, yP, xQ, yQ, xK << 16, yK << 16, m, epts);
 epts.setPoint(0, Qx, Qy);
 return (len);
 
}

int ElipArcTool::qtr_elips(long xP, long yP, long xQ, long yQ, long xK, long yK, long m, QPointArray &epts)
{
 int i, x, y, len;
 FIX vx, ux, vy, uy, w, xJ, yJ;

 vx = xK - xQ;			/* displacements from center */
 ux = xK - xP;
 vy = yK - yQ;
 uy = yK - yP;
 xJ = xP - vx + HALF;		/* center of ellipse J */
 yJ = yP - vy + HALF;
 ux -= (w = ux >> (2 * m + 3));	/* cancel 2nd-order error */
 ux -= (w >>= (2 * m + 4));	/* cancel 4th-order error */
 ux -= w >> (2 * m + 3);	/* cancel 6th-order error */
 ux += vx >> (m + 1);		/* cancel 1st-order error */
 uy -= (w = uy >> (2 * m + 3));	/* cancel 2nd-order error */
 uy -= (w >>= (2 * m + 4));	/* cancel 4th-order error */
 uy -= w >> (2 * m + 3);	/* cancel 6th-order error */
 uy += vy >> (m + 1);		/* cancel 1st-order error */
 len = (PIV2 << m) >> 16;

 epts = QPointArray(len+1);
 for (i = len; i >= 1; --i)
  {
   x = (xJ + vx) >> 16;
   y = (yJ + vy) >> 16;

   epts.setPoint(i, x, y);
 
   ux -= vx >> m;
   vx += ux >> m;
   uy -= vy >> m;
   vy += uy >> m;
  }  
  return (len+1);
}

void ElipArcTool::end()
{
  int i;
  ToolBase::end();
  for (i=0; i<pt; i++)
   draw->refreshArea(xr[i], yr[i], xr[i], yr[i], C_SIZE);
  pt=0;
  drawStarted=0;
  pts = QPointArray();
}

