/***************************************************************************
                          toolInsRemPixels.cpp  -  description
                             -------------------
    begin                : Fri Mar 19 2004
    copyright            : (C) 2003-2011 Arahne, d.o.o.
    email                : arahne@arahne.si
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdlib.h>
#include <qspinbox.h>
#include <qradiobutton.h>
#include <qbuttongroup.h>
#include <qlabel.h>
#include <qcheckbox.h>
#include <qbitmap.h>
#include <qtooltip.h>
#include "apaintdoc.h"
#include "apaintdrawrepeat.h"
#include "apaintconfig.h"
#include "toolInsRemPixels.h"
#include "apaintBrushList.h"

extern char *getCommonPNG(const char *pixName);

static bool insertDuplicate = FALSE;
static DuplicateType duplicateType = COLOR;

InsertRemovePixelsDialog::InsertRemovePixelsDialog(QWidget *parent, const char *name):InsertRemovePixelsDialogUI(parent, name)
{
  setCaption(ApaintConfig::getText(680));
  bgType->setTitle(ApaintConfig::getText(678));
  lbPos->setText(ApaintConfig::getText(677));
  lbNumPix->setText(ApaintConfig::getText(679));
  rbBefore->setText(ApaintConfig::getText(681));
  rbAfter->setText(ApaintConfig::getText(682));
  btnOK->setText(ApaintConfig::getText(OK_MSG));
  btnClose->setText(ApaintConfig::getText(86));
  areaButton->setText(ApaintConfig::getText(761));
  lineButton->setText(ApaintConfig::getText(760));

  rbRemoveVert->setPixmap(QPixmap(getCommonPNG("cutX-minus")));
  rbInsertVert->setPixmap(QPixmap(getCommonPNG("cutX-plus")));
  
  rbRemoveHor->setPixmap(QPixmap(getCommonPNG("cutY-minus")));
  rbInsertHor->setPixmap(QPixmap(getCommonPNG("cutY-plus")));

//  cbUseVectors->setText(ApaintConfig::getText(763)); it does not display text and icon in parallel
  cbUseVectors->setPixmap(QPixmap(getCommonPNG("cut-irregular")));
  QToolTip::add(cbUseVectors, ApaintConfig::getText(763));

  cbFindPath->setPixmap(QPixmap(getCommonPNG("marble")));
  QToolTip::add(cbFindPath, ApaintConfig::getText(836));

  QToolTip::add(colorButton, ApaintConfig::getText(DUPLICATE_MSG));

  connect(btnOK, SIGNAL(clicked()), this, SIGNAL(okBtnClicked()));
  connect(btnClose, SIGNAL(clicked()), this, SLOT(closeDlg()));
  connect(rbBefore, SIGNAL(clicked()), this, SLOT(rbBeforeClicked()));
  connect(rbAfter, SIGNAL(clicked()), this, SLOT(rbAfterClicked()));
  connect(bgType, SIGNAL(clicked(int)), this, SLOT(typeChanged(int)));
  connect(cbUseVectors, SIGNAL(toggled(bool)), this, SIGNAL(selTypeChange(bool)));
  connect(cbUseVectors, SIGNAL(toggled(bool)), this, SLOT(cbUseVectorsClicked(bool)));
  connect(cbFindPath, SIGNAL(toggled(bool)), this, SLOT(cbFindPathClicked(bool)));
  connect(colorButton, SIGNAL(clicked()), this, SLOT(colorClicked()));
  connect(areaButton, SIGNAL(clicked()), this, SLOT(areaClicked()));
  connect(lineButton, SIGNAL(clicked()), this, SLOT(lineClicked()));

  rbBefore->setChecked(true);
}

void InsertRemovePixelsDialog::showDlg()
{
  show();
}

void InsertRemovePixelsDialog::closeDlg()
{
  close();  
}

void InsertRemovePixelsDialog::closeEvent(QCloseEvent *e)
{
  emit closeSig();
  InsertRemovePixelsDialogUI::closeEvent(e);
}

void InsertRemovePixelsDialog::colorClicked()
{
 colorButton->setChecked(true);
 areaButton->setChecked(false);
 lineButton->setChecked(false);
 duplicateType = COLOR;
 emit duplicateChanged(false);
}

void InsertRemovePixelsDialog::areaClicked()
{
  colorButton->setChecked(false);
  areaButton->setChecked(true);
  lineButton->setChecked(false);
  duplicateType = AREA;
  emit duplicateChanged(true);
}

void InsertRemovePixelsDialog::lineClicked()
{
  colorButton->setChecked(false);
  areaButton->setChecked(false);
  lineButton->setChecked(true);
  duplicateType = LINE;
  emit duplicateChanged(true);
}

void InsertRemovePixelsDialog::rbBeforeClicked()
{
  rbBefore->setChecked(true);
  rbAfter->setChecked(false);
}

void InsertRemovePixelsDialog::rbAfterClicked()
{
  rbBefore->setChecked(false);
  rbAfter->setChecked(true);  
}

void InsertRemovePixelsDialog::setPos(int posx, int posy)
{
  if (getOp() == InsertHor || getOp() == RemoveHor)
    sbPos->setValue(posx+1);
  if (getOp() == InsertVert || getOp() == RemoveVert)
    sbPos->setValue(posy);
}

InsertRemovePixelsDialog::OpType InsertRemovePixelsDialog::getOp()
{
  if (rbInsertHor->isChecked())
   return InsertHor;
  else if (rbInsertVert->isChecked())
   return InsertVert;
  else if (rbRemoveHor->isChecked())
   return RemoveHor;
  else if (rbRemoveVert->isChecked())
   return RemoveVert;

  return InsertHor;  
}

void InsertRemovePixelsDialog::setNum(int val)
{
  sbNumPix->setValue(val);
}

int InsertRemovePixelsDialog::getPos()
{
  return sbPos->value();
}

int InsertRemovePixelsDialog::getNum()
{
  return sbNumPix->value();
}

void InsertRemovePixelsDialog::typeChanged(int typ)
{
  QButton *btn = bgType->find(typ);
  if (btn == rbRemoveHor || btn == rbInsertHor)
   sbPos->setMaxValue(maxHor);
  else if (btn == rbRemoveVert || btn == rbInsertVert)
   sbPos->setMaxValue(maxVert);
}

void InsertRemovePixelsDialog::cbUseVectorsClicked(bool set)
{
 cbFindPath->setEnabled(!set);
}

void InsertRemovePixelsDialog::cbFindPathClicked(bool set)
{
 cbUseVectors->setEnabled(!set);
}

void InsertRemovePixelsDialog::setPosLimit(int maxH, int maxV)
{
  maxHor = maxH;
  maxVert = maxV;
  typeChanged(bgType->id(bgType->selected()));
}

InsertRemovePixelsDialog::SelectType InsertRemovePixelsDialog::getSelType()
{
  if (cbUseVectors->isChecked() || cbFindPath->isChecked())
   return(VectorType);
  else
   return(PointType);
}
    
/*************************class InsertRemovePixelsTool*****************************/

InsertRemovePixelsTool::InsertRemovePixelsTool(ToolType type, DrawInRepeat *drawp, ApaintDoc *docp, InsertRemovePixelsDialog *dlgp, PropObject *propObj):ToolBase(type, propObj)
{
 inFirstTile = 0;
 useBgCol = false;
 ptsList.clear();
 ptsList.setAutoDelete(true);
 setToggleBtn(QPixmap(getCommonPNG("insert-lines")), tr("Insert Pixel"));
 bList = ApaintBrushList::getObject();
 if (propObj)
  setUsepWidget(true);
 setAutoScroll(false);
 Q_CHECK_PTR(docp);
 doc = docp;
 Q_CHECK_PTR(dlgp);
 dlg = dlgp;
 Q_CHECK_PTR(drawp);
 draw = drawp;
 connect(dlg, SIGNAL(okBtnClicked()), this, SLOT(doAction()));
 connect(dlg, SIGNAL(selTypeChange(bool)), this, SLOT(selTypeChanged(bool)));
 connect(dlg, SIGNAL(duplicateChanged(bool)), this, SLOT(duplicateChanged(bool)));
 connect(doc, SIGNAL(layerChanged(Layer *)), this, SLOT(layerChangedSlot(Layer *)));
}

void InsertRemovePixelsTool::begin()
{
  initData();
  dlg->showDlg();
}

void InsertRemovePixelsTool::end()
{
  dlg->close();
}

void InsertRemovePixelsTool::layerChangedSlot(Layer *layer)
{
 if (!layer)
  return;
  
 dlg->setPosLimit(layer->layerPos().width(), layer->layerPos().height());
}

void InsertRemovePixelsTool::selTypeChanged(bool useVect)
{
  if (!useVect)
  {
   QRect pos = getBorderRect();
   initData();
   doc->refreshLayer(pos);
  }
}

void InsertRemovePixelsTool::duplicateChanged(bool doDuplicate)
{
 insertDuplicate = dlg->colorButton->isChecked() == false;
 emit updateAreaLineButtons(insertDuplicate);
}

void InsertRemovePixelsTool::initData()
{
  Layer *layer = doc->getLayer();
  
  ptsList.clear();
  drawStarted = false;
  inFirstTile = true;
  btnPressed = false;
  if (layer && layer->getType() == MainLayer && layer->isUsingTiles())
   {
    layer->setDrawDelimiterLines(true);
    doc->refreshLayer();
   }
}

void InsertRemovePixelsTool::drawScreenRect(QPainter *p, int x1, int y1, int x2, int y2)
{
  p->setRasterOp(Qt::XorROP);
  p->setPen(QPen(QColor(128, 128,0),QMAX(4, (int)(zoomX/2))/*, Qt::DashLine*/));
  QRect r = normalRect(x1, y1, x2, y2);
  p->drawRect((int)(r.x()*zoomX), (int)(r.y()*zoomY), (int)(r.width()*zoomX), (int)(r.height()*zoomY));
}         

void InsertRemovePixelsTool::drawScaleLine(QPainter *p, int x1, int y1, int x2, int y2)
{
  p->setRasterOp(Qt::XorROP);
  p->setPen(QPen(QColor(128, 128,0),QMAX(2, (int)(zoomX/2)),Qt::DashLine));
  draw->drawScreenLine(p, x1, y1, x2, y2, zoomX, zoomY);
}

void InsertRemovePixelsTool::refreshPoly(QPainter *p)
{ 
 int x1, y1, x2, y2, i, count;
 QPoint *pt;

 pt = ptsList.first();
 count = ptsList.count();
 if (!count) return;

 x1 = pt->x();
 y1 = pt->y();
 for (i =1; i<count; i++)
  {
   pt = ptsList.next();
   x2 = pt->x();
   y2 = pt->y();
   drawScaleLine(p, x1, y1, x2, y2);
   x1 = x2;
   y1 = y2;
  }
}

void InsertRemovePixelsTool::beforeRefresh(QPainter *p)
{  
 if (dlg->getSelType() == InsertRemovePixelsDialog::PointType)
  {  
    if (!btnPressed) return;
    Layer *sel = doc->getLayer();
    if (!sel) return;
    if ((dlg->getOp() == InsertRemovePixelsDialog::InsertHor) || (dlg->getOp() == InsertRemovePixelsDialog::RemoveHor))
     drawScreenRect(p, firstx, 0,   firstx + (lastx -  firstx), sel->imageBuf.height());
    else if (dlg->getOp() == InsertRemovePixelsDialog::InsertVert || dlg->getOp() == InsertRemovePixelsDialog::RemoveVert) 
     drawScreenRect(p, 0, firsty, sel->imageBuf.width(), firsty + (lasty-firsty));    
  }
 else if (dlg->getSelType() == InsertRemovePixelsDialog::VectorType)
  {
    if (!ptsList.count()) return;
    refreshPoly(p);
    drawScaleLine(p, firstx, firsty, lastx, lasty);
  }
}

void InsertRemovePixelsTool::afterRefresh(QPainter *p)
{ 
 if (dlg->getSelType() == InsertRemovePixelsDialog::PointType)
  {
    if (!btnPressed) return;
    Layer *sel = doc->getLayer();
    if (!sel) return;
    if ((dlg->getOp() == InsertRemovePixelsDialog::InsertHor) || (dlg->getOp() == InsertRemovePixelsDialog::RemoveHor))
     drawScreenRect(p, firstx, 0,   firstx + (lastx -  firstx), sel->imageBuf.height());
    else if (dlg->getOp() == InsertRemovePixelsDialog::InsertVert || dlg->getOp() == InsertRemovePixelsDialog::RemoveVert)
     drawScreenRect(p, 0, firsty, sel->imageBuf.width(), firsty + (lasty-firsty));  
  }
 else if (dlg->getSelType() == InsertRemovePixelsDialog::VectorType)
  {
    if (!ptsList.count()) return;
    refreshPoly(p);
    drawScaleLine(p, firstx, firsty, lastx, lasty);
  }
}

uchar *InsertRemovePixelsTool::getBrushBuffer(int wid, int hei, int dx, int dy)
{
  uchar *srcLine, *dstLine;
  int x, xx, y, yy, bpp, bpl, bhei, bwid,  offx, offy;
  int *transTable;
  QRgb *bColTab;
  Layer *selLayer = doc->getLayer();
  if (!selLayer) return 0;
  uint col = selLayer->getFgCol();
  int colPtr =  selLayer->getFgIndx();

  ApaintImage img = selLayer->imageBuf;
  ApaintImage bimg = bList->getBrush();
  QImage mask = bList->getBrushMask();
  bpp = img.depth()/8;
  bpl = bpp*wid;
  bhei = bimg.height();
  bwid = bimg.width();
  bColTab = bimg.colorTable();
  offx = bList->getOffsetX();
  offy = bhei - (hei % bhei) + bList->getOffsetY();
  if (bList->getAbsOffset())
   {
    offx+=dx;
    offy+=dy;
   }
  uchar *buf = new uchar[hei*bpl];
  if (!buf)
   {
    qDebug("NULL pointer: InsertRemovePixelsTool::getBrushBuffer!");
    return 0;
   }
  transTable = bList->getTransTable();
  if ((img.type() == LOAD_CMAP) && bList->isColorTabDifferent())
   {
    doc->putUndoStack(ChangeColorTable);
    bList->mergeColorTables(selLayer);    
    emit colorTableChanged(selLayer);
   }
  for (y = 0; y<hei; y++)
   {
     yy = (y+offy)%bhei;
     srcLine = bimg.scanLine(yy);
     dstLine = &buf[bpl*y];
     if (img.type() == LOAD_CMAP)
     {
       for (x = 0; x<wid; x++)
       {
          xx = (x+offx)%bwid;
          if (PIX_ON(mask, xx, yy))
            *dstLine++ = transTable[srcLine[xx]];
          else
            *dstLine++ = colPtr;
       }
     }
     else if (img.type() == LOAD_TRUE_COL)
     {
        if (bimg.type() == LOAD_CMAP)
          for (x = 0; x<wid; x++)
          {
            xx = (x+offx)%bwid;
            if (PIX_ON(mask, xx, yy))
             {
              *dstLine++ = qRed(bColTab[srcLine[xx]]);
              *dstLine++ = qGreen(bColTab[srcLine[xx]]);
              *dstLine++ = qBlue(bColTab[srcLine[xx]]);
             }
            else
             {
              *dstLine++ = qRed(col);
              *dstLine++ = qGreen(col);
              *dstLine++ = qBlue(col);
             }
          }
        else if(bimg.type() == LOAD_TRUE_COL)
          for (x = 0; x<wid; x++)
          {
             xx = ((x+offx)%bwid)*bpp;
             if (PIX_ON(mask, ((x+offx)%bwid), yy))
              {
               *dstLine++ = srcLine[xx];
               *dstLine++ = srcLine[xx+1];
               *dstLine++ = srcLine[xx+2];
              }
             else
              {
               *dstLine++ = qRed(col);
               *dstLine++ = qGreen(col);
               *dstLine++ = qBlue(col);
              }
          }        
      }
  }

  return buf;
}

void InsertRemovePixelsTool::getMinMaxY(QPointArray &pts, int &min, int &max)
{
 int yi;
 max = 0;
 min = INT_MAX;
 for (int i=0; i<(int)pts.count(); i++)
  {
   yi = pts.point(i).y();
   if(yi > max)
     max = yi;
   if (yi < min)
     min = yi;
  }
}

void InsertRemovePixelsTool::getMinMaxX(QPointArray &pts, int &min, int &max)
{
 int xi;
 max = 0;
 min = INT_MAX;
 for (int i=0; i<(int)pts.count(); i++)
  {
   xi = pts.point(i).x();
   if(xi > max)
     max = xi;
   if (xi < min)
     min = xi;
  } 
}

void InsertRemovePixelsTool::crackLayer(QPointArray &r1, QPointArray &r2)
{
 QRect newPos;
 int numPix = dlg->getNum();
 int top, bottom, left, right, dummy;
 bool insert=FALSE;
 Layer *newLayer=NULL, *tmpLayer;
 Layer *selLayer = doc->getLayer();
 if (!selLayer) return;
 int selLayerPos = doc->getLayerNum();
 uint fillCol=0;

 if (selLayer->imageBuf.type() == LOAD_TRUE_COL)
  fillCol = selLayer->getFgCol();
 else if (selLayer->imageBuf.type() == LOAD_CMAP)
  fillCol  = selLayer->getFgIndx();

 //create new main layer
 if (dlg->getOp() == InsertRemovePixelsDialog::InsertVert)
  {
   newLayer = new Layer(selLayer->imageBuf.width(), selLayer->imageBuf.height()+numPix, selLayer->imageBuf.depth(),
    fillCol, 2, selLayer->getType(), aspectX, aspectY);
   getMinMaxY(r1, top, dummy);
   getMinMaxY(r2, dummy, bottom);
   left = 0;
   right =  selLayer->imageBuf.width();
   insert = true;
  }
 else  if (dlg->getOp() == InsertRemovePixelsDialog::RemoveVert)
  {
   newLayer = new Layer(selLayer->imageBuf.width(), selLayer->imageBuf.height()-numPix, selLayer->imageBuf.depth(),
    fillCol, 2, selLayer->getType(), aspectX, aspectY);
   insert = false;
  }
 else  if (dlg->getOp() == InsertRemovePixelsDialog::InsertHor)
  {
   newLayer = new Layer(selLayer->imageBuf.width()+numPix, selLayer->imageBuf.height(), selLayer->imageBuf.depth(),
                 fillCol, 2, selLayer->getType(), aspectX, aspectY);
   getMinMaxX(r1, left, dummy);
   getMinMaxX(r2, dummy, right);
   top = 0;
   bottom =  selLayer->imageBuf.height();
   insert = true;
  }
 else if (dlg->getOp() == InsertRemovePixelsDialog::RemoveHor)
  {
   newLayer = new Layer(selLayer->imageBuf.width()-numPix, selLayer->imageBuf.height(), selLayer->imageBuf.depth(),
    fillCol, 2, selLayer->getType(), aspectX, aspectY);
   insert = false;
  }

 selLayer->imageBuf.copyImageDataTo(&newLayer->imageBuf);
 selLayer->origBuf.copyImageDataTo(&newLayer->origBuf);
 //set bg and fg color from selLayer to newLayer
 newLayer->setFgIndx(selLayer->getFgIndx());
 newLayer->setBgIndx(selLayer->getBgIndx());
 newLayer->setFgCol(selLayer->getFgCol());
 newLayer->setBgCol(selLayer->getBgCol());

 // copy first part of selection
 tmpLayer = new Layer(r1, SelectionLayer, Freehand, aspectX, aspectY);
 tmpLayer->loadDataFromLayer(selLayer, (r1.boundingRect()).normalize());
 QRect rect(0,0,selLayer->imageBuf.width(),selLayer->imageBuf.height());
 tmpLayer->loadDataFromLayer(selLayer, rect);
 QImage img = tmpLayer->getMask().copy();

 if (insert)
  {
   ApaintImage image;
   uchar *brushBuf;

   if (insertDuplicate || bList->useBrush())
    {
     int bwid = QABS(right-left);
     int bhei = QABS(bottom-top);

     if (dlg->getOp() == InsertRemovePixelsDialog::InsertHor)
	     bwid += numPix + left;
     else if (dlg->getOp() == InsertRemovePixelsDialog::InsertVert)
	     bhei += numPix + top;

     int bpp = newLayer->imageBuf.depth() / 8;

     if (insertDuplicate)
      {
       image = selLayer->origBuf.copy(0, 0, bwid, bhei);

       if (duplicateType == LINE)
	{
	 if (dlg->getOp() == InsertRemovePixelsDialog::InsertHor)
	  for (int y=0; y<bhei; ++y)
	   for (int x=1; x<bwid; ++x)
	    if ((x>=bwid-numPix) || PIX_ON(img,x,y) == false)
	     {
	      if (image.type() == LOAD_TRUE_COL)
	       image.setPixel(x,y,image.pixel(x-1,y));
	      else
	       image.setPixel(x,y,image.pixelIndex(x-1,y));
	     }
	 else
	  for (int y=0; y<bhei; ++y)
	   for (int x=0; x<bwid; ++x)
	    if ((y>=bhei-numPix) || PIX_ON(img,x,y) == false)
	     {
	      if (image.type() == LOAD_TRUE_COL)
	       image.setPixel(x,y,image.pixel(x,y-1));
	      else
	       image.setPixel(x,y,image.pixelIndex(x,y-1));
	     }
	}
       brushBuf = image.scanLine(0);
      }
     else
      brushBuf = getBrushBuffer(bwid, bhei, left, top);
//printf("getting brush buffer wid %d hei %d start %dx%d\n", bwid, bhei, left, top);
     if (dlg->getOp() == InsertRemovePixelsDialog::InsertVert)
      {
       memcpy(newLayer->imageBuf.scanLine(0) + bpp*left,  brushBuf, bwid*bhei*bpp);
       memcpy(newLayer->origBuf.scanLine(0) + bpp*left,  brushBuf, bwid*bhei*bpp);
      }
     else
      for (int i=0; i<bhei; i++)
       {
	memcpy(newLayer->imageBuf.scanLine(i) + bpp*left, brushBuf + i*bpp*bwid, bpp*bwid);
	memcpy(newLayer->origBuf.scanLine(i) + bpp*left, brushBuf + i*bpp*bwid, bpp*bwid);
       }
     if (!insertDuplicate)
      delete[] brushBuf;
    }
   else
    {
     newLayer->imageBuf.fill(fillCol);
     newLayer->origBuf.fill(fillCol);
    }
  }
 doc->copyImageToLayer(newLayer, tmpLayer->layerPos(), tmpLayer->imageBuf, tmpLayer->origBuf,
  false, tmpLayer->usingMask(), &tmpLayer->getMask(), tmpLayer->getTransColors(), FALSE);

 img.invertPixels();
 delete tmpLayer;

// copy second part of selection
 if (insert)
  {
   tmpLayer = new Layer(r1, SelectionLayer, Freehand, aspectX, aspectY);
   tmpLayer->loadDataFromLayer(selLayer, img.rect());
   tmpLayer->setMask(img);
  }
 else
  {
   tmpLayer = new Layer(r2, SelectionLayer, Freehand, aspectX, aspectY);
   tmpLayer->loadDataFromLayer(selLayer, (r2.boundingRect()).normalize());
  }

 if (dlg->getOp() == InsertRemovePixelsDialog::InsertVert)
  newPos = QRect(tmpLayer->layerPos().x(), tmpLayer->layerPos().y()+numPix, tmpLayer->layerPos().width(), tmpLayer->layerPos().height());
 else if (dlg->getOp() == InsertRemovePixelsDialog::RemoveVert)
  newPos = QRect(tmpLayer->layerPos().x(), QMAX(0, tmpLayer->layerPos().y()-numPix), tmpLayer->layerPos().width(), tmpLayer->layerPos().height());
 else if (dlg->getOp() == InsertRemovePixelsDialog::InsertHor)
  newPos = QRect(tmpLayer->layerPos().x()+numPix, tmpLayer->layerPos().y(), tmpLayer->layerPos().width(), tmpLayer->layerPos().height());
 else if (dlg->getOp() == InsertRemovePixelsDialog::RemoveHor)
  newPos = QRect(QMAX(0, tmpLayer->layerPos().x()-numPix), tmpLayer->layerPos().y(), tmpLayer->layerPos().width(), tmpLayer->layerPos().height());

 doc->copyImageToLayer(newLayer, newPos, tmpLayer->imageBuf, tmpLayer->origBuf,
          false, tmpLayer->usingMask(), &tmpLayer->getMask(), tmpLayer->getTransColors(), FALSE);
 delete tmpLayer;
 doc->setLayer(selLayerPos, newLayer);
}

void InsertRemovePixelsTool::getFirstLastPoint(QPoint &fp, QPoint &lp, QPoint &wrapPoint)
{
 Layer *selLayer = doc->getLayer();
 if (!selLayer) return;

 QPoint *pt, *ptl;  
 int midPoint;
 int xr, yr;

 if (dlg->getOp() == InsertRemovePixelsDialog::InsertVert || dlg->getOp() == InsertRemovePixelsDialog::RemoveVert)
  {
   midPoint = selLayer->imageBuf.width()/2;
   pt = ptsList.first();
   ptl = ptsList.last();
   if (pt->x() > midPoint)
    {
     fp = QPoint(selLayer->imageBuf.width()-1,pt->y());
     lp = QPoint(0,ptl->y());      
    }
   else
    {
     fp = QPoint(0,pt->y());
     lp = QPoint(selLayer->imageBuf.width()-1,ptl->y());
    }
   draw->transPoint(lp.x(), fp.y(), &xr, &yr);
   wrapPoint = QPoint(xr, yr);
  }
 else if (dlg->getOp() == InsertRemovePixelsDialog::InsertHor || dlg->getOp() == InsertRemovePixelsDialog::RemoveHor)
  {
   midPoint = selLayer->imageBuf.height()/2;
   pt = ptsList.first();
   ptl = ptsList.last();
   if (pt->y() > midPoint)
    {
     fp = QPoint(pt->x(), selLayer->imageBuf.height()-1);
     lp = QPoint(ptl->x(), 0);
    }
   else
    {
     fp = QPoint(pt->x(), 0);
     lp = QPoint(ptl->x(), selLayer->imageBuf.height()-1);
    }
   draw->transPoint(fp.x(), lp.y(), &xr, &yr);
   wrapPoint = QPoint(xr, yr);
  } 
}

void InsertRemovePixelsTool::doAction()
{  
  doc->setActiveLayer(0);
  if (dlg->getSelType() == InsertRemovePixelsDialog::PointType)
   doActionPoint();
  else if (dlg->getSelType() == InsertRemovePixelsDialog::VectorType)
  {
    if (drawStarted)
      endCurve();
    doActionVector();
  }
}

void InsertRemovePixelsTool::doActionVector()
{
  if (!ptsList.count()) return;
  //add corner points
  int midPoint, i;
  bool firstToLast;
  int insert = (dlg->getOp() == InsertRemovePixelsDialog::InsertHor) || (dlg->getOp() == InsertRemovePixelsDialog::InsertVert);
  int horizontal = (dlg->getOp() == InsertRemovePixelsDialog::InsertHor) || (dlg->getOp() == InsertRemovePixelsDialog::RemoveHor);
  QPointArray borderPts, borderPts1, curvePoints, curvePoints1;
  QPoint *pt;
  QPoint *ptf;
  Layer *selLayer = doc->getLayer();
  if (!selLayer) return;

// ZZZZ 
  if (horizontal && dlg->isFindPath())
   {
    int numPix = dlg->getNum();
    Layer *newLayer=NULL;
    uint fillCol;
    int maxy = selLayer->imageBuf.height();
    int i, y, insRemPoint[maxy];
    int selLayerPos = doc->getLayerNum();

//printf("new ins %d rem %d points %d\n", dlg->getOp() == InsertRemovePixelsDialog::InsertHor, dlg->getOp() == InsertRemovePixelsDialog::RemoveHor, numPix);

     pt = ptsList.first();
     for (i=0; i<maxy; pt = ptsList.next(), i++)
      insRemPoint[(i+1)%maxy]= max(0, min(pt->x(), selLayer->imageBuf.width()-numPix));

     doc->putUndoStack(selLayer, ChangeLayerForm);

     fillCol  = selLayer->getFgIndx();

     newLayer = new Layer(selLayer->imageBuf.width() + numPix*(insert ? 1 : -1), selLayer->imageBuf.height(), selLayer->imageBuf.depth(),
                   fillCol, 2, selLayer->getType(), aspectX, aspectY);

     selLayer->imageBuf.copyImageDataTo(&newLayer->imageBuf);
     selLayer->origBuf.copyImageDataTo(&newLayer->origBuf);
     //set bg and fg color from selLayer to newLayer
     newLayer->setFgIndx(selLayer->getFgIndx());
     newLayer->setBgIndx(selLayer->getBgIndx());
     newLayer->setFgCol(selLayer->getFgCol());
     newLayer->setBgCol(selLayer->getBgCol());

     for (y=0; y<selLayer->imageBuf.height(); y++)
      {
       memcpy(newLayer->imageBuf.scanLine(y), selLayer->imageBuf.scanLine(y), insRemPoint[y]);
       memcpy(newLayer->origBuf.scanLine(y), selLayer->origBuf.scanLine(y), insRemPoint[y]);
      }

     if (insert)
      {
       for (y=0; y<selLayer->imageBuf.height(); y++)
	{
	 memset(newLayer->imageBuf.scanLine(y)+insRemPoint[y], fillCol, numPix);
         memcpy(newLayer->imageBuf.scanLine(y)+insRemPoint[y]+numPix, selLayer->imageBuf.scanLine(y)+insRemPoint[y], selLayer->imageBuf.width()-insRemPoint[y]);
	 memset(newLayer->origBuf.scanLine(y)+insRemPoint[y], fillCol, numPix);
         memcpy(newLayer->origBuf.scanLine(y)+insRemPoint[y]+numPix, selLayer->origBuf.scanLine(y)+insRemPoint[y], selLayer->origBuf.width()-insRemPoint[y]);
	}
      }
     else
      for (y=0; y<selLayer->imageBuf.height(); y++)
       {
        memcpy(newLayer->imageBuf.scanLine(y)+insRemPoint[y], selLayer->imageBuf.scanLine(y)+insRemPoint[y]+numPix, selLayer->imageBuf.width()-insRemPoint[y]-numPix);
        memcpy(newLayer->origBuf.scanLine(y)+insRemPoint[y], selLayer->origBuf.scanLine(y)+insRemPoint[y]+numPix, selLayer->origBuf.width()-insRemPoint[y]-numPix);
       }
     doc->setLayer(selLayerPos, newLayer);

     doc->refreshLayer();
     if (selLayer->getType() == MainLayer)
      doc->sendSignals(true, true);
     else
      doc->sendSignals(true, false);
    return;
   }

  if (dlg->getOp() == InsertRemovePixelsDialog::InsertVert || dlg->getOp() == InsertRemovePixelsDialog::RemoveVert)
   {
    int width = selLayer->imageBuf.width();
    midPoint = width/2;
    ptf = ptsList.first();
    if (ptf->x() > midPoint)
     firstToLast = false;
    else
     firstToLast = true;
    size_t pointsNum = ptsList.count()+1;

    QPointArray points(pointsNum);

    for (pt = ptsList.first(), i=1; pt; pt = ptsList.next(), i++)
    	points.setPoint(i, pt->x(), pt->y());

    points.at(pointsNum) = points.at(pointsNum-1);

    if (firstToLast)
     {
      points.at(0).setX(0);
      points.at(pointsNum).setX(width-1);
     }
    else
     points.at(0).setX(0);

    borderPts = QPointArray(ptsList.count()+2);
    borderPts1 =  QPointArray(ptsList.count()+2);
    curvePoints =  QPointArray(ptsList.count());
    curvePoints1 =  QPointArray(ptsList.count());

    for (pt = ptsList.first(), i=1; pt; pt = ptsList.next(), i++)
     {
      borderPts.setPoint(i, pt->x(), pt->y());
      borderPts1.setPoint(i, pt->x(), pt->y()+1);
      curvePoints.setPoint(i-1, pt->x(), pt->y());
      curvePoints1.setPoint(i-1, pt->x(), pt->y() + 1);
     }

    if (firstToLast)
     {
      borderPts.setPoint(0, 0, 0);
      borderPts.setPoint(ptsList.count()+1, selLayer->imageBuf.width()-1, 0);
     }
    else
     {
      borderPts.setPoint(0, selLayer->imageBuf.width()-1, 0);
      borderPts.setPoint(ptsList.count()+1, 0, 0);
     }

    if (firstToLast)
     {
      borderPts1.setPoint(0, 0, selLayer->imageBuf.height()-1);
      borderPts1.setPoint(ptsList.count()+1, selLayer->imageBuf.width()-1, selLayer->imageBuf.height()-1);
     }
    else
     {
      borderPts1.setPoint(0, selLayer->imageBuf.width()-1, selLayer->imageBuf.height()-1);
      borderPts1.setPoint(ptsList.count()+1, 0 , selLayer->imageBuf.height()-1);
     }
   }
  else if (dlg->getOp() == InsertRemovePixelsDialog::InsertHor || dlg->getOp() == InsertRemovePixelsDialog::RemoveHor)
   {
    midPoint = selLayer->imageBuf.height()/2;
    ptf = ptsList.first();
    if (ptf->y() > midPoint)
      firstToLast = false;
    else
      firstToLast = true;

    borderPts = QPointArray(ptsList.count()+2);
    borderPts1 =  QPointArray(ptsList.count()+2);
    curvePoints =  QPointArray(ptsList.count());
    curvePoints1 =  QPointArray(ptsList.count());
    for (pt = ptsList.first(), i=1; pt; pt = ptsList.next(), i++)
     {
      borderPts.setPoint(i, pt->x(), pt->y());
      borderPts1.setPoint(i, pt->x()+1, pt->y());
      curvePoints.setPoint(i-1, pt->x(), pt->y());
      curvePoints1.setPoint(i-1, pt->x() + 1, pt->y());
     }

    if (firstToLast)
     {
      borderPts.setPoint(0, 0, 0);
      borderPts.setPoint(ptsList.count()+1, 0, selLayer->imageBuf.height()-1);
     }
    else
     {
      borderPts.setPoint(0, 0, selLayer->imageBuf.height()-1);
      borderPts.setPoint(ptsList.count()+1, 0, 0);
     }

    if (firstToLast)
     {
      borderPts1.setPoint(0, selLayer->imageBuf.width()-1, 0);
      borderPts1.setPoint(ptsList.count()+1, selLayer->imageBuf.width()-1, selLayer->imageBuf.height()-1);
     }
    else
     {
      borderPts1.setPoint(0, selLayer->imageBuf.width()-1, selLayer->imageBuf.height()-1);
      borderPts1.setPoint(ptsList.count()+1, selLayer->imageBuf.width()-1, 0);
     }
   }
  doc->putUndoStack(selLayer, ChangeLayerForm);

  crackLayer(borderPts, borderPts1);
  doc->refreshLayer();

  if (selLayer->getType() == MainLayer)
   doc->sendSignals(true, true);
  else
   doc->sendSignals(true, false);
}

void InsertRemovePixelsTool::doActionPoint()
{
 uchar *brushBuf = 0;
 Layer *selLayer = doc->getLayer();
 Layer *mainLayer = doc->getMainLayer();
 if (!selLayer || !mainLayer) return;
 QRect br = selLayer->layerPos();
 int x =  dlg->getPos() - 1 - br.x();
 int y =  mainLayer->layerPos().height() - (dlg->getPos()-1) - br.y();
 int wid = selLayer->imageBuf.width();
 int hei = selLayer->imageBuf.height();
 int numPix = dlg->getNum();

 if (dlg->getOp() == InsertRemovePixelsDialog::InsertHor || dlg->getOp() == InsertRemovePixelsDialog::InsertVert)
  {
   if (dlg->isAfter())
    {x++; y--;}
  }
 else if (dlg->isAfter() && dlg->getOp() == InsertRemovePixelsDialog::RemoveVert)
  {
   if (y<numPix)
    {
     numPix = y;
     y=0;
    }
   else
    y -=numPix;
  }
 else if (dlg->isBefore() && dlg->getOp() == InsertRemovePixelsDialog::RemoveHor)
  {
   if (x < numPix)
    {
     numPix = x;
     x=0;
    }
   else
    x -=numPix-1;
  }
 
 if (dlg->getOp() == InsertRemovePixelsDialog::InsertHor)
  {
   emit putUndo(true, x, numPix, true);

   if (insertDuplicate)
    {
     ApaintImage image = selLayer->imageBuf.copy(min(x, wid-numPix), 0, numPix, hei);

     if (duplicateType == LINE)
      for (int yy=0; yy<hei; ++yy)
       {
	uint value = (image.type() == LOAD_TRUE_COL) ? image.pixel(0, yy) : image.pixelIndex(0, yy);
	for (int xx=1; xx<numPix; ++xx)
	 image.setPixel(xx, yy, value);
       }

     brushBuf = image.scanLine(0);
     selLayer->insertRemovePixelsHor(x, numPix, true, useBgCol, true, brushBuf);
     brushBuf = 0;
    }
   else if (bList->useBrush())
    {
     brushBuf = getBrushBuffer(numPix, hei, x, 0);
     selLayer->insertRemovePixelsHor(x, numPix, true, useBgCol, true, brushBuf);
    }
   else
    selLayer->insertRemovePixelsHor(x, numPix, true, useBgCol, false, 0);
  }
 else if (dlg->getOp() == InsertRemovePixelsDialog::InsertVert)
  {
   emit putUndo(false, y, numPix, true);

   if (insertDuplicate)
    {
     ApaintImage image = selLayer->imageBuf.copy(0, max(0, y-numPix), wid, numPix);

     if (duplicateType == LINE)
      for (int yy=0; yy<numPix-1; ++yy)
       for (int xx=0; xx<wid; ++xx)
	{
	 if (image.type() == LOAD_TRUE_COL)
	  image.setPixel(xx, yy, image.pixel(xx, numPix-1));
	 else
	  image.setPixel(xx, yy, image.pixelIndex(xx, numPix-1));
	}

     brushBuf = image.scanLine(0);
     selLayer->insertRemovePixelsVert(y, numPix, true, useBgCol, true, brushBuf);
     brushBuf = 0;
    }
   else if (bList->useBrush())
    {
     brushBuf = getBrushBuffer(wid, numPix, 0, y);
     selLayer->insertRemovePixelsVert(y, numPix, true, useBgCol, true, brushBuf);
    }
   else
    selLayer->insertRemovePixelsVert(y, numPix, true, useBgCol, false, 0);
  }
 else if (dlg->getOp() == InsertRemovePixelsDialog::RemoveHor)
  {
   numPix = QMIN(br.width()-x, numPix);   
   emit putUndo(true, x, numPix, false);
   selLayer->insertRemovePixelsHor(x, numPix, false, useBgCol, false, 0);
  }
 else if (dlg->getOp() == InsertRemovePixelsDialog::RemoveVert)
  {
   numPix = QMIN(br.height()-y, numPix);
   emit putUndo(false, y, numPix, false);
   selLayer->insertRemovePixelsVert(y, numPix, false, useBgCol, false, 0);
  }

 doc->refreshLayer();

 if (doc->getLayerType() == MainLayer)
  doc->sendSignals(true, true);
 else
  doc->sendSignals(true, false);
   
 if (brushBuf)
  delete[] brushBuf; 
}

QRect InsertRemovePixelsTool::getBorderRect()
{
  if (!ptsList.count()) return QRect();
  QPoint *pt;
  int i;
  QPointArray borderPts = QPointArray(ptsList.count());  
  for (pt = ptsList.first(), i=0; pt; pt = ptsList.next(), i++)
    borderPts.setPoint(i, pt->x(), pt->y());

  return (borderPts.boundingRect()).normalize();
}

void InsertRemovePixelsTool::endCurve()
{
  QPoint fp, lp, wp;
  QPoint *plp;
  
  drawStarted = false;
  getFirstLastPoint(fp, lp, wp);
  firstx = lastx = lp.x();
  firsty = lasty = lp.y();
  if (!inFirstTile)
   plp = new QPoint(wp);
  else
   plp = new QPoint(lp);
  ptsList.append(plp); 
}

/* return number of equal pixels on left and on right */
void InsertRemovePixelsTool::getLeftRightLen(unsigned char *line, int pos, int max, unsigned char pixel, int *lenLeft, int *lenRight)
{
 int len;

 len = 0;
 while (((pos + len)<max) && (line[pos + len]==pixel))
  len++;
 *lenRight = len;

 len = 0;
 while (((pos - len)>0) && (line[pos - len]==pixel))
  len++;
 *lenLeft = len;
}

/* return number of equal pixels up and down */
/*
void InsertRemovePixelsTool::getLeftRightLen(unsigned char *line, int pos, int max, int bpl, unsigned char pixel, int *lenUp, int *lenDown)
{
 int len;

 len = 0;
 while (((pos + len)<max) && (line[(pos + len)*bpl]==pixel))
  len++;
 *lenDown = len;

 len = 0;
 while (((pos - len)>0) && (line[(pos - len)*bpl]==pixel))
  len++;
 *lenUp = len;
}
*/

void InsertRemovePixelsTool::mousePressed(QMouseEvent *e, QPainter *pc, QPainter *pi)
{  
  int xin = (int)(e->pos().x() / zoomX);
  int yin = (int)(e->pos().y() / zoomY);
  int x, y;
  QPoint fp, lp, wp;
  QPoint *pfp;
  Layer *layer = doc->getLayer();
  int curTile;
  if (!layer) return;
  // you can only do this on main layer!!
  if (layer != doc->getMainLayer())
    return;

  draw->transPoint(xin, yin, &x, &y );

  btnPressed = true;   
  
  if (dlg->getSelType() == InsertRemovePixelsDialog::VectorType)
  {
    if (!drawStarted)
      firstTile = layer->getTileNum(xin, yin);
    curTile = layer->getTileNum(xin, yin);

// ZZZZ
if (dlg->isFindPath())
 {
  int i, oldX, randDir, length, lenLeft, lenRight, jump, numTrials, moveFailed=TRUE, localRetry;
  int wid = layer->imageBuf.width();
  int hei = layer->imageBuf.height();
  ApaintImage img = layer->imageBuf;
  unsigned char *thisLine, *nextLine, pixel;
  unsigned short path[max(wid, hei)];
  int haveSpace;

  /* make artificial path and immediately apply it */

 /* add true color support later */
 if (img.type() != LOAD_CMAP)
  return;

  if ((curTile == -1) || (firstTile != curTile))
   return;
  firstx = lastx = x;
  firsty = lasty = y;
  inFirstTile = true;

thisLine = img.scanLine(y);
pixel = thisLine[x];

ptsList.clear();

i=0;
localRetry = 0;
while (i<hei)
 {
  oldX = x; /* remember original x before moving it */
  thisLine = img.scanLine((i+y)%hei);
  nextLine = img.scanLine((i+y+1)%hei);

  moveFailed = TRUE;
  randDir = rand();

  getLeftRightLen(thisLine, x, wid, pixel, &lenLeft, &lenRight);

  haveSpace = TRUE;
  if ((dlg->getNum()>1) && dlg->rbRemoveHor->isChecked())
   if ((lenLeft + lenRight-3)<dlg->getNum())
    haveSpace = FALSE;

  if (haveSpace)
   {
    /* prefer going up */
    if (((randDir%3) >= 2) && (nextLine[x]==pixel)) /* try up */
     moveFailed = FALSE;

    if (moveFailed && (randDir&1)) /* right */
     {
      length = lenRight-1;
      jump = 0;
      /* find one pixel up which is same color */
      if (length>0)
       {
	numTrials = 0;
	jump = (rand()%length)+1;
	while ((numTrials<=length) &&  (nextLine[x+jump]!=pixel))
	 { jump = (rand()%length)+1; numTrials++; }
       }
      if (nextLine[x+jump]==pixel)
       { x+=jump; moveFailed = FALSE; }
     }
    if (moveFailed) /* left */
     {
      length = lenLeft-1;
      jump = 0;
      /* find one pixel up which is same color */
      if (length>0)
       {
	numTrials = 0;
	jump = (rand()%length)+1;
	while ((numTrials<=length) && (nextLine[x-jump]!=pixel))
	 { jump = (rand()%length)+1; numTrials++; }
       }
      if (nextLine[x-jump]==pixel)
       { x-=jump; moveFailed = FALSE; }
     }

    if (moveFailed && (!(randDir&1))) /* try right again, if we did not do it so far */
     {
      length = lenRight-1;
      jump = 0;
      /* find one pixel up which is same color */
      if (length>0)
       {
	numTrials = 0;
	jump = (rand()%length)+1;
	while ((numTrials<=length) && (nextLine[x+jump]!=pixel))
	 { jump = (rand()%length)+1; numTrials++; }
       }
      if (nextLine[x+jump]==pixel)
       { x+=jump; moveFailed = FALSE; }
     }
   }
     
  if (!moveFailed)
   {
    int lenL, lenR;

    // position the actual cut in the middle of allowable area
    lenR = 0; // test how far can we go
    while (((x+lenR)<wid) && (thisLine[x+lenR]==pixel) && (nextLine[x+lenR]==pixel))
     lenR++;

    lenL = 0; // test how far can we go
    while (((x-lenL)>0) && (thisLine[x-lenL]==pixel) && (nextLine[x-lenL]==pixel) )
     lenL++;

    x = x - lenL/2 + lenR/2;

    path[(i+y)%hei] = x;
    i++;
   }

  /* check if last position and next position are connected */
  if ((i==hei) && (!moveFailed))
   {
    int oldLenLeft;
    int oldLenRight;

    getLeftRightLen(thisLine, oldX, wid, pixel, &oldLenLeft, &oldLenRight);

    x = path[(i+y)%hei];
    thisLine = img.scanLine((i+y)%hei);
    getLeftRightLen(thisLine, x, wid, pixel, &lenLeft, &lenRight);
    moveFailed = ((oldX-oldLenLeft)<=(x+lenRight)) && ((x-lenLeft)<=(oldX+oldLenRight)) ? 0 : 1;
//    printf("the end x %d [%d %d] next %d [%d %d] intersect %d\n", oldX, oldLenLeft, oldLenRight, x, lenLeft, lenRight, ((oldX-oldLenLeft)<=(x+lenRight)) && ((x-lenLeft)<=(oldX+oldLenRight)));
   }
     
  if (moveFailed)    
   {
    localRetry++;
    if ((i>4) && (localRetry<100))
     {
      i = rand()%i+1; /* backtrack part of random path and try again */
      x = path[(i+y)%hei];
     }
    else
     i=MAX_IMAGE; /* dead end, give up */
   }
 }

 if ((i>=MAX_IMAGE) || moveFailed)
  {
   QApplication::beep();
   ptsList.clear();
   return;
  }
 else
  {
   if ((dlg->getNum()>1) && dlg->rbRemoveHor->isChecked())
    for (i=0; i<hei; i++)
     if (path[i]>(dlg->getNum()/2))
      path[i] -= dlg->getNum()/2;

   for (i=0; i<hei; i++)
    {
     QPoint *pt  = new QPoint(path[i], i);
     ptsList.append(pt);
    }
   endCurve();
   doAction();
   ptsList.clear();
   return;
  }
 }







    if (e->button() == LeftButton)
    {
      if ((curTile == -1) || (firstTile != curTile)) return;
      firstx = lastx = x;
      firsty = lasty = y;
      inFirstTile = true;     
      if (!drawStarted)
      {
       drawStarted = true;
       QRect pos = getBorderRect();
       ptsList.clear();
       doc->refreshLayer(pos);
      }
      QPoint *pt  = new QPoint(x, y);
      ptsList.append(pt);
      if (ptsList.count() == 1)
      {
       getFirstLastPoint(fp, lp, wp);
       pfp = new QPoint(fp);
       ptsList.insert(0, pfp);
       drawScaleLine(pc, ptsList.first()->x(), ptsList.first()->y(), x,  y);
      }
    }
    else if (e->button() == RightButton)
    {     
     endCurve();
     doAction(); 
    }
  }
  else if (dlg->getSelType() == InsertRemovePixelsDialog::PointType)
  {
    firstx = lastx = x;
    firsty = lasty = y;
  } 
}

void InsertRemovePixelsTool::mouseMoved(QMouseEvent *e, QPainter *pc, QPainter *pi)
{  
  int xin = (int)(e->pos().x() / zoomX);
  int yin = (int)(e->pos().y() / zoomY);
  int x, y;
  int wid, hei;
  QPoint lp, fp, wp;
  int curTile;
  
  Layer *layer = doc->getLayer();
  if (!layer) return;
// you can only do this on main layer!!
  if (layer != doc->getMainLayer())
    return;
if (dlg->isFindPath())
 return;

  wid = layer->imageBuf.width();
  hei = layer->imageBuf.height();
  draw->transPoint(xin, yin, &x, &y );
  
  curTile = layer->getTileNum(xin, yin);
  if (firstTile == curTile)
    inFirstTile = true;

  if (dlg->getSelType() == InsertRemovePixelsDialog::PointType)
  {
     if (!btnPressed) return;

     if (dlg->getOp() == InsertRemovePixelsDialog::InsertHor || dlg->getOp() == InsertRemovePixelsDialog::RemoveHor)
     {
       drawScreenRect(pc, firstx, 0, firstx+(lastx-firstx), hei);
       lastx = x;
       lasty = y;
       drawScreenRect(pc, firstx, 0, firstx+(lastx-firstx), hei);

       dlg->setNum(QABS(lastx-firstx));
     }
     else if (dlg->getOp() == InsertRemovePixelsDialog::InsertVert || dlg->getOp() == InsertRemovePixelsDialog::RemoveVert)
     {
       drawScreenRect(pc, 0, firsty, wid, firsty+(lasty-firsty));
       lastx = x;
       lasty = y;
       drawScreenRect(pc, 0, firsty, wid, firsty+(lasty-firsty));

       dlg->setNum(QABS(lasty-firsty));
     }
  }
  else if (dlg->getSelType() == InsertRemovePixelsDialog::VectorType)
  {
    if (!drawStarted) return;
    if (curTile != firstTile)
    {
     if (inFirstTile)
     {        
      drawScaleLine(pc, firstx, firsty, lastx, lasty);
      getFirstLastPoint(fp, lp, wp);
      lastx = wp.x();
      lasty = wp.y();
      drawScaleLine(pc, firstx, firsty, lastx, lasty);
      inFirstTile = false;
     }
    }
    else if (btnPressed)
    {
      QPoint *pt  = new QPoint(x, y);
      ptsList.append(pt);
      drawScaleLine(pc, firstx, firsty, x, y);
      firstx = lastx = x;
      firsty = lasty = y;  
    }
    else
    {
      drawScaleLine(pc, firstx, firsty, lastx, lasty);
      lastx = x;
      lasty = y;
      drawScaleLine(pc, firstx, firsty, lastx, lasty);
    }
  }
}

void InsertRemovePixelsTool::mouseReleased(QMouseEvent *e, QPainter *pc, QPainter *pi)
{
  if (!btnPressed) return;
  btnPressed = false;
  
  Layer *selLayer = doc->getLayer();
  if (!selLayer) return;
  // you can only do this on main layer!!
  if (selLayer != doc->getMainLayer())
    return;

  QRect br = selLayer->layerPos();
  if (dlg->getSelType() == InsertRemovePixelsDialog::PointType)
   {
     if (e->button() == MidButton)
      dlg->setPos(firstx+1, br.height()-firsty);
     else
      {
       useBgCol = e->button() == RightButton;
       if (dlg->getOp() == InsertRemovePixelsDialog::RemoveHor)
        {
         if (dlg->isAfter())
            dlg->setPos(QMIN(firstx, lastx), 0);
          else
           dlg->setPos(QMAX(firstx, lastx), 0);
        }
       else if ( dlg->getOp() == InsertRemovePixelsDialog::InsertHor)
        {
         if (dlg->isAfter())
            dlg->setPos(QMAX(firstx, lastx), 0);
          else
           dlg->setPos(QMIN(firstx, lastx), 0);
        }
       else if (dlg->getOp() == InsertRemovePixelsDialog::RemoveVert)
        {
          if (dlg->isAfter())
            dlg->setPos(0, br.height()-QMAX(firsty, lasty));
          else
           dlg->setPos(0, br.height()-QMIN(firsty, lasty));
        }
       else if (dlg->getOp() == InsertRemovePixelsDialog::InsertVert)
        {
           if (dlg->isAfter())
            dlg->setPos(0, br.height()-QMIN(firsty, lasty));
          else
           dlg->setPos(0, br.height()-QMAX(firsty, lasty));
        }
       doAction();
      }
   }  
}
