/***************************************************************************
                          toolbase.h  -  description
                             -------------------
    begin                : Sun Feb 2 2003
    copyright            : (C) 2003-2011 Arahne, d.o.o.
    email                : arahne@arahne.si
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef TOOLBASE_H
#define TOOLBASE_H

#include <qobject.h>
#include <qtoolbar.h>
#include <qcursor.h>
#include <qrect.h>
#include <qpixmap.h>
#include <qimage.h>
#include <qpushbutton.h>
#include "apaintimage.h"

class PropObject: public QObject {
  Q_OBJECT
public:
 PropObject(QWidget *w);
 ~PropObject(){};

 void show();
 void hide();
 void setValidShow(bool enable);
 QWidget *pWidget() {return propWidget;}

signals:
 void hidden();
 void shown();     
private:
 bool widgetOK, validShow;
 QWidget *propWidget; 
};  
  
enum ToolType{DrawTool, SelectionTool, ActionTool};
enum ToolChangeType {NoChange, FillChangeType};

                
class ToolBase: public QObject {
   Q_OBJECT
public: 
  ToolBase(ToolType typ, PropObject *propObj);
  ~ToolBase();
  
 bool isActive();  
 bool usePropertyWin();
 void setAspectRatio(double ax, double ay);
 PropObject *pObject();
 ToolType getType() {return type;} 
 QString getToolName() {return toolName;}
 void setPropWindow(PropObject *propObj);
 void enablePropWindow(bool enable);
 bool usingAutoScroll() {return useAutoScroll; }
 void setToolChangeType(ToolChangeType typ) {changeType = typ;}
 ToolChangeType getChangeType() {return changeType;}
 
 virtual void begin();
 virtual void end();
 virtual void resetTool();
 virtual void setZoom(double zx, double zy);
 virtual void doAction();
 virtual void mousePressed(QMouseEvent *e, QPainter *pc, QPainter *pi);
 virtual void mouseReleased(QMouseEvent *e, QPainter *pc, QPainter *pi);
 virtual void mouseMoved(QMouseEvent *e, QPainter *pc, QPainter *pi);
 virtual void mouseDoubleClick(QMouseEvent *e, QPainter *pc, QPainter *pi);
 virtual void keyPress(QKeyEvent *e, QPainter *pc, QPainter *pi);
 virtual void keyRelease(QKeyEvent *e, QPainter *pc, QPainter *pi);
 virtual void drawPointer(QPainter *p, int x, int y);
 virtual void beforeRefresh(QPainter *p);
 virtual void afterRefresh(QPainter *p);
 virtual bool doChangeTool(ToolChangeType chgType);
 virtual QPixmap getToolPix();
 virtual void showPropWindow(bool enable);
 virtual void showCommonProp(bool enable);
 virtual void repaint(QPainter *pc, QPainter *pi) {}

 QCursor& getCursor();

signals:
 void activated(QObject *sender);
 void activatedDouble(QObject *sender);
 
protected slots:
  void activateSlot();
  void activateSlotDouble();  
  
protected:

  void refreshRect(int xs1, int ys1, int xs2, int ys2, bool finalDraw);
  void refreshRect(int xs1, int ys1, int xs2, int ys2, bool useFg, bool finalDraw);
  void clearRect(int x1, int y1, int x2, int y2);
  void setToggleBtn(QPixmap pix, QString text);
  void setToggleBtn(QPixmap pix, QPixmap pix2);
  void setUsepWidget(bool use);
  void setAutoScroll(bool enable){ useAutoScroll = enable;}
  QRect normalRect(int xs1, int ys1, int xs2, int ys2);
  QRect normalQuadRect(int xs1, int ys1, int xs2, int ys2, double aspx, double aspy);
  QRect normalCircle(int xs1, int ys1, int xs2, int ys2, double aspx, double aspy);
  QRect normalEllipse(int xs1, int ys1, int xs2, int ys2);

  QPixmap toolPix, toolPix2;
  QCursor cur;
  ToolType type;
  double aspectX, aspectY;
  double zoomX, zoomY, invZoomX, invZoomY, oldZoomX, oldZoomY;

  
private:
  bool active;         // iam active
  bool usepWidget;     // i have property widget
  bool useAutoScroll; //...
  QPixmap icon;
  QPushButton *tButton;
  PropObject *pObj;  
  QString toolName;
  ToolChangeType changeType;
};

#endif
