#include <gutenprint/gutenprint.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include "print.h"

static char *dstFile = NULL;
static char *commandStr = NULL;
static int doRemove = FALSE;

static FILE *fout, *fgutenp;
static char *fname, *fnamegutenp;
static int horDpi;

static printBlock *pBlock=NULL;
static unsigned short *lineToBlock;
static int numPrintBlocks;

static const char *Image_get_appname(stp_image_t *image);
static stp_image_status_t Image_get_row(stp_image_t *image, unsigned char *data, size_t byte_limit, int row);
static int Image_height(stp_image_t *image);
static int Image_width(stp_image_t *image);

static stp_image_t stpImage =
{
 NULL,
 NULL,
 Image_width,
 Image_height,
 Image_get_row,
 Image_get_appname,
 NULL,
 NULL
};

double global_c_level;
double global_c_gamma = 1.0;
double global_m_level;
double global_m_gamma = 1.0;
double global_y_level;
double global_y_gamma = 1.0;
double global_k_gamma = 1.0;
double global_gamma = 1.0;
double ink_limit;
double density;
char *printer;
char *ink_type;
char *stp_resolution;
char *media_type;
char *media_size;
char *dither_algorithm = "Ordered";

int printx, printy, offsetx, offsety;
static int bidirectionalPrinting=FALSE;

static void writefunc(void *file, const char *buf, size_t bytes)
{
 FILE *prn = (FILE *)file;
 fwrite(buf, 1, bytes, prn);
}

static int Image_width(stp_image_t *image)
{ return printx; }

static int Image_height(stp_image_t *image)
{ return printy; }

static const char *Image_get_appname(stp_image_t *image)
{ return("agutenp: Arahne's interface to gutenprint"); }


static stp_image_status_t Image_get_row(stp_image_t *image, unsigned char *data, size_t byte_limit, int row)
{
 int x;
 unsigned char *dat, *cmapPtr;
 printBlock *pb = &pBlock[lineToBlock[row]];

 if (!pb->data)
  {
   /* read the print block structure */
   if (!fread(pb, sizeof(printBlock), 1, fout))
    return(STP_IMAGE_STATUS_ABORT);
   /* read a block of data from file if we do not have it yet */
   pb->data = malloc(pb->bpl*pb->unreadLines);
   if (!pb->data)
    return(STP_IMAGE_STATUS_ABORT);
   /* we presume that reading of rows is close to consecutive - we must not jump by more than one block height */
   if (pb->unreadLines!=fread(pb->data, pb->bpl, pb->unreadLines, fout))
    return(STP_IMAGE_STATUS_ABORT);
//fprintf(stderr, "reading block %d lines %d type %d\n", lineToBlock[row], pb->unreadLines, pb->bandType);
  }
 dat = &pb->data[pb->bpl*(row-pb->ypos)];
 /* now interpret the image type and fill in the data with the correct format of data */

 switch (pb->bandType)
  {
   case IMG_BW:   
    for (x=0; x<printx; x++)
     if (!(dat[x>>3]&(128>>(x&7))))
      { *data++ = 255; *data++ = 255; *data++ = 255; }
     else
      { *data++ = 0; *data++ = 0; *data++ = 0; }
    break;
   case IMG_CMAP:   
    for (x=0; x<printx; x++)
     {
      cmapPtr = &pb->cmap[(*dat++)*3];
      *data++ = *cmapPtr++; *data++ = *cmapPtr++; *data++ = *cmapPtr;
     }
    break;
   case IMG_TRUE_COL:   
    memcpy(data, dat, printx*3);
    break;
  }

 /* if we have read all the lines from this block, then free it from the memory */
 pb->unreadLines--;
 if (pb->unreadLines==0)
  {
//fprintf(stderr, "freed block %d\n", lineToBlock[row]);
   free(pb->data);
   pb->data=NULL;
  }
 return(STP_IMAGE_STATUS_OK);
}

/* use it to remove new line character */
static void removeLast(char *str)
{
 str[strlen(str)-1]=0;
}

static void readParameterFile(FILE *fp)
{
 char buf[500];
 int i, j, num, pos;

 fgets(buf, 500, fp); removeLast(buf);
 fname = strdup(buf);

 fgets(buf, 500, fp); removeLast(buf);
 if (strlen(buf))
  dstFile = strdup(buf);
 fgets(buf, 500, fp); removeLast(buf);
 if (strlen(buf))
  commandStr = strdup(buf);
 fgets(buf, 500, fp); removeLast(buf);
 doRemove = atoi(buf);

 fgets(buf, 500, fp); removeLast(buf);
 sscanf(buf, "%d %d %d %d", &printx, &printy, &offsetx, &offsety);

 fgets(buf, 500, fp); removeLast(buf);
 bidirectionalPrinting = atoi(buf);

 fgets(buf, 500, fp); removeLast(buf);
 sscanf(buf, "%lf %lf %lf %lf %lf %lf %lf %lf %lf %lf", &global_c_level, &global_c_gamma, &global_m_level, &global_m_gamma,  &global_y_level, &global_y_gamma, &global_k_gamma, &global_gamma, &ink_limit, &density);

 fgets(buf, 500, fp); removeLast(buf);
 printer = strdup(buf);

 fgets(buf, 500, fp); removeLast(buf);
 ink_type = strdup(buf);

 fgets(buf, 500, fp); removeLast(buf);
 media_type = strdup(buf);

 fgets(buf, 500, fp); removeLast(buf);
 media_size = strdup(buf);

 fgets(buf, 500, fp); removeLast(buf);
 stp_resolution = strdup(buf);

 fgets(buf, 500, fp); removeLast(buf);
 horDpi = atoi(buf);

 fgets(buf, 500, fp); removeLast(buf);
 numPrintBlocks = atoi(buf);

 pBlock = calloc(sizeof(printBlock), numPrintBlocks);
 pos=0;
 lineToBlock = calloc(sizeof(unsigned short), printy);
 for (i=0; i<numPrintBlocks; i++)
  {
   fgets(buf, 500, fp); removeLast(buf);
   num = atoi(buf);
   for (j=0; j<num; j++)
    if (pos<printy)
     lineToBlock[pos++] = (unsigned short) i;
  } 
}

/* return -1 if process can't be created, 0 on failure, 1 on success */
static int executeCommand(char *command)
{
 int status;

 status = system(command);
 if(status<0)
  return(-1);
 if (WIFEXITED(status))
  if (WEXITSTATUS(status)==0)
   return(TRUE);
  else
   return(FALSE);
 else
  return(FALSE);
}

/* does the proper thing (copy, move or print) with the finished print file */
static int enqPrintFile(char *srcFile)
{
 char command[500];
 int status;

 if (!commandStr) /* need to move the file from srcFile to dstFile */
  {
   if (rename(srcFile, dstFile)==0)
    return(PRT_OK);
   else /* rename fails across filesystems, so use mv command */
    strcpy(command, "/bin/mv -f");
  }
 else
  strcpy(command, commandStr); /* use the provided command */

 /* take the command, add srcfile (and dstfile, if present) */
 strcat(command, " \"");
 strcat(command, srcFile);
 strcat(command, "\" ");
 if (dstFile)
  {
   strcat(command, " \"");
   strcat(command, dstFile);
   strcat(command, "\" ");
  }
 status = executeCommand(command);

 if (doRemove)
  if (access(dstFile, F_OK)==0)
   remove(dstFile);
 if (dstFile)
  free(dstFile);
 if (access(srcFile, F_OK)==0)
  remove(srcFile);

 if (status<0)
  return(PRT_INTERNAL_ERROR);
 else
  return((status==0) ? PRT_IO_ERROR : PRT_OK);
}

int main(int argc, char *argv[])
{
 FILE *fparam;
 const stp_printer_t *the_printer;
 int left, right, top, bottom;
 int x, y;
 int paramsOK=TRUE;
 char ink_set[50]="";

 if (argc<1)
  {
   printf("missing parameter file\n");
   return(-1);
  }
 if ((fparam = fopen(argv[1], "rb" )) == NULL)
// if ((fparam = fopen("/tmp/param", "rb" )) == NULL)
  {
   printf("cannot open parameter file %s\n", argv[1]);
   return(-1);
  }

 fnamegutenp = strdup("/tmp/plibgpXXXXXX");
 mkstemp(fnamegutenp);

 readParameterFile(fparam);
 fclose(fparam);
 remove(argv[1]);

 if ((fout = fopen(fname, "rb" )) == NULL)
  {
   printf("cannot open input file %s\n", fname);
   return(-1);
  }

 if ((fgutenp = fopen(fnamegutenp, "wb" )) == NULL)
  {
   printf("cannot open temporary print file\n");
   return(-1);
  }

 stp_init();

stp_vars_t *stpvars = stp_vars_create();
const stp_vars_t *defaults = stp_default_settings();
stp_vars_copy(stpvars, defaults);

stp_set_driver(stpvars, printer);


 the_printer = stp_get_printer_by_driver(printer);
 if (!the_printer)
  {
   printf("gutenprint unsupported printer: %s\n", printer);
   return(-1);
  }

 stp_set_printer_defaults(stpvars, the_printer);
 stp_set_outfunc(stpvars, writefunc);
 stp_set_errfunc(stpvars, writefunc);
 stp_set_outdata(stpvars, fgutenp);
 stp_set_errdata(stpvars, stderr);
 stp_set_float_parameter(stpvars, "Density", density);
 stp_set_string_parameter(stpvars, "Resolution", stp_resolution);
 if (strlen(ink_type))
  stp_set_string_parameter(stpvars, "InkType", ink_type);
 stp_set_string_parameter(stpvars, "MediaType", media_type);

 /* set the matte ink set for matte paper on R2400 */
 if (!strcmp(printer, "escp2-r2400"))
  if (!strcmp(media_type, "MatteHeavy"))
   strcpy(ink_set, "ultra3matte");

 /* set the matte ink set for R3000 */
 if (!strcmp(printer, "escp2-r3000"))
  strcpy(ink_set, "ultra3matte");

 /* set the matte ink set for R1900 */
 if  (!strcmp(printer, "escp2-r1900"))
  strcpy(ink_set, "cmykromatte");

 /* set the matte ink set for  R800*/
 if (!strcmp(printer, "escp2-r800"))
  strcpy(ink_set, "cmykrbmatte");

 if (strlen(ink_set))
  stp_set_string_parameter(stpvars, "InkSet", ink_set);

printf("inkSet %s\n", ink_set);

 if (strlen(media_size))
  stp_set_string_parameter(stpvars, "PageSize", media_size);
// if (media_size)
//  stp_set_media_size(stpvars, media_size);
// if (dither_algorithm)
//  stp_set_dither_algorithm(stpvars, dither_algorithm);

stp_set_string_parameter(stpvars, "InputImageType", "RGB");
stp_set_string_parameter(stpvars, "ChannelBitDepth", "8");
stp_set_string_parameter(stpvars, "PrintingMode", "Color");

stp_set_string_parameter(stpvars, "ColorCorrection", "Uncorrected");
stp_set_string_parameter(stpvars, "PrintingDirection", bidirectionalPrinting ? "Bidirectional" : "Unidirectional");

stp_get_imageable_area(stpvars, &left, &right, &bottom, &top);

stp_describe_resolution(stpvars, &x, &y);

 if (x < 0)
   x = 300;
 if (y < 0)
   y = 300;

 /* stp_papersize_get_height(pt); you may need this function in the future */

 /* we always render things at 1:1 size */
/*
 if (horDpi==150)
  stp_set_scaling(stpvars, -150); // this means we are saying we have 150 dpi
 if (horDpi==300)
  stp_set_scaling(stpvars, -300); // this means we are saying we have 300 dpi
 if (horDpi==600)
  stp_set_scaling(stpvars, -600); // this means we are saying we have 600 dpi
 if (horDpi==1200)
  stp_set_scaling(stpvars, -1200); // this means we are saying we have 1200 dpi
 if (horDpi==360)
  stp_set_scaling(stpvars, -360); // this means we are saying we have 360 dpi
 if ((horDpi==720)||(horDpi==1440))
  stp_set_scaling(stpvars, -720); // this means we are saying we have 720 dpi
*/
int scaleDpi;
 if (horDpi==1440)
  scaleDpi = 720;
 else if (horDpi==2880)
  scaleDpi = 1440;
 else
  scaleDpi = horDpi;
  
 stp_set_left(stpvars, max(left, (72*offsetx)/scaleDpi));
 stp_set_top(stpvars, max(top, (72*offsety)/scaleDpi));

 stp_set_width(stpvars,  (72*printx)/scaleDpi);
 stp_set_height(stpvars, (72*printy)/scaleDpi);


stp_vars_t *tmpvars = stp_vars_create_copy(stpvars);
const stp_vars_t *current_vars = stp_printer_get_defaults(stp_get_printer(stpvars));
stp_merge_printvars(tmpvars, current_vars);

if(stp_verify(tmpvars))
 stp_print(tmpvars, &stpImage);
else
 paramsOK = FALSE;

stp_vars_destroy(tmpvars);

 fclose(fout);
 fclose(fgutenp);
 remove(fname); /* remove it immediately so we lower the necessary disk space */
 free(fname);
 fname=NULL;
 if (paramsOK)
  return(enqPrintFile(fnamegutenp));
 else
  {
   remove(fnamegutenp);
   return(-1);
  }
}
