#define EPSON
/* epson common printer commands file */
#include <string.h>
#include "print.h"

DEPSON unsigned char *ResetPrn(unsigned char *bufPtr)
{
 *bufPtr++ = 27;
 *bufPtr++ = '@';
 return(bufPtr);
}

DEPSON unsigned char *FormFeed(unsigned char *bufPtr)
{
 *bufPtr++ = 12;
 return(bufPtr);
}

DEPSON unsigned char *SetColorCMYK(unsigned char CMYKcomp, unsigned char *bufPtr)
{
 unsigned char color = 0;
// Black 0	Magenta 1	Cyan 2	Violet 3	Yellow 4 Red 5		Green 6

 *bufPtr++ = 27;
 *bufPtr++ = 'r';
 switch (CMYKcomp)
   {
    case 0:
     color = 2;
     break;
    case 1:
     color = 1;
     break;
    case 2:
     color = 4;
     break;
    case 3:
     color = 0;
     break;
   }
 *bufPtr++ = color;
 return (bufPtr);
}

DEPSON unsigned char *SetColorCCMMYK(unsigned char CCMMYKcomp, unsigned char *bufPtr)
{
 unsigned char color = 0, m = 0;
/* Black 0 Magenta 1 Cyan 2 Yellow 4 CyanLight 2 MagentaLight 1 */

 *bufPtr++ = 27;
 *bufPtr++ = '(';
 *bufPtr++ = 'r';
 *bufPtr++ = 2;
 *bufPtr++ = 0;
 switch (CCMMYKcomp)
  {
   case 0: color = 2; break;
   case 1: color = 1; break;
   case 2: color = 4; break;
   case 3: color = 0; break;
   case 4: color = 2; m = 1; break;
   case 5: color = 1; m = 1; break;
  }
 *bufPtr++ = m;
 *bufPtr++ = color;
 return(bufPtr);
}

DEPSON unsigned char *SetGraphicsModeEpson(unsigned char *bufPtr)
{
 *bufPtr++ = 27;
 *bufPtr++ = '(';
 *bufPtr++ = 'G';
 *bufPtr++ = 1;
 *bufPtr++ = 0;
 *bufPtr++ = 1;
 return(bufPtr);
}

DEPSON unsigned char *SetMicroWeave(int microWeaveOn, unsigned char *bufPtr)
{
 *bufPtr++ = 27;
 *bufPtr++ = '(';
 *bufPtr++ = 'i';
 *bufPtr++ = 1;
 *bufPtr++ = 0;
 if (microWeaveOn)
   *bufPtr++ = 1;
 else
   *bufPtr++ = 0;
 return(bufPtr);
}

DEPSON unsigned char *SetUnit(unsigned char unitS, unsigned char *bufPtr)
{ /*5,10,20,30,40,50,60*/
 *bufPtr++ = 27;
 *bufPtr++ = '(';
 *bufPtr++ = 'U';
 *bufPtr++ = 1;
 *bufPtr++ = 0;
 *bufPtr++ = unitS;
 return(bufPtr);
}

DEPSON unsigned char *SetPageLengthU(int PageLength, unsigned char *bufPtr)
{ /*0 < tmp <= 22*/
 *bufPtr++ = 27;
 *bufPtr++ = '(';
 *bufPtr++ = 'C';
 *bufPtr++ = 2;
 *bufPtr++ = 0;
 *bufPtr++ = PageLength % ( /*prnStatus.unitS*/ 256);
 *bufPtr++ = PageLength / 256;
 return (bufPtr);
}

DEPSON unsigned char *SetDirection(unsigned char d, unsigned char *bufPtr)
{
 /* 0 bidirectional, 1 unidirectional */
 *bufPtr++ = 27;
 *bufPtr++ = 'U';
 *bufPtr++ = d;
 return(bufPtr);
}

DEPSON unsigned char *SetPageFormat(int topM, int bottomM, unsigned char *bufPtr)
{
 *bufPtr++ = 27;
 *bufPtr++ = '(';
 *bufPtr++ = 'c';
 *bufPtr++ = 4;
 *bufPtr++ = 0;
 *bufPtr++ = topM % 256;
 *bufPtr++ = topM / 256;
 *bufPtr++ = bottomM % 256;
 *bufPtr++ = bottomM / 256;
 return(bufPtr);
}

/* never used
unsigned char *SetAbsXPos(int position, unsigned char *bufPtr)
{
  *bufPtr++ = 27;
  *bufPtr++ = '$';
  *bufPtr++ = position % 256;
  *bufPtr++ = position / 256;
  return (bufPtr);
}
*/
DEPSON unsigned char *SetRelXPos(int position, unsigned char *bufPtr)
{
 *bufPtr++ = 27;
 *bufPtr++ = 92;
 *bufPtr++ = position % 256;
 *bufPtr++ = position / 256;
 return (bufPtr);
}

/* never used
unsigned char *SetAbsYPos(int position, unsigned char *bufPtr)
{
  *bufPtr++ = 27;
  *bufPtr++ = '(';
  *bufPtr++ = 'V';
  *bufPtr++ = 2;
  *bufPtr++ = 0;
  *bufPtr++ = position % 256;
  *bufPtr++ = position / 256;
  return (bufPtr);
}
*/
DEPSON unsigned char *SetRelYPos(int position, unsigned char *bufPtr)
{
 *bufPtr++ = 27;
 *bufPtr++ = '(';
 *bufPtr++ = 'v';
 *bufPtr++ = 2;
 *bufPtr++ = 0;
 if (position >= 0)
   {
    *bufPtr++ = position % 256;
    *bufPtr++ = position / 256;
   }
 else
   {
    *bufPtr++ = position & 0xff;
    *bufPtr++ = (position >> 8) & 0xff;
   }
 return(bufPtr);
}

DEPSON unsigned char *SetDotSize(unsigned char dotSize, unsigned char *bufPtr)
{
 *bufPtr++ = 27;
 *bufPtr++ = '(';
 *bufPtr++ = 'e';
 *bufPtr++ = 2;
 *bufPtr++ = 0;
 *bufPtr++ = 0;
 *bufPtr++ = dotSize;
 return(bufPtr);
}

DEPSON unsigned char *PrintRasterGraphics(char compression, char vDotDens, char hDotDens,
 char m, int width, unsigned char *bufPtr)
{
 *bufPtr++ = 27;
 *bufPtr++ = '.';
 *bufPtr++ = compression;
 *bufPtr++ = vDotDens;
 *bufPtr++ = hDotDens;
 *bufPtr++ = m;
 *bufPtr++ = width % 256;
 *bufPtr++ = width / 256;
 return(bufPtr);
}

DEPSON unsigned char *SetResolutionRasterImage(int xdpi, int ydpi, unsigned char *bufPtr)
{
 *bufPtr++ = 27;
 *bufPtr++ = '(';
 *bufPtr++ = 'D';
 *bufPtr++ = 4;
 *bufPtr++ = 0;
 *bufPtr++ = 14400 % 256;
 *bufPtr++ = 14400 / 256;
 *bufPtr++ = 14400 / ydpi;
 *bufPtr++ = 14400 / xdpi;
 return(bufPtr);
}


DEPSON unsigned char *TransferRasterImage(char color, char compression, char bitsPerPixel, 
 int width, int height, unsigned char *bufPtr)
{
 switch (color)
  {
   case 0: color = 2; break;
   case 1: color = 1; break;
   case 2: color = 4; break;
   case 3: color = 0; break;
   case 4: color = 18; break;
   case 5: color = 17; break;
  }
 *bufPtr++ = 27;
 *bufPtr++ = 'i';
 *bufPtr++ = color;
 *bufPtr++ = compression;
 *bufPtr++ = bitsPerPixel;
 *bufPtr++ = width % 256;
 *bufPtr++ = width / 256;
 *bufPtr++ = height % 256;
 *bufPtr++ = height / 256;
 return(bufPtr);
}

DEPSON unsigned char *SelectPrintingMode(char mode, unsigned char *bufPtr)
{
 *bufPtr++ = 27;
 *bufPtr++ = '(';
 *bufPtr++ = 'K';
 *bufPtr++ = 2;
 *bufPtr++ = 0;
 *bufPtr++ = 0;
 *bufPtr++ = mode;
 return(bufPtr);
}

/* never used
unsigned char *SetRelXPosNew(int position, unsigned int unit, unsigned char *bufPtr)
{
  *bufPtr++ = 27;
  *bufPtr++ = '(';
  *bufPtr++ = 92;
  *bufPtr++ = 4;
  *bufPtr++ = 0;
  *bufPtr++ = unit % 256;
  *bufPtr++ = unit / 256;
  *bufPtr++ = position % 256;
  *bufPtr++ = position / 256;
  return (bufPtr);
}
*/

DEPSON unsigned char *SelectPrintingSpeed(char speed, unsigned char *bufPtr)
{
 *bufPtr++ = 27;
 *bufPtr++ = '(';
 *bufPtr++ = 's';
 *bufPtr++ = 1;
 *bufPtr++ = 0;
 *bufPtr++ = speed;
 return(bufPtr);
}

/* compress line with RLL encoding */
DEPSON unsigned char *compressLine(unsigned char *lnPtr, unsigned char *printBuf, unsigned char *RLLBuf,int diff)
{
  unsigned char *lPtr, *p = RLLBuf, *endP, *countP, lastPixel;
  int i, counter;

  counter = 0;
  lPtr = lnPtr;
  endP = lPtr + diff;
  lastPixel = *lPtr++;
  while (lPtr < endP)
    {
      while ((lPtr < endP) && (*lPtr == lastPixel))
	{
	  lPtr++;
	  counter++;
	}
      if (counter)
	{
	  counter++;
	  while (counter > 0)
	    {
	      if (counter > 129)
		i = 129;
	      else
		i = counter;
	      *p++ = (unsigned char) (257 - i);
	      *p++ = lastPixel;
	      counter -= i;
	    }
	  lastPixel = *lPtr++;
	  continue;
	}
      countP = p++;
      *countP = 0;
      while ((lPtr < endP) && (*lPtr != lastPixel))
	{
	  *p++ = lastPixel;
	  lastPixel = *lPtr++;
	  (*countP)++;
	  if ((*countP) == 127)
	    break;
	}
      if (*countP)
	{
	  if ((*lPtr != lastPixel) || (lPtr == endP))
	    *p++ = lastPixel;
	  else
	    (*countP)--;
	}
      if (*lPtr == lastPixel)
	counter = 1;
      lastPixel = *lPtr++;
    }
  if (*(endP - 1) != (*(p - 1)))
    {
      if (counter)
	*p++ = 256 - counter;
      else
	*p++ = 0;
      *p++ = *(endP - 1);
    }

  i = p - &RLLBuf[0];
  memcpy(printBuf, RLLBuf, i);
  printBuf += i;
  return(printBuf);
}

/* makes a linearizing dither matrix based on measurments of luminance (L) */
/* point is a measurment, number of measurments must be 17 */
/* ranging from 0 to 255 by step of 16 start with white (~95.0) */
/* and proceed to black (~15), always falling. */
/* Bleed indicates how many values of "black" are "almost" equal, i.e. fully */
/* saturated. The distribution will then be made in such a way, that the extra */
/* points that do not add to saturation, but only to bleeding, will not be present. */
/* This means that fully black will not have all the points, it will still be */
/* dithered. Hopefully this will help against flood on colors with multiple */
/* overlapping saturated colors, e.g. strong dark green, green-violet range */
DEPSON unsigned char *linearizeDither(float *point, float bleed)
{
 unsigned short gamma[256];
 float inters[17], val, first;
 int i, j, s;

 for (i = 1; i < 17; i++)
  if ((point[i-1]-point[i])<0)
   return(0); /* fail on illegal linearization measurments */

 inters[0] = 0.0;
 inters[16] = 256.0;
 /* find equalized linear step intersection values */
 for (i = 1; i < 16; i++)
  {
   val = point[0] - i * ((point[0] - point[17 - 1]) / 16.0);
   j = 0;
   while (val < point[j])
     j++;			/* val lies between point[j-1] and point[j] */
   inters[i] = (j - 1) * 16.0 + 16.0 * ((point[j - 1] - val) / (point[j - 1] - point[j]));
   /* printf("inters %d %f critical %d=%.2f %d=%.2f interdiff %.2f %.2f\n", i, inters[i], j, point[j], j-1, point[j - 1], inters[i]-inters[i-1], point[j-1]-point[j]); */
  }
 /* handle first block separetely to ensure gamma[0]=0 */
 if (bleed>0)
  {
   gamma[0] = (unsigned short) (255 * 16 * bleed);
   s = (int) gamma[0];
  }
 else
  s = gamma[0] = 0;

 /* first block handled separately to implement bleeding */
 first = (inters[16] - inters[15])*17.0 - bleed*255.0*16.0/15.0 + 0.5;
 if (first<0.0) first = 0; /* precaution against too big bleed value */
 for (j = 1; j < 16; j++)
  s += gamma[j] = (unsigned short) first;
 /*printf("bleed %f  0-%d 1-%d i1 %f i2 %f interdiff %f\n", bleed, gamma[0], gamma[1], inters[16], inters[15], inters[16] - inters[15]); */

 for (i = 1; i < 16; i++)
  for (j = 0; j < 16; j++)
   s += gamma[i * 16 + j] = (unsigned short) ((inters[15 - i + 1] - inters[15 - i]) * 16.0 + 0.5);

 if (s < 256 * 256) /* compensate for rounding */
  for (i = s - 256 * 256; i > 0; i--)
   gamma[((i * 11) % 255) + 1]++; /* scatter roundoff error through the whole array */
 /*printf("s=%d\n", s);*/
 return((unsigned char *) makeDither(gamma));
}

/* for each pixel value count the coverage of this value inside */
/* dither matrix m and store the result in the field array */
DEPSON void setDots(unsigned int *field, unsigned char *m)
{
 int i;
 unsigned int numb[256];

 for (i = 0; i < 256; i++)
  field[i] = numb[i] = 0;
 for (i = 0; i < 0x10000; i++)
  numb[m[i]]++;
 field[255] = 0x10000 - numb[0];
 for (i = 254; i >= 0; i--)
  field[i] = field[i + 1] - numb[255 - i];
 for (i = 256; i < 512; i++)
  field[i] = 0xffff;
}

